﻿/*------------------------------------------------------------------------------
* MainWindow.cpp : メイン画面クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "MainWindow.h"
#include "ui_MainWindow.h"


using namespace std;

// static
QString MainWindow::antLibLogFilename = "";
QString MainWindow::agencyForPcv = "";
QString MainWindow::agencyForAntex = "";
int MainWindow::intervalForIFB = 0;
int MainWindow::pathMinForIFB = 0;
double MainWindow::pcvSphVar[2][3] = {{0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
int MainWindow::ifbRejectLoopNum = 0;
double MainWindow::ifbRejectSigma = 0;

// アンテナ位相中心位置のINDEX
#define PC_SYSIDX_GPS   0
#define PC_SYSIDX_GLO   1

/** 推定モード
 * @brief The estMode enum
 */
enum estMode {
    EST_ANT = 0,    // アンテナ位相特性&観測モデル誤差推定
    EST_GLO         // GLONASS周波数間搬送波位相バイアス推定
};

/** コンストラクタ
 * @brief MainWindow::MainWindow
 * @param parent
 */
MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    lastCurDir = ".";
    lastOutputDir = ".";

    // 初期推定モードに応じた入力ファイルの活性/非活性の制御
    slotChangeEstMode();

    // 漢字コード変換用クラスの初期化
    pQTextCodec = QTextCodec::codecForLocale();

    // 排他準備
    initlock(&lockForSubProc);

    // グラフタイプのコンボボックスの準備
    ui->comboBoxGraphType->addItem("PCV Skyplot");
    ui->comboBoxGraphType->addItem("Code Multipath");
    ui->comboBoxGraphType->addItem("Carrier Phase Multipath");

    // ログレベル コンボボックスの準備
#ifdef TRACE
    ui->comboBoxLogLevel->addItem(pQTextCodec->toUnicode("出力なし"));
    ui->comboBoxLogLevel->addItem(pQTextCodec->toUnicode("1(エラーのみ)"));
    ui->comboBoxLogLevel->addItem(pQTextCodec->toUnicode("2(警告以上)"));
    ui->comboBoxLogLevel->addItem(pQTextCodec->toUnicode("3(Info以上)"));
    ui->comboBoxLogLevel->addItem(pQTextCodec->toUnicode("4(デバッグ)"));
    ui->comboBoxLogLevel->addItem(pQTextCodec->toUnicode("5(デバッグ[詳細])"));
    ui->comboBoxLogLevel->setCurrentIndex(2);
#else
    // TRACEなしなら制御しない
    ui->comboBoxLogLevel->setEnabled(false);
#endif

    /** アンテナ位相特性＆観測モデル誤差推定　非選択対応　2014.12.18 */
    ui->tabWidgetOption->setCurrentIndex(1);
    ui->tabWidgetOption->setTabEnabled(0,false);
    ui->tabWidgetOption->setTabText(0,"");
    ui->tabWidgetOption->removeTab(0);
    ui->radioButtonEstimatePcv->setEnabled(false);
    ui->radioButtonEstimatePcv->hide();
    ui->radioButtonEstimateGloBias->setChecked(true);
    slotChangeEstMode();
}

/** デストラクタ
 * @brief MainWindow::~MainWindow
 */
MainWindow::~MainWindow()
{
    delete ui;

    // 子プロセスのクリア
    for (int i = 0; i < (int)vecProcessExecutor.size(); i++) {
        delete(vecProcessExecutor[i]);
    }

    // ログファイルクローズ
    traceclose();
}

/** 初期化
 * @brief MainWindow::initialize
 * @param iniFilepath INIファイルへのパス
 */
void MainWindow::initialize(QString iniFilepath) {

    // [[[ 設定値取得 ]]]
    // 設定ファイルから値を取得
    QSettings settings(iniFilepath, QSettings::IniFormat);
    settings.setIniCodec(QTextCodec::codecForLocale());
    pythonPlotProgname = settings.value("PlotProgname").toString();
    antLibLogFilename = settings.value("LogFilename").toString();
    agencyForPcv = settings.value("AgencyForPcv").toString();
    agencyForAntex = settings.value("AgencyForAntex").toString();
    intervalForIFB = settings.value("IntervalForIFB").toInt();
    pathMinForIFB = settings.value("PathMinForIFB").toInt();
    QString statusBarText =settings.value("StatusBarText").toString();
    ifbRejectLoopNum = settings.value("IFBRejectLoopNum",1).toInt();
    ifbRejectSigma = settings.value("IFBRejectSigma",1).toDouble();

    // 球面調和関数の共分散(周波数別)
    pcvSphVar[0][0] = settings.value("PcvSphVarGpsL1").toDouble();  // GPS(L1)
    pcvSphVar[0][1] = settings.value("PcvSphVarGpsL2").toDouble();  // GPS(L2)
    pcvSphVar[0][2] = settings.value("PcvSphVarGpsL5").toDouble();  // GPS(L5)
    pcvSphVar[1][0] = settings.value("PcvSphVarGloRL1").toDouble(); // GLONASS(RL1)
    pcvSphVar[1][1] = settings.value("PcvSphVarGloRL2").toDouble(); // GLONASS(RL2)

    int graphNum = settings.value("GraphMax").toInt();

    // 設定ファイルから値が取れない場合(デフォルト値)
    graphNum = (graphNum <= 0)? GRAPH_MAX_NUM: graphNum;
    intervalForIFB = (intervalForIFB <= 0)? 3600: intervalForIFB;
    pathMinForIFB = (pathMinForIFB <= 0)? 300: pathMinForIFB;
    for (int sys = 0; sys < 2; sys++) {
        for (int i = 0; i < 3; i++) {
            pcvSphVar[sys][i] = (pcvSphVar[sys][i] == 0.0)? 1e-8: pcvSphVar[sys][i];
        }
    }

    // 設定ファイルからの情報を出力
    outputLog("MainWindow::initialize()");
    outputLog(" codec=" + pQTextCodec->name());
    outputLog(" PlotProgname=" + pythonPlotProgname);
    outputLog(" LogFilename=" + antLibLogFilename);
    outputLog(" AgencyForPcv=" + agencyForPcv);
    outputLog(" AgencyForAntex=" + agencyForAntex);
    outputLog(" IntervalForIFB=" + QString::number(intervalForIFB));
    outputLog(" PathMinForIFB=" + QString::number(pathMinForIFB));
    outputLog(" PcvSphVarGpsL1=" + QString::number(pcvSphVar[0][0]));
    outputLog(" PcvSphVarGpsL2=" + QString::number(pcvSphVar[0][1]));
    outputLog(" PcvSphVarGpsL5=" + QString::number(pcvSphVar[0][2]));
    outputLog(" PcvSphVarGloRL1=" + QString::number(pcvSphVar[1][0]));
    outputLog(" PcvSphVarGloRL2=" + QString::number(pcvSphVar[1][1]));
    outputLog(" GraphMax=" + QString::number(graphNum));
    if (!statusBarText.isEmpty()) {
        statusBar()->showMessage(statusBarText);
    }
    outputLog(" IFBRejectLoopNum=" + QString::number(ifbRejectLoopNum));
    outputLog(" IFBRejectSigma=" + QString::number(ifbRejectSigma));


    // [[[ 推定条件取得 ]]]
    // 観測データ(基準局/測定局)、PCVファイル、観測誤差モデルテーブル、ARP(基準局/測定局)
    ui->lineEditRefObsFile->setText(settings.value("ObsRef").toString());
    ui->lineEditTargetObsFile->setText(settings.value("ObsTag").toString());
    ui->lineEditRefPcvFile->setText(settings.value("PcvRef").toString());
    ui->lineEditNavErrFile->setText(settings.value("NavErr").toString());
    ui->lineEditRefAntARP->setText(settings.value("ArpRef").toString());
    ui->lineEditTargetAntARP->setText(settings.value("ArpTag").toString());

    // 航法データ
    settings.beginGroup("NAVS");
    QStringList navs = settings.childKeys();
    for (int i = 0; i < navs.size(); i++) {
        ui->listWidgetNavFiles->addItem(settings.value(navs.at(i)).toString());
    }


    // [[[ 初期化 ]]]
    // 子プロセス用の初期化
    for (int i = 0; i < graphNum; i++) {
        ProcessExecutor *executor = new ProcessExecutor();
        executor->initialize(this);
        vecProcessExecutor.push_back(executor);
    }

    ui->estGlo_LineEditSigma->setText(QString::number(ifbRejectSigma));
    ui->estGlo_LineEditRepeat->setText(QString::number(ifbRejectLoopNum));

}

/** グラフ表示
 * @brief MainWindow::execPlot
 * @param args 引数
 */
void MainWindow::execPlot(QStringList args) {
    /* [memo]
     * lock()/unlock()については、rtklib.hで定義しているものを使用する。
     */

    lock(&lockForSubProc);  // <<<<< LOCK >>>>>
    // 実行していないプロセスクラスを探す
    int idx = -1;
    for (int i = 0; i < (int)vecProcessExecutor.size(); i++) {
        if (!vecProcessExecutor[i]->isRunning()) {
            // 実行していないものを見つけたので抜ける
            idx = i;
            break;
        }
    }

    if (idx == -1) {
        unlock(&lockForSubProc);    // <<<<< UNLOCK >>>>>

        // 最大数以上、表示していたら警告で表示しない
        openMessageBox(MSGBOX_WARNING, "グラフ表示エラー", "表示最大数を超過しています");
        return;
    }

    // 実行(phythonの実行)
    vecProcessExecutor[idx]->start("py", args);
    unlock(&lockForSubProc);    // <<<<< UNLOCK >>>>>
}

/** ファイルダイアログをオープン
 * @brief MainWindow::openFileDialog
 * @param titleStdStr ダイアログタイトル
 * @param selectFilterStdStr ファイルフィルタ
 * @param outputEdit 選択ファイルの出力先
 */
void MainWindow::openFileDialog(string titleStdStr, string selectFilterStdStr, QLineEdit* outputEdit)
{
    QString title = pQTextCodec->toUnicode(titleStdStr.c_str());
    QString curDir = lastCurDir;
    QString selectFilter = pQTextCodec->toUnicode(selectFilterStdStr.c_str());

    // ファイルダイアログオープン
    QString targetFile = QFileDialog::getOpenFileName(this, title, curDir, selectFilter);
    if (!targetFile.isEmpty()) {
        // 空っぽではないなら取得
        outputEdit->setText(targetFile);

        // ディレクトリを取得
        int pos = targetFile.lastIndexOf("\\");
        lastCurDir = targetFile.left(pos);
    }

    // 推定ボタンの制御
    checkPreparation();
}

/** メッセージボックス表示
 * @brief MainWindow::openMessageBox
 * @param type MessageBoxタイプ
 * @param titleStdStr タイトル
 * @param messageStdStr メッセージ
 */
int MainWindow::openMessageBox(int type, string titleStdStr, string messageStdStr) {
    QString title = pQTextCodec->toUnicode(titleStdStr.c_str());
    QString msg = pQTextCodec->toUnicode(messageStdStr.c_str());

    if (type == MSGBOX_QUESTION) {
        return QMessageBox::question(this, title, msg, QMessageBox::Ok, QMessageBox::Cancel);
    } else {
        return QMessageBox::warning(this, title, msg, QMessageBox::Ok);
    }
}

/** ログ領域に出力
 * @brief MainWindow::outputLog
 * @param msg メッセージ
 * @param append true：追記
 */
void MainWindow::outputLog(QString msg, bool append) {
    // 現在時刻取得(生成時刻)
    QString timeStr = QString::fromStdString(time_str(getNowTime(), 0));

    // 一行分の文字列を生成
    QString lineStr = timeStr + " " + msg;

    if(append) {
        // 追記の場合
        ui->textEditLog->append(lineStr);

    } else {
        // 上書きの場合
        ui->textEditLog->setText(lineStr);
    }
}

/** 現在の推定モード
 * @brief MainWindow::currentEstMode
 * @return 推定モード
 */
int MainWindow::currentEstMode() {
    if( ui->radioButtonEstimatePcv->isChecked()) {
        // アンテナ位相特性推定の場合
        return EST_ANT;

    } else if( ui->radioButtonEstimateGloBias->isChecked()) {
        // GLONASS周波数間搬送波位相バイアス推定の場合
        return EST_GLO;
    }

    return 0; // (それ以外はないけど）
}

/** 推定の準備状況をチェック
 * @brief MainWindow::checkPreparation
 * @return
 */
void MainWindow::checkPreparation() {
    bool notReadyFlag = false;

    // 観測データの入力をチェック
    notReadyFlag |= ui->lineEditRefObsFile->text().isEmpty();
    notReadyFlag |= ui->lineEditTargetObsFile->text().isEmpty();

    // 航法データの入力チェック
    notReadyFlag |= (ui->listWidgetNavFiles->count() == 0);

    // アンテナ位相特性ファイルの入力チェック
    if (ui->pushButtonOpenRefPcvFile->isEnabled()) {
        notReadyFlag |= ui->lineEditRefPcvFile->text().isEmpty();
    }

    // 観測誤差モデルテーブルの入力チェック
    if (ui->pushButtonOpenNavErrFile->isEnabled()) {
        notReadyFlag |= ui->lineEditNavErrFile->text().isEmpty();
    }

    // 出力ディレクトリのチェック
    bool nanOutDir = ui->lineEditOutputDir->text().isEmpty();
    notReadyFlag |= nanOutDir;

    // 推定ボタンの制御
    ui->pushButtonEstimate->setEnabled(!notReadyFlag);

    // グラフ表示ボタンは出力ディレクトリで判定する
    updateComboBoxForGraph();
}

/** ファイルの存在チェック
 * @brief MainWindow::checkInputFile
 * @param lineEdit チェックする入力欄
 * @return true：OK、false：ファイルがない
 */
bool MainWindow::checkInputFile(QLineEdit* lineEdit) {
    return checkInputFile(lineEdit->text());
}

/** ファイルの存在チェック
 * @brief MainWindow::checkInputFile
 * @param filepath ファイル名
 * @return true：OK、false：ファイルがない
 */
bool MainWindow::checkInputFile(QString filepath) {
    if (QFile::exists(filepath)) {
        return true;
    }

    // ファイルが存在しない旨をダイアログで通知
    string msg = "ファイルが存在しません\n";
    msg.append(pQTextCodec->fromUnicode(filepath));
    openMessageBox(MSGBOX_WARNING, "ファイルエラー", msg);
    return false;
}

/** ディレクトリの存在チェック
 * @brief MainWindow::checkDirectory
 * @param lineEdit チェックする入力欄
 * @return true：OK、false：NG
 */
bool MainWindow::checkDirectory(QLineEdit* lineEdit) {
    QDir dir(lineEdit->text());
    if (dir.exists()) {
        return true;
    }

    // ディレクトリがない場合は作成するか確認
    string msg = "次のディレクトリは存在しません。作成しますか？\n";
    msg.append(pQTextCodec->fromUnicode(lineEdit->text()));
    int ret = openMessageBox(MSGBOX_QUESTION, "ディレクトリの作成確認", msg);
    if (ret == QMessageBox::Cancel) {
        // キャンセルならエラーで返す
        return false;
    }

    // ディレクトリ作成
    dir.mkpath(".");
    return true;
}

/** アンテナ位相中心位置取得
 * @brief MainWindow::getPhaseCenter
 * @param lineEdit チェックする入力欄
 * @param posStr 位置情報
 * @param retPos [OUT]位置
 * @param noCheck true：エラーダイアログなし
 * @return true：OK、false：NG
 */
bool MainWindow::getPhaseCenter(QLineEdit* lineEdit, string posStr, double *retPos, bool noCheck) {
    if (lineEdit->text().isEmpty()) {
        if (!noCheck) {
            // 入力が存在しない旨をダイアログで通知
            string msg = "値が設定されていません\n";
            msg.append(posStr);
            openMessageBox(MSGBOX_WARNING, "設定エラー", msg);
        }
        return false;
    }

    // アンテナ位置を取得
    bool ret = setAntennaPosition(lineEdit->text(), retPos);
    if (!ret) {
        if (!noCheck) {
            // 値が不正な旨をダイアログで通知
            string msg = "入力値が不正です\n";
            msg.append(posStr);
            openMessageBox(MSGBOX_WARNING, "設定エラー", msg);
        }
        return false;
    }

    return true;
}

/** グラフ選択コンボボックスの生成
 * @brief MainWindow::createComboBoxForGraph
 */
void MainWindow::updateComboBoxForGraph() {
    ui->pushButtonViewGraph->setEnabled(false);

    int estMode = currentEstMode();
    if (estMode == EST_GLO) {
        // グラフ表示なしの場合、何もしない
        return;
    }
    QString outputDir = ui->lineEditOutputDir->text();
    if (outputDir.isEmpty()) {
        // 出力ディレクトリの指定がないなら、何もしない
        return;
    }

    // コンボボックスクリア
    ui->comboBoxGraph->clear();
    vecGraphParam.clear();

    // グラフタイプ取得
    int graphType = ui->comboBoxGraphType->currentIndex();

    // ファイルからグラフリスト取得
    const string graphTypeStr[3] = {PCV_RESULT_PREFIX, CMP_RESULT_PREFIX, PMP_RESULT_PREFIX};
    vecGraphParam = getGraphList(graphTypeStr[graphType], outputDir, graphType + 1);

    // コンボボックス作成
    for (int i = 0; i < (int)vecGraphParam.size(); i++) {
        QString str = vecGraphParam[i].sysStr + "-" + vecGraphParam[i].freqStr;
        ui->comboBoxGraph->addItem(str);
    }

    // コンボボックスを活性化
    ui->comboBoxGraph->setEnabled(true);

    // グラフ表示ボタンを活性/非活性化
    ui->pushButtonViewGraph->setEnabled(vecGraphParam.size() > 0);
}

/** グラフリスト取得
 * @brief MainWindow::getGraphList
 * @param graphType
 * @param outputDir
 * @param plotMode グラフモード(1:アンテナ位相特性、2:コードマルチパス、3：搬送波位相マルチパス)
 * @return
 */
vector<graphParameter_t> MainWindow::getGraphList(string graphType, QString outputDir, int plotMode) {

    // 検索対象の準備
    QStringList strListFilter;
    strListFilter << QString::fromStdString(graphType) + "*" + QString::fromStdString(RESULT_SUFFIX);

    //　ディレクトリ内のファイル検索
    QDir dir(outputDir);
    QFileInfoList list = dir.entryInfoList(strListFilter, QDir::Files);

    // ファイルリストの取得
    vector<graphParameter_t> result;
    for (int i = 0; i < list.size(); i++) {
        QFileInfo file = list.at(i);
        QString filename = file.fileName();

        // ファイル名の分析
        QStringList strList = filename.split("_");
        QString sysStr = strList.at(2);
        QString freqStr = strList.at(3);
        freqStr.replace(RESULT_SUFFIX, "");

        // ファイル追加
        graphParameter_t param;
        param.dataFilepath = file.filePath();
        param.plotMode = plotMode;
        param.sysStr = sysStr;
        param.freqStr = freqStr;
        result.push_back(param);
    }

    return result;
}

// +----+----+----+----+----+----+----+----+----+----+----+----+
// slot
/** (slot)ファイルダイアログをオープン(基準局観測データファイル)
 * @brief MainWindow::slotOpenRefObsFile
 */
void MainWindow::slotOpenRefObsFile() {
    openFileDialog("基準局観測データファイル選択",
                   "観測データ(*.obs *O);;全て(*.*)",
                   ui->lineEditRefObsFile);
}

/** (slot)ファイルダイアログをオープン(測定局観測データファイル)
 * @brief MainWindow::slotOpenTargetObsFile
 */
void MainWindow::slotOpenTargetObsFile() {
    openFileDialog("測定局観測データファイル選択",
                   "観測データ(*.obs *O);;全て(*.*)",
                   ui->lineEditTargetObsFile);
}

/** (slot)ファイルダイアログをオープン(基準局アンテナ位相特性ファイル)
 * @brief MainWindow::slotOpenRefPcvFile
 */
void MainWindow::slotOpenRefPcvFile() {
    openFileDialog("基準局アンテナ位相特性ファイル選択",
                   "アンテナ位相特性ファイル(*.pcv *.atx);;全て(*.*)",
                   ui->lineEditRefPcvFile);
}

/** (slot)ファイルダイアログをオープン(観測残差モデルテーブルファイル)
 * @brief MainWindow::slotOpenNavErrFile
 */
void MainWindow::slotOpenNavErrFile() {
    openFileDialog("slotOpenNavErrFile",
                   "観測残差モデルテーブル(*.tbl);;全て(*.*)",
                   ui->lineEditNavErrFile);
}

/** (slot)出力ディレクトリを選択
 * @brief MainWindow::slotSelectOutputDir
 */
void MainWindow::slotSelectOutputDir() {
    QString title = pQTextCodec->toUnicode("出力ディレクトリ選択");
    QString curDir = lastOutputDir;

    // フォルダ選択ダイアログをオープン
    QString targetDir = QFileDialog::getExistingDirectory(this, title, curDir,
                                                          QFileDialog::ShowDirsOnly);
    if (!targetDir.isEmpty()) {
        // 空っぽではないなら取得
        ui->lineEditOutputDir->setText(targetDir);
        lastOutputDir = targetDir;
    }

    // 推定ボタンの制御
    checkPreparation();
}

/** (Slot)航法データ追加
 * @brief MainWindow::slotAddNavFile
 */
void MainWindow::slotAddNavFile() {
    QString title = pQTextCodec->toUnicode("航法データ追加");
    QString curDir = lastCurDir;
    QString selectFilter = pQTextCodec->toUnicode("Navigationファイル(*.nav *N *G *Q);;全て(*.*)");

    // ファイルダイアログオープン
    QStringList files = QFileDialog::getOpenFileNames(this, title, curDir, selectFilter);
    if (files.size() > 0) {
        // 空っぽではないなら取得
        ui->listWidgetNavFiles->addItems(files);

        // ディレクトリを取得
        int pos = files[0].lastIndexOf("\\");
        lastCurDir = files[0].left(pos);
    }

    // 推定ボタンの制御
    checkPreparation();
}

/** (Slot)航法データ削除
 * @brief MainWindow::slotRemoveNavFile
 */
void MainWindow::slotRemoveNavFile() {
    // 確認ダイアログをオープン
    int ret = openMessageBox(MSGBOX_QUESTION, "選択項目削除", "選択された項目を削除しますか？");
    if (ret == QMessageBox::Cancel) {
        // キャンセルなら何もしない
        return;
    }

    // 選択されている項目を削除
    QList<QListWidgetItem*> list = ui->listWidgetNavFiles->selectedItems();
    for (int i = 0; i < list.size(); i++) {
        for (int j = 0; j < ui->listWidgetNavFiles->count(); j++) {
            QListWidgetItem *item = ui->listWidgetNavFiles->item(j);
            if (list[i] == item) {
                // 同じものなら削除
                ui->listWidgetNavFiles->takeItem(j);
                break;
            }
        }
    }

    // 推定ボタンの制御
    checkPreparation();
}

/** (Slot)入力ファイルを直接修正
 * @brief MainWindow::slotTextChanged
 */
void MainWindow::slotTextChanged() {
    // 推定ボタンの制御
    checkPreparation();
}

/** (Slot)ラジオボタン変更(推定モード変更)
 * @brief MainWindow::slotChangeEstMode
 */
void MainWindow::slotChangeEstMode() {
    int estModeIdx = currentEstMode();

    // オプションタブを変更
    ui->tabWidgetOption->setCurrentIndex(estModeIdx);

    // 入力ファイルの活性/非活性制御
    if (estModeIdx == EST_ANT) {
        // アンテナ位相特性推定の場合
        ui->pushButtonOpenRefPcvFile->setEnabled(true);
        ui->lineEditRefPcvFile->setEnabled(true);

        ui->pushButtonOpenNavErrFile->setEnabled(true);
        ui->lineEditNavErrFile->setEnabled(true);

    } else if (estModeIdx == EST_GLO) {
        // GLONASS周波数間搬送波位相バイアス推定の場合
        ui->pushButtonOpenRefPcvFile->setEnabled(false);
        ui->lineEditRefPcvFile->setEnabled(false);

        ui->pushButtonOpenNavErrFile->setEnabled(false);
        ui->lineEditNavErrFile->setEnabled(false);

        ui->comboBoxGraph->clear();
    }

    // 推定ボタンの制御
    checkPreparation();
}

/** (Slot)ログレベル変更
 * @brief MainWindow::slotChangeLogLevel
 */
void MainWindow::slotChangeLogLevel() {
    // ログレベルを設定
    int level = ui->comboBoxLogLevel->currentIndex();
    tracelevel(level);
    AntLibBase::setLogLevel(level);
}

/** (Slot)グラフタイプ変更
 * @brief MainWindow::slotChangeGraphType
 */
void MainWindow::slotChangeGraphType() {
    updateComboBoxForGraph();
}

/** (Slot)推定ボタン押下
 * @brief MainWindow::slotEstimate
 */
void MainWindow::slotEstimate() {
    // 入力ファイルのチェック(共通)
    if (ui->lineEditRefObsFile->text() == ui->lineEditTargetObsFile->text()) {
        // 基準局と測定局のファイルが同じ
        openMessageBox(MSGBOX_WARNING, "ファイルエラー", "基準局と測定局の観測データが同じです");
        return;
    }
    if (!checkInputFile(ui->lineEditRefObsFile)) {  // 観測データ(基準局)
        return;
    }
    if (!checkInputFile(ui->lineEditTargetObsFile)) {  // 観測データ(測定局)
        return;
    }
    for (int i = 0; i < ui->listWidgetNavFiles->count(); i++) { // 航法データ
        QListWidgetItem *item = ui->listWidgetNavFiles->item(i);
        if (!checkInputFile(item->text())) {
            return;
        }
    }
    if (ui->lineEditRefPcvFile->isEnabled()) {
        if (!checkInputFile(ui->lineEditRefPcvFile)) {  // アンテナ位相特性ファイル
            return;
        }
    }
    if (ui->lineEditNavErrFile->isEnabled()) {
        if (!checkInputFile(ui->lineEditNavErrFile)) {  // 観測誤差モデルテーブル
            return;
        }
    }

    // 出力フォルダのチェック
    if (!checkDirectory(ui->lineEditOutputDir)) {
        return;
    }

    // ARPのチェック
    double work[3];
    if (!getPhaseCenter(ui->lineEditRefAntARP, "基準局のアンテナ位置(ARP)", work)) {
        return;
    }
    if (!getPhaseCenter(ui->lineEditTargetAntARP, "測定局のアンテナ位置(ARP)", work)) {
        return;
    }

    // 推定文字列取得＆入力値チェック
    string errTitle = "オプション・エラー";
    int estModeIdx = currentEstMode();
    QByteArray titleByte = "N/A";
    if (estModeIdx == EST_ANT) {
        titleByte = "アンテナ位相特性推定";

        // [[[ アンテナ位相特性推定のチェック ]]]
        // アンテナ位相特性ファイルのチェック
        if (!checkInputFile(ui->lineEditRefPcvFile)) {
            return;
        }

        // 球面調和関数の範囲チェック
        int nMax = ui->antEst_LineEditShpereOrder->text().toInt();
        int mMax = ui->antEst_LineEditSphereAngle->text().toInt();
        if (nMax <= 0 || nMax > ANTEST_NMAX) {
            ostringstream oStr;
            oStr << "球面調和関数オーダの値が不正です(1-" <<  ANTEST_NMAX << ")";
            openMessageBox(MSGBOX_WARNING, errTitle, oStr.str());
            return;
        }
        if (mMax <= 0 || mMax > nMax) {
            openMessageBox(MSGBOX_WARNING, errTitle, "球面調和関数角の値が不正です");
            return;
        }

        // [[[ 観測誤差モデル推定のチェック ]]]
        // 観測残差モデルテーブルのチェック
        if (!checkInputFile(ui->lineEditNavErrFile)) {
            return;
        }

        // 単純移動平均の対象データ数の範囲チェック
        int rangeForAMP = ui->navErr_LineEditAmpRange->text().toInt();
        if (rangeForAMP <= 0 || rangeForAMP > 100) {
            openMessageBox(MSGBOX_WARNING, errTitle, "単純移動平均の対象データ数の値が不正です(1-100)");
            return;
        }

        // 観測更新の単位とする仰角範囲の範囲チェック
        double deltaBeta = ui->navErr_LineEditDeltaBeta->text().toDouble();
        if (deltaBeta <= 0 || deltaBeta > 2) {
            openMessageBox(MSGBOX_WARNING, errTitle, "単位とする仰角範囲の値が不正です(0<,<=2)");
            return;
        }
    } else if (estModeIdx == EST_GLO) {
        titleByte = "GLONASS周波数間搬送波位相バイアス推定";
    }

    QString title = pQTextCodec->toUnicode(titleByte);
    QString labelStr = pQTextCodec->toUnicode("しばらくお待ちください");


    // 推定ボタンを非活性(連打防止)
    ui->pushButtonEstimate->setEnabled(false);


    // 進捗ダイアログ準備
    // ※フラグで閉じるボタンをOFFする
    QProgressDialog dialog(this, Qt::CustomizeWindowHint);
    dialog.setWindowTitle(title);
    dialog.setLabelText(labelStr);
    dialog.setRange(0, 0);
    dialog.setCancelButton(NULL);
    dialog.setWindowModality(Qt::WindowModal);
    dialog.resize(350, 50);

    // スレッドの準備＆進捗ダイアログとの接続
    QFutureWatcher<QString> futureWatcher;
    QObject::connect(&futureWatcher, SIGNAL(finished()), &dialog, SLOT(reset()));

    // 推定処理実施
    std::vector<QString> warningMessage;
    outputLog(title + pQTextCodec->toUnicode("を開始します"), false);
    QFuture<QString> future = QtConcurrent::run(&MainWindow::estimate, ui, estModeIdx, &warningMessage);
    futureWatcher.setFuture(future);

    // 進捗ダイアログオープン
    dialog.exec();

    // 終了待ち
    futureWatcher.waitForFinished();

    // 推定ボタンを活性
    ui->pushButtonEstimate->setEnabled(true);

    // 推定結果を表示
    QString errMessage = future.result();
    if (errMessage.isEmpty()) {
        // 警告メッセージ出力
        for (int i = 0; i < (int)warningMessage.size(); i++) {
            outputLog(warningMessage[i]);
        }

        // 終了メッセージ(正常)
        outputLog(title + pQTextCodec->toUnicode("を終了しました"));

    } else {
        // エラー時はログに出力
        outputLog(errMessage);

        // 終了メッセージ(失敗)
        outputLog(title + pQTextCodec->toUnicode("が異常終了しました"));
    }

    // 終了処理
    updateComboBoxForGraph();
}

/** (Slot)グラフ表示
 * @brief MainWindow::slotViewGraph
 */
void MainWindow::slotViewGraph() {
    if (vecGraphParam.size() == 0) {
        // このケースはないけど、一応エラー処理
        return;
    }

    // グラフ情報を取得
    int idx = ui->comboBoxGraph->currentIndex();
    graphParameter_t param = vecGraphParam[idx];

    // ファイルの存在チェック
    if (!QFile::exists(param.dataFilepath)) {
        // ファイルが存在しない場合
        outputLog("not find file. filepath=" + param.dataFilepath);
        openMessageBox(MSGBOX_WARNING, "グラフ表示エラー", "プロットデータのファイルが存在しません");
        return;
    }

    // 引数生成
    QStringList args;
    args << pythonPlotProgname
            << param.dataFilepath
            << QString::number(param.plotMode)
            << param.freqStr;

    // 実行
    outputLog(args.join(' '));
    execPlot(args);
}

/** (Slot)終了
 * @brief MainWindow::slotExit
 */
void MainWindow::slotExit() {
    // 確認ダイアログをオープン
    int ret = openMessageBox(MSGBOX_QUESTION, "終了", "プログラムを終了しますか？");
    if (ret == QMessageBox::Cancel) {
        // キャンセルなら何もしない
        return;
    }

    close();
}


// +----+----+----+----+----+----+----+----+----+----+----+----+
/** ファイルパスからファイル名の取得
 * @brief getFilenameOnly
 * @param filepath ファイルパス
 * @return ファイル名
 */
QString MainWindow::getFilenameOnly(QString filepath) {
    QStringList list = filepath.split("/");
    int lastIdx = list.size() - 1;
    return list.at(lastIdx);
}

/** 推定処理
 * @brief MainWindow::estimate
 * @param ui
 * @param estModeIdx 推定モード
 * @param warningMessage [OUT] 警告メッセージ
 * @return エラーメッセージ
 */
QString MainWindow::estimate(Ui::MainWindow *ui, int estModeIdx, vector<QString> *warningMessage) {
    AntLibMain antLibMain;

    // 出力先
    QString outputDirQStr = ui->lineEditOutputDir->text();
    outputDirQStr.append("/");
    string outputDir = toStdStringFileSystem(outputDirQStr);
    antLibMain.outputDir = outputDir;

    // ログオープン
    ofstream logStream;
    if (!antLibLogFilename.isEmpty()) {
        QString logFilepath = outputDirQStr + antLibLogFilename;
        logStream.open(toStdStringFileSystem(logFilepath));

        // Logストリームセット
        AntLibBase::setStream(&logStream);
    }

    // 値確認用データの出力先をチェック
#ifdef DEBUG_OUTPUT_DATA
    QTextCodec *codec = QTextCodec::codecForLocale();
    for (int i = 0; i < 3; i++) {
        string outputDataDir = outputDir + "data_" + antLibMainFreqName[i] + "/";
        QDir dir(codec->toUnicode(outputDataDir.c_str()));
        if (!dir.exists()) {
            dir.mkdir(".");
        }
    }
#endif

    // 初期化
    antLibMain.initialize();


    // [[[ 入力ファイルパスの取得(共通項目) ]]]
    // 観測データ(基準局)
    vector<string> vecRefObsFile;
    vecRefObsFile.push_back(toStdStringFileSystem(ui->lineEditRefObsFile->text()));

    // 観測データ(測定局)
    vector<string> vecTargetObsFile;
    vecTargetObsFile.push_back(toStdStringFileSystem(ui->lineEditTargetObsFile->text()));

    // 航法データ
    vector<string> vecNavFile;
    for (int i = 0; i < ui->listWidgetNavFiles->count(); i++) {
        QListWidgetItem *item = ui->listWidgetNavFiles->item(i);
        vecNavFile.push_back(toStdStringFileSystem(item->text()));
    }

    // ログレベル
    int logLevel = ui->comboBoxLogLevel->currentIndex();

    // 推定別処理
    QString errMessage = "";
    try {
        if (estModeIdx == EST_ANT) {
            // アンテナ位相中心フラグ保管用(true：算出)
            bool targetPhaseCenterFlag[2][3] = {{false, false, false}, {false, false, false}};

            //アンテナ位相中心保管用
            double targetPhaseCenter[2][3][3] = {{{0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}},
                                                {{0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}}};

            // [[[ アンテナ位相特性推定の場合 ]]]
            // アンテナ特性ファイル
            QString antFilepath = ui->lineEditRefPcvFile->text();
            QString antFilename = getFilenameOnly(antFilepath);
            string pcvFilepath = toStdStringFileSystem(antFilepath);
            string outputPcvFilepath = outputDir + toStdStringFileSystem(antFilename);

            // ファイルコピー
            copyFile(antFilepath, outputDirQStr + antFilename);

            // 局位置設定(ARP)
            setAntennaPosition(&antLibMain, ui->lineEditRefAntARP->text(), RECEIVER_NO_REFERENCE);
            setAntennaPosition(&antLibMain, ui->lineEditTargetAntARP->text(), RECEIVER_NO_TARGET);

            // オプションの取得
            double minForSN = ui->antEst_LineEditCutOffSNR->text().toDouble();
            int nMax = ui->antEst_LineEditShpereOrder->text().toInt();
            int mMax = ui->antEst_LineEditSphereAngle->text().toInt();
            double azOffset = ui->antEst_LineEditOffsetOfAntAZ->text().toDouble();
            double elMin = ui->antEst_LineEditCutOffEL->text().toDouble();
            double thresslip = ui->antEst_LineEditSlipTh->text().toDouble();


            // パラメータチェック
            if (nMax <= 0 || nMax > ANTEST_NMAX) {
                ostringstream oStr;
                oStr << "invalid nMax. nMax=" << nMax;
                throw AntLibException(oStr.str(), "MainWindow::estimate");
            }
            if (mMax <= 0 || mMax > ANTEST_NMAX) {
                ostringstream oStr;
                oStr << "invalid mMax. mMax=" << mMax;
                throw AntLibException(oStr.str(), "MainWindow::estimate");
            }


            // 推定処理
            antLibMain.estimatePCV(vecRefObsFile, vecTargetObsFile, vecNavFile,
                                   pcvFilepath, outputPcvFilepath,
                                   minForSN, nMax, mMax,
                                   azOffset, elMin, thresslip, pcvSphVar,
                                   agencyForPcv.toUtf8().constData(), agencyForAntex.toUtf8().constData(), logLevel,
                                   targetPhaseCenterFlag, targetPhaseCenter);

            // [[[ 観測モデル誤差推定 ]]]
            // 観測誤差モデルテーブル
            QString navErrFilepath = ui->lineEditNavErrFile->text();
            QString navErrFilename = getFilenameOnly(navErrFilepath);
            string orgNavErrFilepath = toStdStringFileSystem(navErrFilepath);
            string outputNavErrFilepath = outputDir + toStdStringFileSystem(navErrFilename);

            // ファイルコピー
            copyFile(navErrFilepath, outputDirQStr + navErrFilename);

            // オプションの取得
            int rangeForAMP = ui->navErr_LineEditAmpRange->text().toInt();
            double deltaBeta = ui->navErr_LineEditDeltaBeta->text().toDouble();
            double sigmaA = ui->navErr_LineEditSigmaA->text().toDouble();
            double sigmaB = ui->navErr_LineEditSigmaB->text().toDouble();
            double minEl = ui->navErr_LineEditMinEL->text().toDouble();
            double maxEl = ui->navErr_LineEditMaxEL->text().toDouble();

            // 推定処理
            antLibMain.estimateMultipass(vecRefObsFile, vecTargetObsFile, vecNavFile,
                                         orgNavErrFilepath, outputNavErrFilepath,
                                         rangeForAMP, deltaBeta, sigmaA, sigmaB,
                                         thresslip, minEl, maxEl,
                                         targetPhaseCenterFlag, targetPhaseCenter);

        } else if (estModeIdx == EST_GLO) {
            // [[[ GLONASS周波数間搬送波位相バイアス推定の場合 ]]]
            // 局位置設定(ARP)
            setAntennaPosition(&antLibMain, ui->lineEditRefAntARP->text(), RECEIVER_NO_REFERENCE);
            setAntennaPosition(&antLibMain, ui->lineEditTargetAntARP->text(), RECEIVER_NO_TARGET);

            // バイアス推定ファイル
            string outputEstBiasFile = outputDir + ANTLIB_GLO_IFB_FILENAME;

            // オプションの取得
            double sigmaPhai = ui->estGlo_LineEditSigmaPhai->text().toDouble();
            double sigmaX = ui->estGlo_LineEditSigmaX->text().toDouble();
            double thresslip = ui->estGlo_LineEditSlipTh->text().toDouble();
            double sigma = ui->estGlo_LineEditSigma->text().toDouble();
            int repeat = ui->estGlo_LineEditRepeat->text().toInt();

            // 推定処理
            antLibMain.estimateGlonass(vecRefObsFile, vecTargetObsFile, vecNavFile,
                                       outputEstBiasFile, sigmaPhai, sigmaX, thresslip,
                                       intervalForIFB, pathMinForIFB,
                                       sigma, repeat);
        }
    } catch(AntLibException ex) {
        QString pos = QString::fromStdString(ex.getPosition());
        QString msg = QString::fromStdString(ex.getMessage());
        errMessage = "[error] " + pos + "\n" + msg;
    }

    // 警告メッセージ取得(エラーがないときのみ)
    if (errMessage.isEmpty()) {
        for (int i = 0; i < (int)antLibMain.vecMessage.size(); i++) {
            warningMessage->push_back(QString::fromStdString(antLibMain.vecMessage[i]));
        }
    }

    // ログクローズ＆Logストリームクリア
    if (logStream.is_open()) {
        logStream.close();
    }
    AntLibBase::clearStream();

    return errMessage;
}

/** 使用システムを取得
 * @brief MainWindow::getSys
 * @param comboBox コンボボックス
 * @return 使用システム
 */
int MainWindow::getSys(QComboBox *comboBox) {
    // initializeComboForSysの並びと合わせること。
    int sysMode[] = {SYS_GPS, SYS_GLO, SYS_QZS, SYS_ALL};
    int idx = comboBox->currentIndex();
    return sysMode[idx];
}

/** 局位置設定
 * @brief MainWindow::setAntennaPosition
 * @param antLibMain 推定クラス
 * @param str 局位置(文字列)
 * @param target 対象局
 */
void MainWindow::setAntennaPosition(AntLibMain *antLibMain, QString &str, int target) {
    // 局位置取得
    double pos[3];
    if (!setAntennaPosition(str, pos)) {
        return;
    }

    // 局位置設定
    if (target == RECEIVER_NO_TARGET) {
        antLibMain->setTargetPosition(pos[0], pos[1], pos[2]);
    } else {
        antLibMain->setReferencePosition(pos[0], pos[1], pos[2]);
    }
}

/** 局位置設定
 * @brief MainWindow::setAntennaPosition
 * @param str 局位置(文字列)
 * @param pos [OUT]局位置(配列)
 * @return true:指定あり
 */
bool MainWindow::setAntennaPosition(QString &str, double *pos) {
    QStringList list = str.split(",");
    if (list.size() < 3) {
        // 要素が足りないなら無視
        return false;
    }

    // 局位置取得
    for (int i = 0; i < 3; i++) {
        pos[i] = list.at(i).toDouble();
    }
    return true;
}

/** QStringからstringへの変換(ファイルシステム用)
 * @brief MainWindow::toStdStringFileSystem
 * @param str 文字列(QString)
 * @return 文字列(string)
 */
string MainWindow::toStdStringFileSystem(const QString str) {
    QTextCodec *codec = QTextCodec::codecForLocale();
    QByteArray qByte = codec->fromUnicode(str);
    return string(qByte);
}

/** ファイルのコピー処理(上書き)
 * @brief MainWindow::copyFile
 * @param from コピー元(ファイルパス)
 * @param to コピー先(ファイルパス)
 */
void MainWindow::copyFile(QString from, QString to) {

    // コピー先にファイルがある場合は、削除する
    // ※QFile::copyでは上書きできないため
    if (QFile::exists(to)) {
        QFile::remove(to);
    }

    // コピー処理
    QFile::copy(from, to);
}
