﻿/*------------------------------------------------------------------------------
* MainWindow.h : メイン画面クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#define NOMINMAX        // "windows.h"内でmin/maxのマクロ定義の回避
#include <winsock2.h>   // QtCoreでwinsock.hを呼び出しているので、先にインクルードする

#include <QtCore>
#include <QtConcurrent/QtConcurrent>
#include <QMainWindow>
#include <QLineEdit>
#include <QFileDialog>
#include <QTextCodec>
#include <QMessageBox>
#include <QProgressDialog>
#include <QComboBox>
#include <QTextEdit>
#include <QSettings>
#include <QWaitCondition>
#include <QMutex>
#include "rtklib.h"
#include "ProcessExecutor.h"
#include "AntLibMain.h"

#define GRAPH_MAX_NUM     50    // グラフ表示最大数(デフォルト)
#define MSGBOX_QUESTION    1    // MessageBox:質問
#define MSGBOX_WARNING     2    // MessageBox:警告

#define ANTLIB_INIFILENAME "AntApp.ini"         // 設定ファイル名
#define ANTLIB_GLO_IFB_FILENAME  "gloifb.tbl"   // GLONASS IFB テーブル
#define ANTLIB_ANTEST_SPH_PREFIX "sph_"         // アンテナ位相推定のアプリオリ誤差共分散行列ファイル名(接頭文字列)


/** グラフ表示用情報
 * @brief The graphParameter_t struct
 */
struct graphParameter_t {
    /** データファイルへのパス
     * @brief dataFilepath
     */
    QString dataFilepath;

    /** プロットモード
     * @brief plotMode
     */
    int plotMode;

    /** 衛星システム文字列
     * @brief sysStr
     */
    QString sysStr;

    /** 周波数文字列
     * @brief freqStr
     */
    QString freqStr;
};

namespace Ui {
class MainWindow;
}

/** メイン画面
 * @brief The MainWindow class
 */
class MainWindow : public QMainWindow
{
    Q_OBJECT
    
public:
    explicit MainWindow(QWidget *parent = 0);
    ~MainWindow();

    void initialize(QString iniFilepath);
    void execPlot(QStringList args);

private:
    void openFileDialog(std::string titleStdStr, std::string selectFilterStdStr, QLineEdit* outputEdit);
    int openMessageBox(int type, std::string titleStdStr, std::string messageStdStr);
    void outputLog(QString msg, bool append = true);
    int currentEstMode();
    void checkPreparation();
    bool checkInput(QLineEdit* lineEdit);
    bool checkInputFile(QLineEdit* lineEdit);
    bool checkInputFile(QString filepath);
    bool checkDirectory(QLineEdit* lineEdit);
    bool getPhaseCenter(QLineEdit* lineEdit, std::string posStr, double *retPos, bool noCheck = false);
    void updateComboBoxForGraph();
    std::vector<graphParameter_t> MainWindow::getGraphList(std::string graphType, QString outputDir, int plotMode);

private slots:
    void slotOpenRefObsFile();
    void slotOpenTargetObsFile();
    void slotOpenRefPcvFile();
    void slotOpenNavErrFile();
    void slotSelectOutputDir();
    void slotAddNavFile();
    void slotRemoveNavFile();
    void slotTextChanged();
    void slotChangeEstMode();
    void slotChangeLogLevel();
    void slotChangeGraphType();
    void slotEstimate();
    void slotViewGraph();
    void slotExit();


private:
    // static
    static QString getFilenameOnly(QString filepath);
    static QString estimate(Ui::MainWindow *ui, int estModeIdx, std::vector<QString> *warningMessage);
    static int getSys(QComboBox *comboBox);
    static void setAntennaPosition(AntLibMain *antLibMain, QString &str, int target);
    static bool setAntennaPosition(QString &str, double *pos);
    static std::string toStdStringFileSystem(const QString str);
    static void copyFile(QString from, QString to);

    /** メイン画面
     * @brief ui
     */
    Ui::MainWindow *ui;

    /** 最後に選択されたディレクトリ
     * @brief lastCurDir
     */
    QString lastCurDir;

    /** 最後に選択された出力ディレクトリ
     * @brief lastOutputDir
     */
    QString lastOutputDir;

    /** グラフ表示用Pythonファイル名
     * @brief pythonPlotProgname
     */
    QString pythonPlotProgname;

    /** グラフ表示用プロセス
     * @brief arrayProcess
     */
    std::vector<ProcessExecutor*> vecProcessExecutor;

    /** アンテナ位相特性ファイル名
     * @brief outFilenameForAnt
     */
    QString outFilenameForAnt;

    /** 排他処理用
     * @brief lockForSubProc
     */
    lock_t lockForSubProc;

    /** グラフ表示情報
     * @brief vecGraphParam
     */
    std::vector<graphParameter_t> vecGraphParam;

    /** 漢字コード変換用
     * @brief codec
     */
    QTextCodec *pQTextCodec;

    // static
    /** AntLib用ログファイル名
     * @brief antLibLogFilename
     */
    static QString antLibLogFilename;

    /** agency(PCVファイル用)
     * @brief agencyForPcv
     */
    static QString agencyForPcv;

    /** agency(ANTEXファイル用)
     * @brief agencyForAntex
     */
    static QString agencyForAntex;

    /** 統計処理間隔[sec](GLONASS IFB推定)
     * @brief intervalForIFB
     */
    static int intervalForIFB;

    /** 有効パス最小時間[sec](GLONASS IFB推定)
     * @brief pathMinForIFB
     */
    static int pathMinForIFB;

    /** PCV推定(球面調和関数)の共分散
     * @brief pcvSphVal
     */
    static double pcvSphVar[2][3];

    /** IFB棄却処理繰り返し回数
     * @brief
     */
    static int ifbRejectLoopNum;

    /** IFB棄却処理σ
     * @brief
     */
    static double ifbRejectSigma;

};

#endif // MAINWINDOW_H
