﻿/*------------------------------------------------------------------------------
* ProcessExecutor.cpp : プロセス実行クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "ProcessExecutor.h"

/** コンストラクタ
 * @brief ProcessExecutor::ProcessExecutor
 */
ProcessExecutor::ProcessExecutor()
{
    mainWindow = NULL;
    QObject::connect(&process, SIGNAL(readyReadStandardError()), this, SLOT(slotStdError()));
    QObject::connect(&process, SIGNAL(error(QProcess::ProcessError)), this, SLOT(slotProcError(QProcess::ProcessError)));
}

/** デストラクタ
 * @brief ProcessExecutor::~ProcessExecutor
 */
ProcessExecutor::~ProcessExecutor()
{
    mainWindow = NULL;
    if (isRunning()) {
        // 実行していたら終了する
        process.kill();
    }
}

/** 初期化処理
 * @brief ProcessExecutor::initialize
 * @param window 画面クラス
 */
void ProcessExecutor::initialize(QMainWindow *window) {
    mainWindow = window;
}

/** プログラム実行(非同期)
 * @brief ProcessExecutor::start
 * @param prog 実行するプログラム
 * @param args 引数
 */
void ProcessExecutor::start(QString prog, QStringList args) {
    progName = prog;
    progArgs = args;

    // プログラム実行(非同期)
    process.start(prog, args);
}

/** プログラム動作状態
 * @brief ProcessExecutor::isRunning
 * @return true：実行中
 */
bool ProcessExecutor::isRunning() {
    return (process.state() != QProcess::NotRunning);
}

/** 警告メッセージ
 * @brief ProcessExecutor::warning
 * @param titleByte タイトル
 * @param messageByte メッセージ
 * @param listDetail メッセージ(詳細な情報)
 */
void ProcessExecutor::warning(QByteArray titleByte, QByteArray messageByte, QStringList &listDetail) {
    if (mainWindow == NULL) {
        // ウィンドウの指定がない場合、何もしない
        return;
    }

    QTextCodec *codec = QTextCodec::codecForLocale();
    QString title = codec->toUnicode(titleByte);
    QString msg = codec->toUnicode(messageByte);

    // 詳細な情報を追加
    for (int i = 0; i < listDetail.size(); i++) {
        msg += "\n" + listDetail.at(i);
    }

    QMessageBox::warning(mainWindow, title, msg, QMessageBox::Ok);
}

/** (slot)エラー発生時
 * @brief ProcessExecutor::slotStdError
 */
void ProcessExecutor::slotStdError() {
    // 標準エラー出力を取得
    QStringList list;
    QByteArray output = process.readAllStandardError();
    list << QString::fromLocal8Bit( output );

    warning((QByteArray)"実行エラー", (QByteArray)"プロセスでエラーが発生しました", list);
}

/** (slot)エラー発生時
 * @brief ProcessExecutor::slotProcError
 * @param error
 */
void ProcessExecutor::slotProcError(QProcess::ProcessError error) {
    // 詳細な情報を追加
    QStringList list;
    list << "program=" + progName;
    for (int i = 0; i < progArgs.size(); i++) {
        list << "arg[" + QString::number(i) +"]=" + progArgs.at(i);
    }
    list << "error=" +  QString::number(error);

    warning((QByteArray)"実行エラー", (QByteArray)"プロセスの起動に失敗しました", list);
}
