﻿/*------------------------------------------------------------------------------
* AntLibMain.ccp : アンテナ位相特性推定処理メインクラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "AntLibMain.h"

using namespace std;

#if 1   // debug
#define DEBUG_ANTLIBMAIN(msg, val) cout << msg << val << endl
#else
#define DEBUG_ANTLIBMAIN(msg, val)
#endif


/** コンストラクタ
 * @brief AntLibMain::AntLibMain
 */
AntLibMain::AntLibMain()
{
    maxGdop = 30.0;

    inputFileNum = 0;
    for (int i = 0; i < MAXINFILE; i++) {
        inputFiles[i] = NULL;
    }
}

/** デストラクタ
 * @brief AntLibMain::AntLibMain
 */
AntLibMain::~AntLibMain()
{
    freeInputFiles();
}

/** 初期化処理
 * @brief AntLibMain::initialize
 */
void AntLibMain::initialize() {
    // アンテナクラスの初期化
    antennaManager.initialize(&observationManagerForReference, &navigationManager);
    antennaEstimater.initialize(&observationManagerForTarget, &navigationManager);
    navErrEstimater.initialize(&observationManagerForTarget, &navigationManager);

    // GLONASS受信機チャンネル間バイアス推定クラス初期化
    glonassEstimater.initialize(&navigationManager);

    // 状態のクリア
    antennaManager.reset();
    antennaEstimater.reset();
    navErrEstimater.reset();
    vecMessage.clear();
}

/** 基準局位置設定(ECEF)
 * @brief AntLibMain::setReferencePosition
 * @param x 局位置のX成分[m]
 * @param y 局位置のY成分[m]
 * @param z 局位置のZ成分[m]
 */
void AntLibMain::setReferencePosition(double x, double y, double z) {
    antennaManager.setPosition(x, y, z);
}

/** 測定局位置設定(ECEF)
 * @brief AntLibMain::setTargetPosition
 * @param x 局位置のX成分[m]
 * @param y 局位置のY成分[m]
 * @param z 局位置のZ成分[m]
 */
void AntLibMain::setTargetPosition(double x, double y, double z) {
    antennaEstimater.setPosition(x, y, z);
}

/** アンテナ位相特性推定
 * @brief AntLibMain::estimatePCV
 * @param rinexFilesForReference 基準局の観測データ
 * @param rinexFilesForTarget 測定局の観測データ
 * @param rinexNavFilest 航法データ
 * @param pcvFile 基準局のアンテナ位相特性データ(NGSPCV or ANTEX)
 * @param outputPcvFile 基準局のアンテナ位相特性データ(出力先)
 * @param minForSN カットオフSNR/CN0
 * @param nMax 球面調和関数・オーダー
 * @param mMax 球面調和関数・角
 * @param azOffset アンテナ方位角オフセット
 * @param elMin カットオフ仰角
 * @param thresslip サイクルスリップ検出の閾値(L1-L2)
 * @param sphVar 球面調和関数推定の共分散(衛星システム別、L1/L2/L5)
 * @param agencyForAntex agency(ANTEXファイル用)
 * @param logLevel ログレベル
 * @param phaseCenterFlag [OUT]測定局のアンテナ位相中心位置フラグ(true：中心位置算出済み)(衛星システム別、L1/L2/L5)
 * @param phaseCenter [OUT]測定局のアンテナ位相中心位置(衛星システム別、L1/L2/L5)
 */
void AntLibMain::estimatePCV(vector<string> &rinexFilesForReference,
            vector<string> &rinexFilesForTarget,
            vector<string> &rinexNavFilest,
            string pcvFile, string outputPcvFile,
            const double minForSN, const int nMax, const int mMax,
                             const double azOffset, const double elMin, const double thresslip, double sphVar[2][3],
            const string agencyForPcv, const string agencyForAntex, int logLevel,
            bool phaseCenterFlag[2][3], double phaseCenter[2][3][3]) {
    // ログ出力
    outputLog(TRACE_LEVEL_DEBUG, "pcvFile=" + pcvFile, "AntLibMain::estimatePCV");
    outputLog(TRACE_LEVEL_DEBUG, "outputPcvFile=" + outputPcvFile, "AntLibMain::estimatePCV");

    // [[[ 推定準備 ]]]
    // 観測データ&航法データ設定(基準局)
    antennaManager.readRinex(&rinexFilesForReference, &rinexNavFilest);

    // 観測データ設定(測定局)
    antennaEstimater.readRinex(&rinexFilesForTarget, NULL);

    // アンテナ位相特性データ設定(基準局)
    observationManagerForReference.readPcv(pcvFile);
    int pcvType = NGS_PCV;
    if ((pcvFile.find(".atx") != string::npos)
            || (pcvFile.find(".ATX") != string::npos)) {
        pcvType = ANTEX;
    }

    // 観測データの最初の時刻を取得
    vector<gtime_t> vecTime = observationManagerForReference.getAllDataTime();
    gtime_t startTime = vecTime[0];

    // [[[ 推定パラメータの準備 ]]]
    // オプション設定
    /* ※次はアンテナ位相中心位置/アンテナ位相特性で違うので、各処理前で設定
     *  -mode
     *  -soltype
     *  -ru[]
     *  -antestmode
     *  -modear
     */
    prcopt_t opt = prcopt_default;
    opt.mode=3;
    opt.elmin = elMin * D2R;    // [rad]
//    opt.snrmin = minForSN;    // rtklib.h修正により一時的にコメントアウト
    opt.thresslip = thresslip;
    opt.maxgdop = maxGdop;
    opt.dcbratio = 1.0;

    // アンテナ位相特性用
    opt.antestnmax[0] = nMax;
    opt.antestnmax[1] = mMax;

    // (基準局位置)
    double *referencePosition = antennaManager.getPosition();
    opt.refpos = 0;
    for (int i = 0; i < 3; i++) {
        opt.rb[i] = referencePosition[i];
    }

    // 解オプション
    solopt_t solopt = solopt_default;
    solopt.posf = SOLF_XYZ;
    solopt.outopt = 1;
    solopt.trace = logLevel;
    strcpy(solopt.sep, ",");

    // ファイルオプション
    filopt_t fileopt = { 0 };
    strcpy(fileopt.satantp, pcvFile.c_str());
    strcpy(fileopt.rcvantp, pcvFile.c_str());

    // 推定クラスへの設定
    antennaEstimater.minForSN = minForSN;
    antennaEstimater.nMax[0] = nMax;
    antennaEstimater.nMax[1] = mMax;
    antennaEstimater.azOffset = azOffset * D2R;    // [rad]
    antennaEstimater.agencyForPcv = agencyForPcv;
    antennaEstimater.paramForAntex.agency = agencyForAntex;

    // 入力ファイル情報の準備
    createInputFiles(rinexFilesForReference, rinexFilesForTarget, rinexNavFilest);

    // GPS&QZSとGLONASSに分ける
    int arrayUseSatSys[2];
    int satSysNum = 2;
    arrayUseSatSys[0] = SYS_GPS | SYS_QZS;
    arrayUseSatSys[1] = SYS_GLO;

    // 衛星システムによるループ
    bool estimateAllCheck = false;
    for (int satSysIdx = 0; satSysIdx < satSysNum; satSysIdx++) {
        // 状態のリセット
        int freqNum = (arrayUseSatSys[satSysIdx] == SYS_GLO)? 2:NFREQ;
        opt.ionoopt=0;
        if(satSysIdx==0){
            opt.oprfrq[0]=0;
            opt.oprfrq[1]=1;
            opt.oprfrq[2]=2;
            opt.oprfrq[3]=-1;
        }
        else if(satSysIdx==1){
            opt.oprfrq[0]=0;
            opt.oprfrq[1]=1;
            opt.oprfrq[2]=-1;
        }
        bool estimatedPcv[NFREQ];
        antennaEstimater.reset();
        for (int i = 0; i < freqNum; i++) {
            estimatedPcv[i] = false;
        }


        // 衛星システム設定
        string satSysStr = getSatSysStr(arrayUseSatSys[satSysIdx]);
        DEBUG_ANTLIBMAIN("satsys=" , satSysStr);
        opt.navsys = arrayUseSatSys[satSysIdx];
        antennaEstimater.useSys = arrayUseSatSys[satSysIdx];

        // 基準局のアンテナ位相特性を設定
        observationManagerForReference.setPcv(startTime, arrayUseSatSys[satSysIdx], 1,
                                              navigationManager.getAll(), &opt);

        // [[[ 推定処理 ]]]
        // (周波数別処理)
        for (int i = 0; i < freqNum; i++) {
            string procStr = "sys=" + satSysStr + " freq=" + antLibMainFreqName[i];
#if 1   // debug
    obs_t *allObs = observationManagerForReference.getAll();
    if (allObs->n > 0) {
        unsigned char code = allObs->data[0].code[i];
        cout << "obscode=" << code2obs(code, &i) << " (code=" << (int)code << ")" << endl;
    }
#endif

            // 周波数指定
            antennaEstimater.useFreq = i;
            opt.antestfreq = i;
            opt.nfreq = 1;
            if (i == ANTLIB_FREQ_L2) {
                opt.nfreq = 2;
            } else if (i == ANTLIB_FREQ_L5) {
                opt.nfreq = 3;
            }

            // アンテナ位相特性の設定(周波数別)
            opt.antestvar = sphVar[satSysIdx][i];

            // 測定局のアンテナ位相中心を推定
            try {
                DEBUG_ANTLIBMAIN("estimatePhaseCenter() " ,procStr);
                outputLog(TRACE_LEVEL_INFO, "estimatePhaseCenter", "AntLibMain::estimatePCV");
                string outputFilepath = outputDir + getResultFilenameForPhaseCenter(arrayUseSatSys[satSysIdx], i);
                double* pc = antennaEstimater.estimatePhaseCenter(opt, &solopt, &fileopt,
                                                            inputFiles, inputFileNum, outputFilepath);
                // 結果を保管
                phaseCenterFlag[satSysIdx][i] = true;
                for (int j = 0; j < 3; j++) {
                    phaseCenter[satSysIdx][i][j] = pc[j];
                }
            } catch(AntLibException ex) {
                phaseCenterFlag[satSysIdx][i] = false;
                vecMessage.push_back("[warning] etimate(phase center) failed. " + procStr
                                            + "\n " + ex.getMessage());
                continue;
            }


            // 測定局のPCVを推定
            string outputFilepathForPos = outputDir + getResultFilenameForPcvPos(arrayUseSatSys[satSysIdx], i);
            string outputFilepathForSph = outputDir + getResultFilenameForPcvSph(arrayUseSatSys[satSysIdx], i);
            string outputFilepathForPhaseError = outputDir + getResultFilenameForPcvPhaseErr(arrayUseSatSys[satSysIdx], i);
            DEBUG_ANTLIBMAIN("estimatePCV() ", procStr);
            outputLog(TRACE_LEVEL_INFO, "estimatePCV() " + procStr, "AntLibMain::estimatePCV");
            estimatedPcv[i] = antennaEstimater.estimatePCV(opt, &solopt, &fileopt,
                                                    inputFiles, inputFileNum,
                                                    outputFilepathForPos, outputFilepathForSph,
                                                    outputFilepathForPhaseError);
            outputLog(TRACE_LEVEL_INFO, "estimatePCV() ret=", estimatedPcv[i], "AntLibMain::estimatePCV");

            // (推定できたらデータを生成)
            if (estimatedPcv[i]) {
                // アンテナ位相特性の推定結果を読込
                DEBUG_ANTLIBMAIN("readPcvEstFile() ", procStr);
                antennaEstimater.readPcvEstFile(outputFilepathForSph);

                // アンテナ位相特性ファイル用データ生成
                DEBUG_ANTLIBMAIN("createAntennaCalibrations() ", procStr);
                outputLog(TRACE_LEVEL_INFO, "createAntennaCalibrations() " + procStr, "AntLibMain::estimatePCV");
                bool ret = antennaEstimater.createAntennaCalibrations(pcvType);

                if (ret) {
                    // 推定結果を出力(プロット用)
                    DEBUG_ANTLIBMAIN("outputPcvResult() ", procStr);
                    outputLog(TRACE_LEVEL_INFO, "outputPcvResult() " + procStr, "AntLibMain::estimatePCV");
                    string outputFilepath = outputDir + getResultFilenameForPcv(arrayUseSatSys[satSysIdx], i);
                    antennaEstimater.outputPcvResult(outputFilepath);
                }
            }
        }


        // 推定状況のチェック
        bool estimateFlag = false;
        for (int i = 0; i < freqNum; i++) {
            string procStr = "sys=" + satSysStr + " freq=" + antLibMainFreqName[i];
            if (estimatedPcv[i]) {
                estimateFlag = true;
                vecMessage.push_back("etimated pcv. " + procStr);
            }
        }

        if (estimateFlag) {
            // アンテナ位相特性ファイルに追記
            DEBUG_ANTLIBMAIN("outputAntennaCalibrations", "");
            outputLog(TRACE_LEVEL_INFO, "outputAntennaCalibrations", "AntLibMain::estimatePCV");
            antennaEstimater.outputAntennaCalibrationsForSys(pcvType, freqNum, outputPcvFile);
        }
        estimateAllCheck |= estimateFlag;
    }

    if (estimateAllCheck && (pcvType == ANTEX)) {
        antennaEstimater.outputAntennaCalibrationsForAntex(outputPcvFile);
    }

    // 後処理
    freeInputFiles();

#if 0   // temp
    if (!estimateAllCheck) {
        throw AntLibException("estimate failed, no valid data", "AntLibMain::estimatePCV");
    }
#endif
}

/** GLONASS受信機チャンネル間バイアス推定
 * @brief AntLibMain::estimateGlonass
 * @param rinexFilesForReference 基準局の観測データ
 * @param rinexFilesForTarget 測定局の観測データ
 * @param rinexNavFilest 航法データ
 * @param outputEstBiasFile
 * @param sigp 搬送波位相観測誤差の標準偏差
 * @param sigx アプリオリ共分散行列:σx
 * @param thresslip サイクルスリップ検出の閾値(L1-L2)
 * @param phaseCenterForRef 基準局のアンテナ位相中心位置(L1/L2)
 * @param phaseCenter 測定局のアンテナ位相中心位置(L1/L2)
 * @param sigNum IFB棄却処理Sigma
 * @param repeat IFB棄却処理繰り返し回数
 */
void AntLibMain::estimateGlonass(vector<string> &rinexFilesForReference,
            vector<string> &rinexFilesForTarget,
            vector<string> &rinexNavFilest,
            string outputEstBiasFile,
            const double sigp, const double sigx, const double thresslip,
            int calcInterval, int pathMinRange,
            const double sigNum, const int repeat) {
    // ログ出力
    outputRinexFile(rinexFilesForReference, rinexFilesForTarget, rinexNavFilest, "AntLibMain::estimateGlonass");
    outputLog(TRACE_LEVEL_DEBUG, "outputEstBiasFile=" + outputEstBiasFile, "AntLibMain::estimateGlonass");

    // [[[ 推定準備 ]]]
    // オプション設定
    prcopt_t opt = prcopt_default;
    opt.thresslip = thresslip;

    // 観測データ&航法データ設定(基準局)
    antennaManager.readRinex(&rinexFilesForReference, &rinexNavFilest);
    if (navigationManager.getGloNavNum() <= 0) {
        // GLONASSの航法データがない場合
        throw AntLibException("estimate failed, no glonass navigation data", "AntLibMain::estimateGlonass");
    }

    // 観測データ設定(測定局)
    antennaEstimater.readRinex(&rinexFilesForTarget, NULL);

    // 推定前の初期化
    glonassEstimater.reset();

    // [[[ 推定処理 ]]]
    int freqNum = 2;
    for (int i = 0; i < freqNum; i++) {
        // GLONASS受信機チャンネル間バイアス推定
        DEBUG_ANTLIBMAIN("estimateBias() freq=", antLibMainFreqName[i]);
        outputLog(TRACE_LEVEL_INFO, "estimateBias() freq=" + antLibMainFreqName[i], "AntLibMain::estimateGlonass");
        string outputFilepathForBias = outputDir + getResultFilenameForGloIFBBias(i);
        string outputFilepath = outputDir + getResultFilenameForGloIFB(i);
        glonassEstimater.estimateBias(i, opt, &observationManagerForReference, antennaManager.getPosition(),
                                      &observationManagerForTarget, antennaEstimater.getPosition(),
                                      sigx, sigp, calcInterval, pathMinRange, outputFilepathForBias, outputFilepath,
                                      sigNum, repeat, vecMessage);
    }

    // 結果出力
    DEBUG_ANTLIBMAIN("outputErrorModel", "");
    outputLog(TRACE_LEVEL_INFO, "outputErrorModel", "AntLibMain::estimateGlonass");
    glonassEstimater.outputResult(outputEstBiasFile);
}

/**
 * @brief AntLibMain::estimateMultipass
 * @param rinexFilesForReference 基準局の観測データ＆航法データ
 * @param rinexFilesForTarget 測定局の観測データ
 * @param rinexNavFilest 航法データ
 * @param navErrFile 観測誤差モデル
 * @param outputNavErrFile 観測誤差モデル(出力先)
 * @param rangeForAMP 単純移動平均の対象データ数
 * @param deltaBeta 観測更新の単位とする仰角範囲
 * @param sigmaA アプリオリ共分散行列:σａ
 * @param sigmaB アプリオリ共分散行列:σb
 * @param thresslip サイクルスリップ検出の閾値(L1-L2)
 * @param minEl 推定対象衛星仰角下限値
 * @param maxEl 推定対象衛星仰角上限値
 * @param phaseCenterFlag 測定局のアンテナ位相中心位置フラグ(true：中心位置指定あり)(衛星システム別、L1/L2/L5)
 * @param phaseCenter 測定局のアンテナ位相中心位置(衛星システム別、L1/L2/L5)
 */
void AntLibMain::estimateMultipass(vector<string> &rinexFilesForReference,
                                   vector<string> &rinexFilesForTarget,
                                   vector<string> &rinexNavFilest,
                                   string navErrFile, string outputNavErrFile,
                                   const int rangeForAMP,
                                   const double deltaBeta, const double sigmaA, const double sigmaB,
                                   const double thresslip, const double minEl, const double maxEl,
                                   bool phaseCenterFlag[2][3], double phaseCenter[2][3][3]) {
    // ログ出力
    outputLog(TRACE_LEVEL_DEBUG, "navErrFile=" + navErrFile, "AntLibMain::estimateMultipass");
    outputLog(TRACE_LEVEL_DEBUG, "outputNavErrFile=" + outputNavErrFile, "AntLibMain::estimateMultipass");

    // [[[ 推定準備 ]]]
    // 観測データ&航法データ設定(基準局)
    antennaManager.readRinex(&rinexFilesForReference, &rinexNavFilest);

    // 観測データ設定(測定局)
    navErrEstimater.readRinex(&rinexFilesForTarget, NULL);

    // 共通するコードを取得
    vector<string> vecPseudorangeCode[NFREQ];
    vector<string> vecPhaseCode[NFREQ];
    for (int freq = 0; freq < NFREQ; freq++) {
        // (疑似距離)
        cout << "### freq=" << antLibMainFreqName[freq] << endl;    // debug
        vector<string> vecRef = antennaManager.getPseudorangeCode(freq);
        vector<string> vecTag = antennaEstimater.getPseudorangeCode(freq);
        for (int i = 0; i < (int)vecRef.size(); i++) {
            for (int j = 0; j < (int)vecTag.size(); j++) {
                if (vecRef[i] == vecTag[j]) {
                    // 共通のものを取得
                    vecPseudorangeCode[freq].push_back(vecRef[i]);
                    cout << "###  PseudorangeCode-" << i << ":" << vecRef[i] << endl;   // debug
                }
            }
        }

        // (搬送波位相)
        vecRef = antennaManager.getPhaseCode(freq);
        vecTag = antennaEstimater.getPhaseCode(freq);
        for (int i = 0; i < (int)vecRef.size(); i++) {
            for (int j = 0; j < (int)vecTag.size(); j++) {
                if (vecRef[i] == vecTag[j]) {
                    // 共通のものを取得
                    vecPhaseCode[freq].push_back(vecRef[i]);
                    cout << "###  PhaseCode-" << i << ":" << vecRef[i] << endl;   // debug
                }
            }
        }
    }

    // 擬似距離コードが同じ周波数に複数あるかチェック
    bool rereadFlag = false;
    for (int freq = 0; freq < NFREQ; freq++) {
        rereadFlag |= (vecPseudorangeCode[freq].size() > 1);
    }


    // [[[ 推定パラメータの準備 ]]]
    // オプション設定
    prcopt_t opt = prcopt_default;
    opt.thresslip = thresslip;
    opt.maxgdop = maxGdop;

    // ALLの場合は、GPS/GLONASSに分ける
    int satSysNum = 2;
    int arrayUseSatSys[2];
    arrayUseSatSys[0] = SYS_GPS;
    arrayUseSatSys[1] = SYS_GLO;

    // 衛星システムによるループ
    bool estimateAllCheck = false;
    for (int satSysIdx = 0; satSysIdx < satSysNum; satSysIdx++) {
        navErrEstimater.clearResult();

        // 衛星システム設定
        string satSysStr = getSatSysStr(arrayUseSatSys[satSysIdx]);
        DEBUG_ANTLIBMAIN("satsys=" , satSysStr);
        opt.navsys = arrayUseSatSys[satSysIdx];
        navErrEstimater.useSys = arrayUseSatSys[satSysIdx];

        // [[[ 推定処理 ]]]
        // (全信号数分処理)
        int freqNum = (arrayUseSatSys[satSysIdx] == SYS_GLO)? 2:NFREQ;
        for (int freq = 0; freq < freqNum; freq++) {
            string procStr = "sys=" + satSysStr + " freq=" + antLibMainFreqName[freq];

            // 周波数指定
            navErrEstimater.useFreq = freq;

            // 測定局のアンテナ位相中心位置(測定局)
            int pcIdx = (arrayUseSatSys[satSysIdx] == SYS_GLO)? 1:0;
            if (!phaseCenterFlag[pcIdx][freq]) {
                vecMessage.push_back("[warning] not input phase center. " + procStr);
                continue;
            }
            navErrEstimater.reset();
            navErrEstimater.setPosition(phaseCenter[pcIdx][freq][0], phaseCenter[pcIdx][freq][1], phaseCenter[pcIdx][freq][2]);

            // 搬送波位相マルチパス＋受信機雑音データ読込
            DEBUG_ANTLIBMAIN("readPhaseErrorFile() ", procStr);
            outputLog(TRACE_LEVEL_INFO, "readPhaseErrorFile() " + procStr, "AntLibMain::estimateMultipass");
            string inputFilepath = outputDir + getResultFilenameForPcvPhaseErr(arrayUseSatSys[satSysIdx], freq);
            navErrEstimater.readPhaseErrorFile(inputFilepath);

            // 周波数別に観測データにある全信号数を取得
            for (int i = 0; i < (int)vecPseudorangeCode[freq].size(); i++) {
                bool needNotPmp = (i != 0);
                string cCode = vecPseudorangeCode[freq].at(i);
                string pCode = (vecPhaseCode[freq].size() > 0)? vecPhaseCode[freq].at(0): "";

                // 擬似距離コード別に読み込む必要がある場合
                if (rereadFlag) {
                    const char* codeStr = cCode.c_str();

                    // 観測データ(基準局)
                    antennaManager.readRinex(&rinexFilesForReference, NULL, codeStr[2]);

                    // 観測データ設定(測定局)
                    navErrEstimater.readRinex(&rinexFilesForTarget, NULL, codeStr[2]);
                }

                // マルチパスを推定
                DEBUG_ANTLIBMAIN("estimateMultipass() ", procStr);
                outputLog(TRACE_LEVEL_INFO, "estimateMultipass() " + procStr, "AntLibMain::estimateMultipass");
                navErrEstimater.estimateMultipass(&observationManagerForReference, opt, rangeForAMP,
                                                    minEl, maxEl, needNotPmp,
                                                    outputDir + "data_" + antLibMainFreqName[freq] + "/");

                // 観測誤差モデル推定&プロット用データ出力
                DEBUG_ANTLIBMAIN("estimateErrorModel() ", procStr);
                outputLog(TRACE_LEVEL_INFO, "estimateErrorModel() " + procStr, "AntLibMain::estimateMultipass");
                string outputCmpFilepath = outputDir + getResultFilenameForCmp(arrayUseSatSys[satSysIdx], cCode);
                string outputPmpFilepath = outputDir + getResultFilenameForPmp(arrayUseSatSys[satSysIdx], pCode);
                navErrEstimater.estimateErrorModel(deltaBeta, sigmaA, sigmaB, cCode, pCode,
                                                   outputCmpFilepath, outputPmpFilepath);
            }
        }

        // 推定状況のチェック
        // (電離層遅延量の算出状態チェック)
        vector<estNavErrorData_t> vecNaveError = navErrEstimater.getVecEstNavError();
        if (!navErrEstimater.getCalculateIono() && vecNaveError.size() > 0) {
            string procStr = "sys=" + satSysStr;
            vecMessage.push_back("[warning] can't calculate Ionospheric Delay Variations. " + procStr);
        }
        for (int i = 0; i < (int)vecNaveError.size(); i++) {
            string procStr = "sys=" + satSysStr + " code=" + vecNaveError[i].code;
            vecMessage.push_back("etimated nav_error_model. " + procStr);
        }
        estimateAllCheck |= (vecNaveError.size() > 0);

        // 観測誤差モデルの出力
        DEBUG_ANTLIBMAIN("outputErrorModel sys=", satSysStr);
        outputLog(TRACE_LEVEL_INFO, "outputErrorModel", "AntLibMain::estimateMultipass");
        navErrEstimater.outputErrorModel(outputNavErrFile, (satSysIdx == 0));
    }
    if (!estimateAllCheck) {
        throw AntLibException("estimate failed, no valid data", "AntLibMain::estimateMultipass");
    }

}

/** ログ出力
 * @brief AntLibMain::outputRinexFile
 * @param rinexFilesForReference 基準局の観測データ
 * @param rinexFilesForTarget 測定局の観測データ
 * @param rinexNavFilest 航法データ
 * @param pos 発生個所
 */
void AntLibMain::outputRinexFile(vector<string> &rinexFilesForReference,
               vector<string> &rinexFilesForTarget,
               vector<string> &rinexNavFilest, string pos) {
    vector<string>* vec[3] = {&rinexFilesForReference, &rinexFilesForTarget, &rinexNavFilest};
    string vecName[3] = {"refObs[", "targetObs[", "navigation["};
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < (int)vec[i]->size(); j++) {
            ostringstream oStr;
            oStr << vecName[i] << j << "]=" << vec[i]->at(j);
            outputLog(TRACE_LEVEL_DEBUG, oStr.str(), pos);
        }
    }
}

/** 結果ファイル名取得(位相中心の結果用)
 * @brief AntLibMain::getResultFilenameForPhaseCenter
 * @param sys 衛星システム
 * @param freq 周波数
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForPhaseCenter(int sys, int freq) {
    return PC_RESULT_PREFIX + getSatSysStr(sys) + "_" + antLibMainFreqName[freq] + RESULT_SUFFIX;
}

/** 結果ファイル名取得(PCVの結果用)
 * @brief AntLibMain::getResultFilenameForPcv
 * @param sys 衛星システム
 * @param freq 周波数
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForPcv(int sys, int freq) {
    return PCV_RESULT_PREFIX + getSatSysStr(sys) + "_" + antLibMainFreqName[freq] + RESULT_SUFFIX;
}

/** 結果ファイル名取得(相対測位結果用)
 * @brief AntLibMain::getResultFilenameForPcvPos
 * @param sys 衛星システム
 * @param freq 周波数
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForPcvPos(int sys, int freq) {
    return PCV_POS_RESULT_PREFIX + getSatSysStr(sys) + "_" + antLibMainFreqName[freq] + RESULT_SUFFIX;
}

/** 結果ファイル名取得(球面調和関数係数結果用)
 * @brief AntLibMain::getResultFilenameForPcvSph
 * @param sys 衛星システム
 * @param freq 周波数
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForPcvSph(int sys, int freq) {
    return PCV_SPH_RESULT_PREFIX + getSatSysStr(sys) + "_" + antLibMainFreqName[freq] + RESULT_SUFFIX;
}

/** 結果ファイル名取得(搬送波位相一重差の残差)
 * @brief AntLibMain::getResultFilenameForPcvPhaseErr
 * @param sys 衛星システム
 * @param freq 周波数
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForPcvPhaseErr(int sys, int freq) {
    return PCV_PHASEERR_RESULT_PREFIX + getSatSysStr(sys) + "_" + antLibMainFreqName[freq] + RESULT_SUFFIX;
}

/** 結果ファイル名取得(GLONASS周波数間バイアスの結果用、衛星バイアス推定)
 * @brief AntLibMain::getResultFilenameForGloIFBBias
 * @param freq 周波数
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForGloIFBBias(int freq) {
    return GLOIFB_BIAS_PREFIX + antLibMainFreqName[freq] + RESULT_SUFFIX;
}

/** 結果ファイル名取得(GLONASS周波数間バイアスの結果用)
 * @brief AntLibMain::getResultFilenameForGloIFB
 * @param freq 周波数
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForGloIFB(int freq) {
    return GLOIFB_RESULT_PREFIX + antLibMainFreqName[freq] + RESULT_SUFFIX;
}

/** 結果ファイル名取得(擬似距離マルチパスの結果用)
 * @brief AntLibMain::getResultFilenameForCmp
 * @param sys 衛星システム
 * @param code 観測コード
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForCmp(int sys, string code) {
    return CMP_RESULT_PREFIX + getSatSysStr(sys) + "_" + code + RESULT_SUFFIX;
}

/** 結果ファイル名取得(搬送波位相マルチパスの結果用)
 * @brief AntLibMain::getResultFilenameForPmp
 * @param sys 衛星システム
 * @param code 観測コード
 * @return ファイル名
 */
string AntLibMain::getResultFilenameForPmp(int sys, string code) {
    return PMP_RESULT_PREFIX + getSatSysStr(sys) + "_" + code + RESULT_SUFFIX;
}

/** 衛星システム名取得
 * @brief AntLibMain::getSatSysStr
 * @param sys 衛星システム
 * @return 衛星システム名
 */
string AntLibMain::getSatSysStr(int sys) {
    string str = "GPS";
    if (sys == SYS_GLO) {
        str = "GLO";
    } else if (sys == SYS_QZS) {
        str = "QZS";
    }
    return str;
}

/** GSILIB用の入力ファイル情報を生成
 * @brief AntLibMain::estimatePCV
 * @param rinexFilesForReference 基準局の観測データ
 * @param rinexFilesForTarget 測定局の観測データ
 * @param rinexNavFilest 航法データ
 */
void AntLibMain::createInputFiles(vector<string> &rinexFilesForReference,
            vector<string> &rinexFilesForTarget,
            vector<string> &rinexNavFilest) {
    // 入力ファイルの準備
    // ※一つ目を測定局の観測データとする(postpostのルール)
    vector<string> rinexFiles;
    for (int i = 0; i < (int)rinexFilesForTarget.size(); i++) {
        rinexFiles.push_back(rinexFilesForTarget[i]);
    }
    for (int i = 0; i < (int)rinexFilesForReference.size(); i++) {
        rinexFiles.push_back(rinexFilesForReference[i]);
    }
    for (int i = 0; i < (int)rinexNavFilest.size(); i++) {
        rinexFiles.push_back(rinexNavFilest[i]);
    }

    // 入力ファイルの数をチェック
    inputFileNum = rinexFiles.size();
    if (inputFileNum > MAXINFILE) {
        inputFileNum = 0;
        throw AntLibException("inputfile is too many.", "AntennaEstimater::estimatePhaseCenter");
    }

    // GSILIB用に置き換え
    for (int i = 0; i < inputFileNum; i++) {
        if (!(inputFiles[i] = (char *)malloc(1024))) {
            for (;i>=0;i--) free(inputFiles[i]);
            throw AntLibException("Reservation of a memory failed for inputfile.", "AntennaEstimater::estimatePhaseCenter");
        }
        strcpy(inputFiles[i], rinexFiles[i].c_str());
    }
}

/**  GSILIB用の入力ファイル情報の解放
 * @brief AntLibMain::freeInputFiles
 */
void AntLibMain::freeInputFiles() {
    for (int i = 0; i < inputFileNum; i++) {
        if (inputFiles[i] == NULL) {
            continue;
        }

        free(inputFiles[i]);
        inputFiles[i] = NULL;
    }
    inputFileNum = 0;
}
