﻿/*------------------------------------------------------------------------------
* AntLibMain.h : アンテナ位相特性推定処理メインクラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#ifndef ANTLIBMAIN_H
#define ANTLIBMAIN_H
#include <string>
#include <vector>
#include "AntennaManager.h"
#include "AntennaEstimater.h"
#include "GlonassEstimater.h"
#include "NavErrEstimater.h"

// 結果ファイル用周波数名
static std::string antLibMainFreqName[] = {"L1", "L2", "L5"};
#define MAXINFILE   1000        // 入力ファイルの最大数

// 結果ファイル名
#define PC_RESULT_PREFIX "pc_est_"
#define PCV_RESULT_PREFIX "pcv_res_"
#define PCV_POS_RESULT_PREFIX "pcv_pos_"
#define PCV_SPH_RESULT_PREFIX "pcv_sph_"
#define PCV_PHASEERR_RESULT_PREFIX "pcv_phe_"
#define GLOIFB_BIAS_PREFIX "gloifb_bias_"
#define GLOIFB_RESULT_PREFIX "gloifb_"
#define CMP_RESULT_PREFIX "cmp_res_"
#define PMP_RESULT_PREFIX "pmp_res_"
#define RESULT_SUFFIX ".dat"


/** アンテナ位相特性推定処理メインクラス
 * @brief The AntLibMain class
 */
class AntLibMain : public AntLibBase
{
public:
    AntLibMain();
    virtual ~AntLibMain();

    void initialize();
    void setReferencePosition(double x, double y, double z);
    void setTargetPosition(double x, double y, double z);

    void estimatePCV(std::vector<std::string> &rinexFilesForReference,
                std::vector<std::string> &rinexFilesForTarget,
                std::vector<std::string> &rinexNavFilest,
                std::string pcvFile, std::string outputPcvFile,
                const double minForSN, const int nMax, const int mMax,
                const double azOffset, const double elMin, const double thresslip, double sphVar[2][3],
                const std::string agencyForPcv, const std::string agencyForAntex, int logLevel,
                bool phaseCenterFlag[2][3], double phaseCenter[2][3][3]);
    void estimateGlonass(std::vector<std::string> &rinexFilesForReference,
                std::vector<std::string> &rinexFilesForTarget,
                std::vector<std::string> &rinexNavFilest,
                std::string outputEstBiasFile,
                const double sigp, const double sigx, const double thresslip,
                int calcInterval, int pathMinRange,
                const double sigNum, const int repeat);
    void estimateMultipass(std::vector<std::string> &rinexFilesForReference,
                std::vector<std::string> &rinexFilesForTarget,
                std::vector<std::string> &rinexNavFilest,
                std::string navErrFile, std::string outputNavErrFile,
                const int rangeForAMP,
                const double deltaBeta, const double sigmaA, const double sigmaB,
                const double thresslip, const double minEl, const double maxEl,
                bool phaseCenterFlag[2][3], double phaseCenter[2][3][3]);

    // log
    void outputRinexFile(std::vector<std::string> &rinexFilesForReference,
                   std::vector<std::string> &rinexFilesForTarget,
                   std::vector<std::string> &rinexNavFilest, std::string pos);

    // static
    static std::string getResultFilenameForPhaseCenter(int sys, int freq);
    static std::string getResultFilenameForPcv(int sys, int freq);
    static std::string getResultFilenameForPcvPos(int sys, int freq);
    static std::string getResultFilenameForPcvSph(int sys, int freq);
    static std::string getResultFilenameForPcvPhaseErr(int sys, int freq);
    static std::string getResultFilenameForGloIFBBias(int freq);
    static std::string getResultFilenameForGloIFB(int freq);
    static std::string getResultFilenameForCmp(int sys, std::string code);
    static std::string getResultFilenameForPmp(int sys, std::string code);
    static std::string getSatSysStr(int sys);

public:
    /** 出力ディレクトリ("/"で終わること)
     * @brief outputDir
     */
    std::string outputDir;

    /** 測位計算時のGDOP閾値
     * @brief maxGdop
     */
    double maxGdop;

    /** 情報メッセージ
     * @brief vecMessage
     */
    std::vector<std::string> vecMessage;

private:
    void createInputFiles(std::vector<std::string> &rinexFilesForReference,
                          std::vector<std::string> &rinexFilesForTarget,
                          std::vector<std::string> &rinexNavFilest);
    void freeInputFiles();

    /** 観測データクラス(基準局用)
     * @brief observationManager
     */
    ObservationManager observationManagerForReference;

    /** 観測データクラス(測定局用)
     * @brief observationManager
     */
    ObservationManager observationManagerForTarget;

    /** 航法MSGクラス
     * @brief navigationManager
     */
    NavigationManager navigationManager;

    /** 基準局のアンテナクラス
     * @brief antennaManager
     */
    AntennaManager antennaManager;

    /** 測定局のアンテナクラス
     * @brief antennaEstimater
     */
    AntennaEstimater antennaEstimater;

    /** GLONASS受信機チャンネル間バイアス推定クラス
     * @brief glonassEstimater
     */
    GlonassEstimater glonassEstimater;

    /** 観測誤差モデル推定のアンテナクラス
     * @brief navErrEstimater
     */
    NavErrEstimater navErrEstimater;

    /** 入力ファイルの数（GSILIB用）
     * @brief inputFileNum
     */
    int inputFileNum;

    /** 入力ファイル（GSILIB用）
     * @brief inputFiles
     */
    char *inputFiles[MAXINFILE];
};

#endif // ANTLIBMAIN_H
