﻿/*------------------------------------------------------------------------------
* AntLibTool.h : ANTLib共通ライブラリクラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#ifndef ANTLIBTOOL_H
#define ANTLIBTOOL_H

#include <vector>
#include <string>
#include "rtklib.h"

// TRACEレベル
#define TRACE_LEVEL_ERROR           1
#define TRACE_LEVEL_WARNING         2
#define TRACE_LEVEL_INFO            3
#define TRACE_LEVEL_DEBUG           4
#define TRACE_LEVEL_DEBUG_DETAIL    5

/** 文字列分割
 * @brief splitString
 * @param str 文字列
 * @param delim 区切り文字
 * @return 分割文字列
 */
inline std::vector<std::string> splitString(const std::string &str, char delim) {
    std::vector<std::string> ret;
    size_t pos = 0;
    size_t targetPos;
    while((targetPos = str.find_first_of(delim, pos))  != std::string::npos) {
        ret.push_back(std::string(str, pos, targetPos - pos));
        pos = targetPos + 1;
    }
    ret.push_back(std::string(str, pos, str.size() - pos));
    return ret;
}

/** 文字列内の文字置換
 * @brief replaceString
 * @param str 文字列
 * @param targetChar 検索文字
 * @param replaceStr 置換文字
 * @return 置換後の文字列
 */
inline std::string replaceString(const std::string &str, char targetChar, std::string replaceStr) {
    std::string ret = str;
    size_t pos = 0;
    size_t targetPos;
    while((targetPos = str.find_first_of(targetChar, pos))  != std::string::npos) {
        ret = ret.replace(targetPos, 1, replaceStr);
        pos = targetPos + 1;
    }
    return ret;
}

/** 日付文字列を時刻に変換
 * @brief stringToGtime
 * @param dateStr 日付文字列(YYYY/MM/DD hh:mm:ss)
 * @return 時刻
 */
inline gtime_t stringToGtime(std::string dateStr) {
    // YYYY/MM/DD hh:mm:ss
    gtime_t datatime;
    dateStr = replaceString(dateStr, '/', " ");
    dateStr = replaceString(dateStr, ':', " ");
    str2time(dateStr.c_str(), 0, 19, &datatime);

    return datatime;
}

/** 現在時刻の取得
 * @brief getNowTime
 * @return 現在時刻
 */
inline gtime_t getNowTime() {
    // 現在時刻取得
    gtime_t gtime = timeget();  // UTC
    return gtime;
}

/** １重差行列作成
 * @brief matrix1
 * @param num データ数
 * @return １重差行列
 */
inline double* matrix1(int num, int baseIdx = 0)
{
    double* D = zeros(num-1, num);

    for (int i=0;i<num;i++)
    {
        double *colD = &D[i*(num-1)];

        if (i == baseIdx) {
            for (int j = 0; j < num-1; j++) {
                colD[j] = 1;
            }
        } else if (i > baseIdx) {
            colD[i-1] = -1;
        } else {
            colD[i] = -1;
        }
    }

    return D;
}

/** 転置行列作成
 * @brief matrixTrans
 * @param mat 行列(n x m)
 * @param outMat [OUT] 転置行列[m x n]
 * @param n
 * @param m
 */
inline void matrixTrans(double *mat, double *outMat, int n, int m) {
    for (int i = 0; i < n; i++) {
        for (int j = 0; j < m; j++) {
            outMat[j + i * m] = mat[i + j * n];
        }
    }
}

#endif // ANTLIBTOOL_H
