﻿/*------------------------------------------------------------------------------
* AntennaBase.ccp : アンテナ基底クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "AntennaBase.h"

using namespace std;

#define NUMSYS      6                   /* number of systems */

extern "C" {
// in rinex4pcv.c
extern int get_tobs(int sys, int targetFreq, int targetType, char ret[][4]);
extern void set_targetcode(char code);
}


/** コンストラクタ
 * @brief AntennaBase::AntennaBase
 */
AntennaBase::AntennaBase()
{
    // 局位置初期化
    positionFlag = false;
    for (int i = 0; i < 3; i++) {
        position[i] = 0.0;
    }
}

/** デストラクタ
 * @brief AntennaBase::~AntennaBase
 */
AntennaBase::~AntennaBase()
{
}

/** 初期処理
 * @brief AntennaBase::initialize
 * @param obsMng 観測データ管理クラスへのポインタ
 * @param navMng 航法データ管理クラスへのポインタ
 */
void AntennaBase::initialize(ObservationManager *obsMng, NavigationManager *navMng) {
    observationManager = obsMng;
    navigationManager = navMng;
}

/** クリア処理(推定準備前の初期化)
 * @brief AntennaBase::reset
 */
void AntennaBase::reset()
{
}

/** 局位置設定(ECEF)
 * @brief AntennaBase::setPosition
 * @param x 局位置のX成分[m]
 * @param y 局位置のY成分[m]
 * @param z 局位置のZ成分[m]
 */
void AntennaBase::setPosition(double x, double y, double z) {
    positionFlag = true;
    position[0] = x;
    position[1] = y;
    position[2] = z;
}

/** RINEXの読込
 * @brief AntennaBase::readRinex
 * @param vecFilepath RINEXファイルパス(複数、観測データ)
 * @param vecNavFilepath RINEXファイルパス(複数、航法データ)
 * @param targetCode 指定疑似距離コード('\0'：指定なし)
 */
void AntennaBase::readRinex(vector<string> *vecFilepath, vector<string> *vecNavFilepath, char targetCode) {
    obs_t obs = { 0 };
    nav_t nav = { 0 };
    sta_t sta = { 0 };

    // 疑似距離コードのクリア
    for (int freq = 0; freq < NFREQ; freq++) {
        vecPseudorangeCode[freq].clear();
    }

    // 疑似距離コード設定
    set_targetcode(targetCode);

    // RINEXファイル読込(観測データ)
    for (int i = 0; i < (int)vecFilepath->size(); i++) {
        readrnx(vecFilepath->at(i).c_str(), receiverNo, "", &obs, &nav, &sta);

        // 観測データに含まれる信号コードを取得
        char obsTypes[MAXOBSTYPE][4];
        for (int freq = 0; freq < NFREQ; freq++) {
            for (int sys = 0; sys < NUMSYS; sys++) {
                // (擬似距離)
                int num = get_tobs(sys, freq, 0, obsTypes);
                for (int j = 0; j < num; j++) {
                    string type = obsTypes[j];

                    // 既にあるタイプかチェック
                    vector<string>::iterator ite = find(vecPseudorangeCode[freq].begin(), vecPseudorangeCode[freq].end(), type);
                    if (ite != vecPseudorangeCode[freq].end()) {
                        // 存在するなら追加しない
                        continue;
                    }
                    vecPseudorangeCode[freq].push_back(type);
                }

                // (搬送波位相)
                num = get_tobs(sys, freq, 1, obsTypes);
                for (int j = 0; j < num; j++) {
                    string type = obsTypes[j];

                    // 既にあるタイプかチェック
                    vector<string>::iterator ite = find(vecPhaseCode[freq].begin(), vecPhaseCode[freq].end(), type);
                    if (ite != vecPhaseCode[freq].end()) {
                        // 存在するなら追加しない
                        continue;
                    }
                    vecPhaseCode[freq].push_back(type);
                }
            }
        }
    }

    // 観測データを設定
    observationManager->setData(&obs, &sta);

    // データ数チェック
    if (observationManager->getTotalNum() <= 0) {
        ostringstream strStream;
        strStream << "[getTotalNum=" << observationManager->getTotalNum() << "]\n";
        strStream << "[receiverNo=" << receiverNo << "]";
        throw AntLibException("no observation data" + strStream.str(), "AntennaBase::readRinex");
    }

    // 航法データを設定
    if (vecNavFilepath != NULL) {
        // RINEXファイル読込(航法データ)
        for (int i = 0; i < (int)vecNavFilepath->size(); i++) {
            readrnx(vecNavFilepath->at(i).c_str(), receiverNo, "", &obs, &nav, &sta);
        }

        navigationManager->setData(&nav);
        if (navigationManager->getTotalNum() <= 0) {
            ostringstream oStr;
            oStr << "[receiverNo=" << receiverNo << "]";
            throw AntLibException("no navigation message" + oStr.str(), "AntennaBase::readRinex");
        }
    }
}

/** 観測データにある全疑似距離コードを返す
 * @brief AntennaBase::getPseudorangeCode
 * @param freq 周波数
 * @return 全疑似距離コード
 */
vector<string> AntennaBase::getPseudorangeCode(int freq) {
    return vecPseudorangeCode[freq];
}

/** 観測データにある全搬送波位相コードを返す
 * @brief AntennaBase::getPhaseCode
 * @param freq 周波数
 * @return 全搬送波位相コード
 */
vector<string> AntennaBase::getPhaseCode(int freq) {
    return vecPhaseCode[freq];
}

/** 局位置計算
 * @brief AntennaBase::calculatePosition
 * @param opt
 * @return 局位置
 */
double* AntennaBase::calculatePosition(prcopt_t *opt) {
    if (positionFlag) {
        // 局位置が設定されている場合は何もしない
        return position;
    }

    // オプションを設定
    prcopt_t defOpt = prcopt_default;
    if (opt == NULL) {
        opt = &defOpt;
    }

    // 局位置計算結果保管用
    double result[3] = {0.0, 0.0, 0.0};

    // 300秒毎に局位置を算出
    gtime_t targetTime;
    vector<obsd_t> vecObsd;
    obsd_t obs_data[MAXOBS];
    int maxNum = observationManager->getAllDataTime().size();
    int calcNum = 0;
    for (int n = 0; n < maxNum; n++) {
        // 観測データを取得
        if (n == 0) {
            // 初回の計算時刻はデータの開始時刻とする
            vecObsd = observationManager->getData(NULL, &targetTime, false);
        } else {
            // 次の計算時刻が観測データの期間外かチェック
            targetTime.time += 300;  // 300秒加算
            if (!observationManager->checkPeriod(targetTime)) {
                break;
            }

            vecObsd = observationManager->getData(&targetTime, NULL, true);
        }

#if 0   // (debug) 計算時刻出力
        cout << "AntennaBase::calculatePosition " << time_str(targetTime, 2)
                << " vecObsd.size=" << vecObsd.size() << endl;
#endif

        // データ数チェック
        int dataNum = vecObsd.size();
        if (dataNum == 0) {
            // データなしの場合は何もしない
            continue;
        }
        if (dataNum > MAXOBS) {
            string msg = "too many obsData. time=";
            msg.append(time_str(targetTime, 2));
            throw AntLibException(msg, "AntennaBase::calculatePosition");
        }

        // pntpos関数用に設定
        for (int i = 0; i < dataNum; i++) {
            obs_data[i] = vecObsd[i];
        }

        // モニタ局位置計算
        //  ・azimuth/elevation angleは不要なのでNULL指定
        //  ・satellite statusは不要なのでNULL指定
        //  ・計算時刻はobs_dataの一つ目の時刻となる
        char msg[MAXSTRMSG];
        sol_t sol = { 0 };
        int ret = pntpos(obs_data, dataNum, navigationManager->getAll(),
                          opt, &sol, NULL, NULL, msg);

        // エラーチェック
        if (ret != 1) {
            ostringstream oStr;
            oStr << msg << "[receiverNo=" << receiverNo << "]";
            outputLog(TRACE_LEVEL_WARNING, oStr.str(), "AntennaBase::calculatePosition");
            continue;
        }

        // 局位置の結果を保管
        calcNum++;
        for (int i = 0; i < 3; i++) {
            result[i] += sol.rr[i];
        }
    }

    // 一回も計算できていない場合
    if (calcNum == 0) {
        ostringstream oStr;
        oStr << "no result [receiverNo=" << receiverNo << "]";
        throw AntLibException(oStr.str(), "AntennaBase::calculatePosition");
    }

    // 局位置を設定(求めた結果を平均)
    for (int i = 0; i < 3; i++) {
        position[i] = result[i] / calcNum;
    }

    // ログ出力
    ostringstream oStr;
    oStr << setprecision(12)
         << "position=" << position[0] << "," << position[1] << "," << position[2];
    oStr << "[receiverNo=" << receiverNo << "]";
    outputLog(TRACE_LEVEL_DEBUG, oStr.str(), "AntennaBase::calculatePosition");

    return position;
}

/** 局位置取得(ECEF)
 * @brief AntennaBase::getPosition
 * @return 局位置へのポインタ
 */
double* AntennaBase::getPosition() {
    return position;
}
