﻿/*------------------------------------------------------------------------------
* AntennaCalibrationsWriter.ccp : アンテナ特性出力クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "AntennaCalibrationsWriter.h"

using namespace std;

/** コンストラクタ
 * @brief AntennaCalibrationsWriter::AntennaCalibrationsWriter
 */
AntennaCalibrationsWriter::AntennaCalibrationsWriter()
{
}

/** デストラクタ
 * @brief AntennaCalibrationsWriter::~AntennaCalibrationsWriter
 */
AntennaCalibrationsWriter::~AntennaCalibrationsWriter()
{
    close();
}

/** ファイルオープン(追記)
 * @brief AntennaCalibrationsWriter::open
 * @param filename ファイルパス
 */
void AntennaCalibrationsWriter::open(string filename) {
    // ファイルのオープン(追記)
    formatWriter.open(filename, true);
}

/** ファイルクローズ
 * @brief AntennaCalibrationsWriter::close
 */
void AntennaCalibrationsWriter::close() {
    // ファイルのクローズ
    formatWriter.close();
}

/** アンテナ特性出力(NGS PCV用、PCV一つ)
 * @brief AntennaCalibrationsWriter::outputOneCalibrationsForPcv
 * @param sta 局情報
 * @param pcvs PCVの構造体
 * @param createTime 生成時刻
 * @param agency agency
 */
void AntennaCalibrationsWriter::outputOneCalibrationsForPcv(sta_t* sta, pcv_t* pcv, struct tm *createTime, string agency) {

    // [[[ Antenna Information(1行目) ]]]
    // Antenna type
    formatWriter.write(15, sta->antdes);
    formatWriter.writeSpace(1);

    // radome type
    formatWriter.write(4, "" );

    // Description of the Antenna type
    formatWriter.writeSpace(1);
    formatWriter.write(40, sta->antdes);

    // Data Source e.g. Agency Abbreviation
    formatWriter.writeSpace(1);
    formatWriter.write(3, agency);
    formatWriter.writeSpace(1);

    // # of individual antennas calibrated
    formatWriter.write("(");
    formatWriter.write(3, (int)1, true);
    formatWriter.write(")");

    // Date created
    formatWriter.writeSpace(1);
    outputDateForPcv(createTime);

    formatWriter.flush();

    // 周波数別のループ(L1/L2)
    for (int i = 0; i < 2; i++) {
        // [[[ NORTH EAST UP ]]]
        for (int j = 0; j < 3; j++) {
            double val = pcv->off[i][j] * 1e3;  // [mm]
            formatWriter.write(10, 1, val, true);
        }
        formatWriter.flush();

        // [[[ Elevation dependent pattern(1) ]]]
        for (int j = 0; j < 10; j++) {
            double val = pcv->var[i][j] * 1e3;  // [mm]
            formatWriter.write(6, 1, val, true);
        }
        formatWriter.flush();

        // [[[ Elevation dependent pattern(2) ]]]
        for (int j = 10; j < 19; j++) {
            double val = pcv->var[i][j] * 1e3;  // [mm]
            formatWriter.write(6, 1, val, true);
        }
        formatWriter.flush();
    }
}

/** 日付出力(NGS PCV用)
 * @brief AntennaCalibrationsWriter::outputDateForPcv
 * @param dateTime 日付
 */
void AntennaCalibrationsWriter::outputDateForPcv(struct tm *dateTime) {
    // YY/MM/DD
    int year = dateTime->tm_year % 100;
    formatWriter.write(2, year);
    formatWriter.write("/");
    formatWriter.write(2, dateTime->tm_mon + 1, '0', true);
    formatWriter.write("/");
    formatWriter.write(2, dateTime->tm_mday, '0', true);
}

/** アンテナ特性出力(ANTEX用、PCV一つ)
 * @brief AntennaCalibrationsWriter::outputOneCalibrationsForAntex
 * @param sta 局情報
 * @param pcv PCVの構造体
 * @param createTime 生成時刻
 * @param param ANTEX用パラメータ
 * @param vecFreqData 周波数別データ
 */
void AntennaCalibrationsWriter::outputOneCalibrationsForAntex(sta_t* sta, pcv_t* pcv,
                                                              struct tm *createTime, paramForAntex_t &param,
                                                              vector<string> *vecFreqData) {

    // [[[ START OF ANTENNA ]]]
    formatWriter.writeSpace(60);

    formatWriter.flush("START OF ANTENNA");


    // [[[ TYPE / SERIAL NO ]]]
    // Antenna type
    formatWriter.write(20, sta->antdes);

    // Serial number
    formatWriter.write(20, sta->antsno);

    // 受信機の場合はblank
    formatWriter.writeSpace(20);

    formatWriter.flush("TYPE / SERIAL NO");


    // [[[ METH / BY / # / DATE ]]]
    // Calibration method
    formatWriter.writeSpace(20);

    // Name of agency
    formatWriter.write(20, param.agency);

    // Number of individual antennas calibrated
    formatWriter.write(6, 2, true);
    formatWriter.writeSpace(4);

    // Date
    formatWriter.write(10, toDateForAntex(createTime));

    formatWriter.flush("METH / BY / # / DATE");


    // [[[ DAZI ]]]
    // Increment of the azimuth
    formatWriter.writeSpace(2);
    formatWriter.write(6, 1, param.dazi);

    formatWriter.writeSpace(52);
    formatWriter.flush("DAZI");


    // [[[ ZEN1 / ZEN2 / DZEN ]]]
    // 受信機：Definition of the grid in zenith angle
    // 衛星：Definition of the grid in nadir angle
    formatWriter.writeSpace(2);
    formatWriter.write(6, 1, param.zen1);
    formatWriter.write(6, 1, param.zen2);
    formatWriter.write(6, 1, param.dzen);

    formatWriter.writeSpace(40);
    formatWriter.flush("ZEN1 / ZEN2 / DZEN");


    // [[[ # OF FREQUENCIES ]]]
    // Number of frequencies
    formatWriter.write(6, vecFreqData->size(), true);

    formatWriter.writeSpace(54);
    formatWriter.flush("# OF FREQUENCIES");


    // [[[ VALID FROM ]]]
    outputDateForAntex(pcv->ts);

    formatWriter.writeSpace(17);
    formatWriter.flush("VALID FROM");


    // [[[ VALID UNTIL ]]]
    outputDateForAntex(pcv->te);

    formatWriter.writeSpace(17);
    formatWriter.flush("VALID UNTIL");


    // [[[ SINEX CODE ]]]
    formatWriter.write(10, "");

    formatWriter.writeSpace(17);
    formatWriter.flush("SINEX CODE");

    // 周波数分繰り返し
    for (int i = 0; i < vecFreqData->size(); i++) {
        formatWriter.write(vecFreqData->at(i));
    }

    // [[[ END OF ANTENNA ]]]
    formatWriter.writeSpace(60);

    formatWriter.flush("END OF ANTENNA");
}

/** 周波数別データ生成
 * @brief AntennaCalibrationsWriter::createFreqDataForAntex
 * @param pcv PCVの構造体
 * @param useSys 使用システム
 * @param freqNum 周波数の数
 * @param vecAzData 方位角別のデータ
 * @param mapNoAzData NOAZIのデータ
 * @return 文字列(周波数別データ)
 */
vector<string> AntennaCalibrationsWriter::createFreqDataForAntex(pcv_t* pcv, int useSys, int freqNum,
                                                              vector<pcvAzDataForAntex_t> *vecAzData,
                                                              map<double, double> *mapNoAzData) {
    // 周波数分繰り返し
    vector<string> vecString;
    for (int i = 0; i < freqNum; i++) {
        if (vecAzData[i].size() == 0) {
            // データがない場合はスキップ
            continue;
        }

        ostringstream oStr;

        // [[[ START OF FREQUENCY ]]]
        oStr << FormatWriter::toStringForSpace(3);
        oStr << outputFreqNameForAntex(useSys, i);

        oStr << FormatWriter::toStringForSpace(54);
        oStr << "START OF FREQUENCY" << endl;

        // [[[ NORTH / EAST / UP ]]]
        for (int j = 0; j < 3; j++) {
            double val = pcv->off[i][j] * 1e3;  //[mm]
            oStr << FormatWriter::toString(10, 2, val, true);
        }

        oStr << FormatWriter::toStringForSpace(30);
        oStr << "NORTH / EAST / UP" << endl;

        // [[[ Values ]]]
        // PCV出力(方位角依存なし)
        oStr << FormatWriter::toString(8, "   NOAZI");
        map<double, double>::iterator ite;
        for (ite = mapNoAzData[i].begin(); ite != mapNoAzData[i].end(); ite++) {
            double val = ite->second * 1e3;    // [mm]
            oStr << FormatWriter::toString(8, 2, val, true);
        }
        oStr << endl;

        // PCV出力(方位角依存)
        for (int azIdx = 0; azIdx < (int)vecAzData[i].size(); azIdx++) {
            oStr << FormatWriter::toString(8, 1, vecAzData[i].at(azIdx).az, true);
            for (int elIdx = 0; elIdx < (int)vecAzData[i].at(azIdx).vecPcv.size(); elIdx++) {
                double val = vecAzData[i].at(azIdx).vecPcv[elIdx] * 1e3;    // [mm]
                oStr << FormatWriter::toString(8, 2, val, true);
            }
            oStr << endl;
        }

        // [[[ END OF FREQUENCY ]]]
        oStr << FormatWriter::toStringForSpace(3);
        oStr << outputFreqNameForAntex(useSys, i);

        oStr << FormatWriter::toStringForSpace(54);
        oStr << "END OF FREQUENCY" << endl;

        // 周波数別データを追加
        vecString.push_back(oStr.str());
    }

    return vecString;
}

/** 日付出力(ANTEX用)
 * @brief AntennaCalibrationsWriter::outputDateForAntex
 * @param dateTime 日付
 */
void AntennaCalibrationsWriter::outputDateForAntex(gtime_t gtime) {
    struct tm *dateTime = localtime(&gtime.time);

    // YY/MM/DD
    int year = dateTime->tm_year + 1900;
    double sec = dateTime->tm_sec + gtime.sec;
    formatWriter.write(6, year, true);
    formatWriter.write(6, dateTime->tm_mon + 1, true);
    formatWriter.write(6, dateTime->tm_mday, true);
    formatWriter.write(6, dateTime->tm_hour, true);
    formatWriter.write(6, dateTime->tm_min, true);
    formatWriter.write(13, 7, sec, true);
}

/** 周波数名出力
 * @brief AntennaCalibrationsWriter::outputFreqNameForAntex
 * @param useSys 使用システム
 * @param useFreq 使用周波数
 */
string AntennaCalibrationsWriter::outputFreqNameForAntex(int useSys, int useFreq) {
    string freqName = "N/A";
    if (useSys == SYS_GLO) {
        // GLONASS
        if (useFreq == 0) {
            freqName = "R01";
        } if (useFreq == 1) {
            freqName = "R02";
        }
    } else if (useSys == SYS_QZS) {
        // QZS
        if (useFreq == 0) {
            freqName = "J01";
        } if (useFreq == 1) {
            freqName = "J02";
        } if (useFreq == 2) {
            freqName = "J05";
        }
    } else {
        // GPS
        if (useFreq == 0) {
            freqName = "G01";
        } if (useFreq == 1) {
            freqName = "G02";
        } if (useFreq == 2) {
            freqName = "G05";
        }
    }
    return freqName;
}

/** ANTEX用日付の取得
 * @brief AntennaCalibrationsWriter::toDateForAntex
 * @param dateTime 日付
 * @return 日付
 */
string AntennaCalibrationsWriter::toDateForAntex(struct tm *dateTime) {
    string monthStr[] = {"JAN", "FEB", "MAR", "APR", "MAY", "JUN",
                         "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};

    // YY-MMM-DD
    ostringstream oStr;
    oStr << setfill('0') << setw(2) <<  (dateTime->tm_year % 100);
    oStr << "-";
    oStr << setfill('0') << setw(2) <<  monthStr[dateTime->tm_mon];
    oStr << "-";
    oStr << setfill('0') << setw(2) <<  dateTime->tm_mday;

    return oStr.str();
}
