﻿/*------------------------------------------------------------------------------
* AntennaEstimater.ccp : 測定局のアンテナクラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include <iomanip>
#include "AntennaEstimater.h"
#include "AntennaBase.h"


extern "C" {
// in rtkpos4pcv.c
extern int rtkopenantest(const char *file);
extern void rtkcloseantest(void);
extern int rtkopenphase(const char *file);
extern void rtkclosephase(void);
// in sphharmonic.c
extern int sphharmonic(const double *azel, const prcopt_t *opt, double *sphCoef);
}

// debug用
//#define DEBUG_EKF                   // カルマンフィルタの入力行列出力
static int countForEkfDebug = 1;    // デバッグ用カウンタ

using namespace std;

/** コンストラクタ
 * @brief AntennaEstimater::AntennaEstimater
 */
AntennaEstimater::AntennaEstimater()
{
    receiverNo = RECEIVER_NO_TARGET;

    useFreq = 0;
    nMax[0] = 10;
    nMax[1] = 5;

    pcv_t pcvZero = { 0 };
    estimatePcvForFile = pcvZero;

    // PCV用パラメータ設定
    agencyForPcv = "NA";

    // ANTEX用パラメータ設定
    paramForAntex.agency = "NA";
    paramForAntex.dazi = 5.0;
    paramForAntex.zen1 = 0.0;
    paramForAntex.zen2 = 85.0;
    paramForAntex.dzen = 5.0;
}

/** デストラクタ
 * @brief AntennaEstimater::~AntennaEstimater
 */
AntennaEstimater::~AntennaEstimater()
{
}

/** 状態リセット
 * @brief AntennaEstimater::reset
 */
void AntennaEstimater::reset() {
    AntennaBase::reset();

    // 推定状況のリセット
    for (int i = 0; i < NFREQ; i++) {
        for (int j = 0; j < 3; j++) {
            phaseCenter[i][j] = 0.0;
        }
    }
}

/** アンテナ位相中心推定処理
 * @brief AntennaEstimater::estimatePhase
 * @param opt 処理オプション
 * @param solopt 解オプション
 * @param fileopt ファイルオプション
 * @param inputFiles 入力ファイルリスト
 * @param inputFileNum 入力ファイル数
 * @param outputFilepath 出力ファイルパス
 * @return アンテナ位相中心位置
 */
double* AntennaEstimater::estimatePhaseCenter(prcopt_t opt, const solopt_t *solopt, const filopt_t *fileopt,
                                              char **inputFiles, int inputFileNum, string outputFilepath) {
    // オプション設定
    opt.mode = PMODE_STATIC;
    opt.soltype = 0;            // 0:forward
    opt.modear = ARMODE_OFF;
    opt.antestmode = ANTEST_MODE_PC;

    // 相対測位の実行
    gtime_t timeAll = { 0 };
    int ret = postpos(timeAll, timeAll, 0.0, 0.0,
                      &opt, solopt, fileopt, inputFiles, inputFileNum,
                      (char*)outputFilepath.c_str(), "", "");

    // 計算できていない場合
    if (ret != 0) {
        stringstream oStr;
        oStr << "postpos failed. ret=" << ret;
        throw AntLibException(oStr.str(), "AntennaEstimater::estimatePhaseCenter");
    }

    // ログ出力
    outputVector(TRACE_LEVEL_DEBUG, "phaseCenter=", phaseCenter[useFreq], 3, 12, "AntennaEstimater::estimatePhaseCenter");

    // 位相中心の結果ファイルを読込＆結果を取得
    return readPcEstFile(outputFilepath);
}

/** アンテナ位相中心推定結果ファイル読込
 * @brief AntennaEstimater::readPcEstFile
 * @param inputFilepath 入力ファイルパス
 */
double* AntennaEstimater::readPcEstFile(string inputFilepath) {

    // ファイルオープン
    ifstream ifs;
    ifs.open(inputFilepath.c_str(), ios::in);
    if (!ifs.is_open()) {
        throw AntLibException("can't open file. filepath=" + inputFilepath, "AntennaEstimater::readPcEstFile");
    }

    // データ読込
    /* ※カンマ区切り
     *  0：時刻
     *  1：X(ECEF)
     *  2：Y(ECEF)
     *  3：Z(ECEF)
     *  4：ステータス
     */
    string buf;
    double result[3] = {0.0, 0.0, 0.0};
    int num = 0;
    while(ifs && getline(ifs, buf)) {
        if (buf[0] == '%') {
            // コメント行は無視
            continue;
        }

        // 対象データ読込
        vector<string> vecData = splitString(buf, ',');

        // 状態を取得
        int stat = atoi(vecData[4].c_str());
        if (stat == SOLQ_FIX || stat == SOLQ_FLOAT) {
            // FIX/Floatの解だけ取得
            num++;
            result[0] = atof(vecData[1].c_str());
            result[1] = atof(vecData[2].c_str());
            result[2] = atof(vecData[3].c_str());
        }
    }

    // ファイルクローズ
    ifs.close();

    // データ数チェック
    if (num == 0) {
        throw AntLibException("no result", "AntennaEstimater::readPcEstFile");
    }

    // 平均値を算出
    // ※最終解を採用
    for (int i = 0; i < 3; i++) {
        phaseCenter[useFreq][i] = result[i];
    }
    return phaseCenter[useFreq];
}

/** アンテナ位相中心変動パラメータ推定処理
 * @brief AntennaEstimater::estimatePhase
 * @param opt 処理オプション
 * @param solopt 解オプション
 * @param fileopt ファイルオプション
 * @param inputFiles 入力ファイルリスト
 * @param inputFileNum 入力ファイル数
 * @param outputFilepathForPos 出力ファイルパス(相対測位結果用)
 * @param outputFilepathForSph 出力ファイルパス(球面調和関数係数結果用)
 * @param outputFilepathForPhaseError 出力ファイルパス(搬送波位相一重差の残差)
 * @return true：推定実施、false：一回も推定できていない
 */
bool AntennaEstimater::estimatePCV(prcopt_t opt, const solopt_t *solopt, const filopt_t *fileopt,
                                   char **inputFiles, int inputFileNum,
                                   string outputFilepathForPos, string outputFilepathForSph,
                                   string outputFilepathForPhaseError) {
    outputLog(TRACE_LEVEL_INFO, "opt.antestvar=", opt.antestvar, "AntennaEstimater::estimatePCV");
#ifdef DEBUG_EKF    // debug
    countForEkfDebug = 1;
#endif

    // オプション設定
    opt.mode = PMODE_FIXED;
    opt.soltype = 2;            // 2:combined
    opt.modear = ARMODE_CONT;
    opt.tropopt = TROPOPT_EST;
    opt.antestmode = ANTEST_MODE_PCV;
    for (int i = 0; i < 3; i++) {
        opt.ru[i] = phaseCenter[useFreq][i];
    }

    // 球面調和関数係数&搬送波位相一重差の残差の結果出力(オープン)
    rtkopenantest(outputFilepathForSph.c_str());
    rtkopenphase(outputFilepathForPhaseError.c_str());

    // 相対測位の実行
    gtime_t timeAll = { 0 };
    int ret = postpos(timeAll, timeAll, 0.0, 0.0,
                      &opt, solopt, fileopt, inputFiles, inputFileNum,
                      (char*)outputFilepathForPos.c_str(), "", "");

    // 球面調和関数係数&搬送波位相一重差の残差の結果出力(クローズ)
    rtkcloseantest();
    rtkclosephase();

    // 計算できていない場合
    if (ret != 0) {
        return false;
    }

    // 推定実施
    return true;
}

/** アンテナ位相特性推定結果ファイル読込
 * @brief AntennaEstimater::readPcvEstFile
 * @param inputFilepath 入力ファイルパス
 */
void AntennaEstimater::readPcvEstFile(string inputFilepath) {

    // ファイルオープン
    ifstream ifs;
    ifs.open(inputFilepath.c_str(), ios::in);
    if (!ifs.is_open()) {
        throw AntLibException("can't open file. filepath=" + inputFilepath, "AntennaEstimater::readPcvEstFile");
    }

    // データ読込
    /* ※カンマ区切り
     *  0：時刻
     *  1：ステータス
     *  2-：球面調和関数係数Anm,Bnm
     */
    int estValNum = sphHarmonic(0, 0, nMax[0], nMax[1], NULL);
    double* estVal = mat(estValNum, 1);
    string buf;
    int num = 0;
    while(ifs && getline(ifs, buf)) {
        if (buf[0] == '%') {
            // コメント行は無視
            continue;
        }

        // 対象データ読込
        vector<string> vecData = splitString(buf, ',');

        // 状態を取得
        int stat = atoi(vecData[1].c_str());
        if (stat == SOLQ_FIX) {
            // FIXの解だけ取得
            num++;
            for (int i = 0; i < estValNum; i++) {
                estVal[i] = atof(vecData[i+2].c_str());
            }
        }
    }

    // ファイルクローズ
    ifs.close();

    // データ数チェック
    if (num == 0) {
        free(estVal);
        throw AntLibException("no result", "AntennaEstimater::readPcvEstFile");
    }

    // PCVを取得
    // ※最終解を採用
    estimatePcv.clear();
    for (int i = 0; i < estValNum; i++) {
        estimatePcv.push_back(estVal[i]);
    }
    free(estVal);
}

/** アンテナ位相特性ファイル生成
 * @brief AntennaEstimater::createAntennaCalibrations
 * @param type ファイルタイプ
 * @return true:PCO算出OK
 */
#define MASK_NORTH_AZ 45    // NOAZI計算時の北方向のマスク角
bool AntennaEstimater::createAntennaCalibrations(int type) {

    // [[[ PCV生成 ]]]
    // 測定局の緯度/経度取得
    // ※局位置設定は上位で行っておく
    double pos[3];
    ecef2pos(position, pos);

    // アンテナ位相中心の相対位置(ENU)を算出[m]
    double ecef[3];
    double enu[3];
    for (int i = 0; i < 3; i++) {
        ecef[i] = phaseCenter[useFreq][i] - position[i];
    }
    ecef2enu(pos, ecef, enu);
    outputVector(TRACE_LEVEL_INFO, "pos=", pos, 3, 12, "AntennaEstimater::createAntennaCalibrations");
    outputVector(TRACE_LEVEL_INFO, "ecef=", ecef, 3, 12, "AntennaEstimater::createAntennaCalibrations");
    outputVector(TRACE_LEVEL_INFO, "phaseCenter=", phaseCenter[useFreq], 3, 12, "AntennaEstimater::createAntennaCalibrations");
    outputVector(TRACE_LEVEL_INFO, "position=", position, 3, 12, "AntennaEstimater::createAntennaCalibrations");
    outputVector(TRACE_LEVEL_INFO, "enu=", enu, 3, 12, "AntennaEstimater::createAntennaCalibrations");

    // アンテナ位相中心のオフセット
    for(int i = 0; i < 3; i++) {
        estimatePcvForFile.off[useFreq][i] = enu[i];   // [m]
    }

    // アンテナ位相特性算出
    if (type == NGS_PCV) {
        // NGS PCVの場合
        vector<pcvAzDataForAntex_t> vecAz;
        map<double, double> mapPcv = calculatePcv(5, 0, 90, 5, MASK_NORTH_AZ, 360-MASK_NORTH_AZ, &vecAz);
        int idx = 0;
        for (double el = 90; el >= 0.0; el -= 5) {
            // PCVに設定
            double zen = 90 - el;
            estimatePcvForFile.var[useFreq][idx++] = mapPcv[zen];
        }

    } else if (type == ANTEX) {
        // ANTEXの場合
        vecAzDataForAntex[useFreq].clear();
        mapNoAziDataForAntex[useFreq] = calculatePcv(paramForAntex.dazi,
                        paramForAntex.zen1, paramForAntex.zen2, paramForAntex.dzen,
                        MASK_NORTH_AZ, 360-MASK_NORTH_AZ, &vecAzDataForAntex[useFreq]);
    }

    return true;
}

/** 推定結果を出力
 * @brief AntennaEstimater::outputPcvResult
 * @param outputFilepath 出力ファイルパス
 */
void AntennaEstimater::outputPcvResult(string outputFilepath) {
    // 出力ファイルのオープン
    ResultWriter resultWriter;
    resultWriter.open(outputFilepath, false);

    // (推定条件)
    resultWriter.writeLine(RESULT_PCV_PARAM_TAG);
    vector<double> vecParam;
    vecParam.push_back(nMax[0]);    // 球面調和関数オーダー
    vecParam.push_back(nMax[1]);    // 球面調和関数角
    resultWriter.writeLine(vecParam);

    // (球面調和関数出力)
    resultWriter.writeLine(RESULT_PCV_TAG);
    resultWriter.writeLine(estimatePcv);

    // アンテナ位相中心
    // (PCO)
    resultWriter.writeLine(RESULT_PCO_TAG);
    resultWriter.writeLine(estimatePcvForFile.off[useFreq], 3, 12);

    // (中心位置(ECEF))
    resultWriter.writeLine(RESULT_PC_ECEF_TAG);
    resultWriter.writeLine(phaseCenter[useFreq],3, 12);

    // ファイルクローズ
    resultWriter.close();
}

/** アンテナ位相特性ファイル出力(システム別)
 * @brief AntennaEstimater::outputAntennaCalibrationsForSys
 * @param type ファイルタイプ
 * @param freqNum 周波数の数
 * @param outputFilepath 出力ファイルパス
 */
void AntennaEstimater::outputAntennaCalibrationsForSys(int type, int freqNum, string outputFilepath) {
    outputLog(TRACE_LEVEL_DEBUG, "agencyForPcv=" + agencyForPcv, "AntennaEstimater::outputAntennaCalibrations");

    // 現在時刻取得(生成時刻)
    gtime_t curTime = getNowTime();
    struct tm *nowTime;
    nowTime = localtime(&curTime.time);

    // 測定局の観測データの時刻を取得
    vector<gtime_t> vecDataTime = observationManager->getAllDataTime();

    // データ期間
    estimatePcvForFile.ts = vecDataTime[0];
    estimatePcvForFile.te = vecDataTime[vecDataTime.size() - 1];

    // 局情報取得
    sta_t* sta = observationManager->getStationParam();

    // [[[ PCV出力 ]]]
    if (type == NGS_PCV && (useSys & SYS_GPS)) {
        // 出力ファイルのオープン(追記)
        AntennaCalibrationsWriter writer;
        writer.open(outputFilepath);
        writer.outputOneCalibrationsForPcv(sta, &estimatePcvForFile, nowTime, agencyForPcv);

        // ファイルクローズ
        writer.close();

    } else if (type == ANTEX) {
        // 周波数別データを生成
        vector<string> vecStr = AntennaCalibrationsWriter::createFreqDataForAntex(&estimatePcvForFile, useSys, freqNum,
                                             vecAzDataForAntex, mapNoAziDataForAntex);

        for (int i = 0; i < (int)vecStr.size(); i++) {
            vecFreqDataForAntex.push_back(vecStr[i]);
        }
    }
}

/** アンテナ位相特性ファイル出力(ANTEX用)
 * @brief AntennaEstimater::outputAntennaCalibrationsForAntex
 * @param outputFilepath 出力ファイルパス
 */
void AntennaEstimater::outputAntennaCalibrationsForAntex(string outputFilepath) {
    outputLog(TRACE_LEVEL_DEBUG, "agencyForAntex=" + paramForAntex.agency, "AntennaEstimater::outputAntennaCalibrations");

    // 現在時刻取得(生成時刻)
    gtime_t curTime = getNowTime();
    struct tm *nowTime;
    nowTime = localtime(&curTime.time);

    // 測定局の観測データの時刻を取得
    vector<gtime_t> vecDataTime = observationManager->getAllDataTime();

    // データ期間
    estimatePcvForFile.ts = vecDataTime[0];
    estimatePcvForFile.te = vecDataTime[vecDataTime.size() - 1];

    // 局情報取得
    sta_t* sta = observationManager->getStationParam();

    // [[[ PCV出力 ]]]
    // 出力ファイルのオープン(追記)
    AntennaCalibrationsWriter writer;
    writer.open(outputFilepath);
    writer.outputOneCalibrationsForAntex(sta, &estimatePcvForFile, nowTime, paramForAntex,
                                         &vecFreqDataForAntex);
}

/** 球面調和関数近似係数計算
 * @brief AntennaEstimater::sphharmonic
 * @param az 方位角[rad]
 * @param el 仰角[rad]
 * @param nMax オーダー
 * @param mMax 角
 * @param sphCoef [OUT]sphCoef(A00 A10 A11 A20 A21 ... B11, B21 ...) NULLの場合：係数の数をカウントするだけ
 * @return 係数の数
 */
int AntennaEstimater::sphHarmonic(const double az, const double el, int nMax, int mMax, double *sphCoef)
{
    double azel[2] = {az, el};
    prcopt_t opt = prcopt_default;
    opt.antestnmax[0] = nMax;
    opt.antestnmax[1] = mMax;

    // 球面調和関数算出
    return sphharmonic(azel, &opt, sphCoef);
}

/** 同じ衛星データの抽出
 * @brief AntennaEstimater::getSameSat
 * @param vecReferenceObsData 観測データ(基準局)
 * @param vecTargetObsData 観測データ(測定局)
 * @param useSys 使用システム
 * @param referenceObsData [OUT]抽出後の観測データ(基準局)
 * @param targetObsData [OUT]抽出後の観測データ(測定局)
 * @return 衛星数
 */
int AntennaEstimater::getSameSat(vector<obsd_t> &vecReferenceObsData, vector<obsd_t> &vecTargetObsData, int useSys,
                                  obsd_t *referenceObsData, obsd_t *targetObsData) {

    // 同じ衛星データのみを抽出
    vector<unsigned char> vecSat;
    for (int i = 0; i < (int)vecReferenceObsData.size(); i++) {
        for (int j = 0; j < (int)vecTargetObsData.size(); j++) {
            // 使用するシステムによる衛星の取捨
            int sys = satsys(vecReferenceObsData[i].sat, NULL);
            if (!(sys & useSys)) {
                continue;
            }

            if (vecReferenceObsData[i].sat == vecTargetObsData[j].sat) {
                vecSat.push_back(vecReferenceObsData[i].sat);
            }
        }
    }

    // データ数のチェック
    int dataNum = vecSat.size();
    if (dataNum > MAXOBS) {
        return 0;
    }

    // 衛星別のデータの準備
    for (int i = 0; i < dataNum; i++) {
        // 基準局の観測データ取得
        for (int j = 0; j < (int)vecReferenceObsData.size(); j++) {
            if (vecSat[i] == vecReferenceObsData[j].sat) {
                referenceObsData[i] = vecReferenceObsData[j];
            }
        }

        // 測定局の観測データ取得
        for (int j = 0; j < (int)vecTargetObsData.size(); j++) {
            if (vecSat[i] == vecTargetObsData[j].sat) {
                targetObsData[i] = vecTargetObsData[j];
            }
        }
    }

    return dataNum;
}

/** アンテナ位相特性からアンテナ位相中心変動の算出
 * @brief AntennaEstimater::calculatePcvByAntennaModel
 * @param antIdx アンテナINDEX
 * @param opt アンテナ位相特性
 * @param azel 方位角/仰角
 * @return アンテナ位相中心変動
 */
double AntennaEstimater::calculatePcvByAntennaModel(int antIdx, obsd_t *obs, prcopt_t *opt, double *azel) {
    // アンテナ位相中心変動の算出
    double dantr[NFREQ]={0};
    antmodel(&opt->pcvr[antIdx], satsys(obs->sat,NULL), opt->antdel[antIdx], azel, opt->posopt[1], dantr);

    return dantr[useFreq];
}

/** PCV算出
 * @brief AntennaEstimater::calculatePcv
 * @param daz 方位角の間隔[deg]
 * @param startZen 天頂角の開始角度[deg]
 * @param endZen 天頂角の終了角度[deg]
 * @param dZen 天頂角の間隔[deg]
 * @param noAziStart 天頂角の開始角度(NOAZI用)[deg]
 * @param noAziEnd　天頂角の終了角度(NOAZI用)[deg]
 * @param vecAz [OUT]方位角別のPCV
 * @return PCV
 */
map<double, double> AntennaEstimater::calculatePcv(double daz, double startZen, double endZen, double dZen,
                                                     double noAziStart, double noAziEnd,
                                                     vector<pcvAzDataForAntex_t> *vecAz) {
    // 推定対象：球面調和関数(Anm,Bnm)
    int estValNum = estimatePcv.size();
    double* estVal = mat(estValNum, 1);
    for (int i = 0; i < estValNum; i++) {
        estVal[i] = estimatePcv[i];
    }

    // 方位角別にPCVを算出
    map<double, int> mapDataNum;
    map<double, double> mapNoAziData;
    double sphCoef[ANTEST_NMAX * ANTEST_NMAX];
    double azel[2];
    for (double az = 0; az <= 360; az += daz) {
        // 方位角を保存
        pcvAzDataForAntex_t azData = { 0 };
        azData.az = az;

        // 天頂角(仰角)別にデータを生成
        for (double zen = startZen; zen <= endZen; zen += dZen) {
            // 測定局のアンテナ位相中心変動
            double el = 90 - zen;
            azel[0] = az * D2R;
            azel[1] = el * D2R;
            sphHarmonic(azel[0] + azOffset, azel[1], nMax[0], nMax[1], sphCoef);
            double targetOffset = dot(sphCoef, estVal, estValNum);

            // PCVに設定
            azData.vecPcv.push_back(targetOffset);

            // 天頂角別データの算出
            if (az >= noAziStart && az <= noAziEnd) {
                map<double, int>::iterator ite = mapDataNum.find(zen);
                if (ite == mapDataNum.end()) {
                    mapDataNum.insert(pair<double, int>(zen, 1));
                    mapNoAziData.insert(pair<double, double>(zen, targetOffset));
                } else {
                    mapDataNum[zen] += 1;
                    mapNoAziData[zen] += targetOffset;
                }
            }
        }
        vecAz->push_back(azData);
    }

    // 天頂角別のPCVを算出(方位角方向平均)
    for (double zen = startZen; zen <= endZen; zen += dZen) {
        map<double, int>::iterator ite = mapDataNum.find(zen);
        if (ite == mapDataNum.end()) {
            continue;
        }

        int num = ite->second;
        if (num == 0) {
            continue;
        }

        // 平均値を算出
        mapNoAziData[zen] /= num;
    }

    // メモリ解放
    free(estVal);

    return mapNoAziData;
}
