﻿/*------------------------------------------------------------------------------
* AntennaEstimater.h : 測定局のアンテナクラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#ifndef ANTENNAESTIMATER_H
#define ANTENNAESTIMATER_H
#include <map>
#include "AntennaBase.h"
#include "AntLibException.h"
#include "ResultWriter.h"
#include "AntennaCalibrationsWriter.h"


#define RESULT_PC_DATA_TAG "<pc_data>"          // 結果ファイルの出力タグ：アンテナ位相中心推定データ
#define RESULT_PCV_PARAM_TAG "<pcv_param>"      // 結果ファイルの出力タグ：推定条件
#define RESULT_PCV_TAG "<pcv>"                  // 結果ファイルの出力タグ：球面調和関数(Anm,Bnm)
#define RESULT_PCO_TAG "<pco>"                  // 結果ファイルの出力タグ：アンテナ位相中心オフセット
#define RESULT_PC_ECEF_TAG "<pc_ecef>"          // 結果ファイルの出力タグ：アンテナ位相中心位置(ECEF)


/** 測定局のアンテナクラス
 * @brief The AntennaEstimater class
 */
class AntennaEstimater : public AntennaBase
{
public:
    AntennaEstimater();
    ~AntennaEstimater();

    virtual void reset();
    double* estimatePhaseCenter(prcopt_t opt, const solopt_t *solopt, const filopt_t *fileopt,
                                char **inputFiles, int inputFileNum, std::string outputFilepath);
    double* readPcEstFile(std::string inputFilepath);
    bool estimatePCV(prcopt_t opt, const solopt_t *solopt, const filopt_t *fileopt,
                     char **inputFiles, int inputFileNum,
                     std::string outputFilepathForPos, std::string outputFilepathForSph,
                     std::string outputFilepathForPhaseError);
    void readPcvEstFile(std::string inputFilepath);
    bool createAntennaCalibrations(int type);
    void outputPcvResult(std::string outputFilepath);
    void outputAntennaCalibrationsForSys(int type, int freqNum, std::string outputFilepath);
    void outputAntennaCalibrationsForAntex(std::string outputFilepath);

    // static
    static int sphHarmonic(const double az, const double el, int nMax, int mMax, double *sphCoef);
    static int getSameSat(std::vector<obsd_t> &vecReferenceObsData, std::vector<obsd_t> &vecTargetObsData, int useSys,
                                      obsd_t *referenceObsData, obsd_t *targetObsData);


    // -----< 推定用パラメータ >-----
    /** 使用システム
     * @brief useSys
     */
    int useSys;

    /** 使用周波数(0：L1、1：L2、2：L5)
     * @brief useFreq
     */
    int useFreq;

    /** カットオフSNR/CN0
     * @brief sigError
     */
    double minForSN;

    /** 球面調和関数オーダー/角
     * 0：オーダー[次]、1：角[度]
     * @brief nMax
     */
    int nMax[2];

    /** アンテナ方位角オフセット[rad]
     * @brief azOffset
     */
    double azOffset;

    /** PCV用パラメータ
     * @brief agencyForPcv
     */
    std::string agencyForPcv;

    /** ANTEX用パラメータ
     * @brief paramForAntex
     */
    paramForAntex_t paramForAntex;

private:
    double calculatePcvByAntennaModel(int antIdx, obsd_t *obs, prcopt_t *opt, double *azel);
    std::map<double, double> calculatePcv(double daz, double startZen, double endZen, double dZen,
                                                         double noAziStart, double noAziEnd,
                                                         std::vector<pcvAzDataForAntex_t> *vecAz);

    /** アンテナ位相中心
     * @brief phaseCenter
     */
    double phaseCenter[NFREQ][3];

    /** アンテナ位相特性パラメータ(バイアス(衛星),Anm, Bnm)
     * @brief estimatePcv
     */
    std::vector<double> estimatePcv;

    /** アンテナ位相特性パラメータ(アンテナ位相特性ファイル用)
     * @brief estimatePcvForFile
     */
    pcv_t estimatePcvForFile;

    /** アンテナ位相特性パラメータ(アンテナ位相特性ファイル用, ANTEX用)
     * @brief vecAzData
     */
    std::vector<pcvAzDataForAntex_t> vecAzDataForAntex[NFREQ];

    /** アンテナ位相特性パラメータ(アンテナ位相特性ファイル用NOAZI, ANTEX用)
     * @brief mapNoAziDataForAntex
     */
    std::map<double, double> mapNoAziDataForAntex[NFREQ];

    /** 周波数別データ(ANTEX用)
     * @brief vecFreqDataForAntex
     */
    std::vector<std::string> vecFreqDataForAntex;
};

#endif // ANTENNAESTIMATER_H
