﻿/*------------------------------------------------------------------------------
* FormatWriter.ccp : 書式出力クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "FormatWriter.h"

using namespace std;

/** コンストラクタ
 * @brief FormatWriter::FormatWriter
 */
FormatWriter::FormatWriter()
{
}

/** デストラクタ
 * @brief FormatWriter::~FormatWriter
 */
FormatWriter::~FormatWriter()
{
    close();
}

/** 文字列出力
 * @brief FormatWriter::write
 * @param str 文字列
 */
void FormatWriter::write(string str) {
    if (!isOpen()) {
        // 開いてないなら何もしない
        return;
    }

    // 左詰めで指定文字数で出力
    outputStream << str;
}

/** 文字列出力
 * @brief FormatWriter::write
 * @param num 文字数
 * @param str 文字列
 */
void FormatWriter::write(int num, string str) {
    if (!isOpen()) {
        // 開いてないなら何もしない
        return;
    }

    // 左詰めで指定文字数で出力
    outputStream << toString(num, str);
}

/** 数値出力(整数)
 * @brief FormatWriter::write
 * @param num 桁数
 * @param val 数値
 * @param rightFlag true:右詰
 */
void FormatWriter::write(int num, int val, bool rightFlag) {
    write(num, val, ' ', rightFlag);
}

/** 数値出力(整数)
 * @brief FormatWriter::write
 * @param num 桁数
 * @param val 数値
 * @param fillChar 埋める文字
 * @param rightFlag true:右詰
 */
void FormatWriter::write(int num, int val, char fillChar, bool rightFlag) {
    if (!isOpen()) {
        // 開いてないなら何もしない
        return;
    }

    outputStream << toString(num, val, fillChar, rightFlag);
}

/** 数値出力(少数以下あり)
 * @brief FormatWriter::write
 * @param num 桁数
 * @param precision 少数以下の桁数
 * @param val 数値
 * @param rightFlag true:右詰
 */
void FormatWriter::write(int num, int precision, double val, bool rightFlag) {
    if (!isOpen()) {
        // 開いてないなら何もしない
        return;
    }

    outputStream << toString(num, precision, val, rightFlag);
}

/** 空白出力
 * @brief FormatWriter::writeSpace
 * @param num スペースの数
 */
void FormatWriter::writeSpace(int num) {
    if (!isOpen()) {
        // 開いてないなら何もしない
        return;
    }

    outputStream << toStringForSpace(num);
}

/** 改行＆flush
 * @brief FormatWriter::flush
 */
void FormatWriter::flush() {
    if (!isOpen()) {
        // 開いてないなら何もしない
        return;
    }

    // 改行＆flush
    outputStream << endl;
    outputStream.flush();
}

/** 改行＆flush(文字列付き)
 * @brief FormatWriter::write
 * @param str 文字列
 */
void FormatWriter::flush(string str) {
    if (!isOpen()) {
        // 開いてないなら何もしない
        return;
    }

    // 左詰めで指定文字数で出力
    outputStream << str << endl;
    outputStream.flush();
}

/** 文字列化
 * @brief FormatWriter::toString
 * @param num 文字数
 * @param str 文字列
 */
string FormatWriter::toString(int num, string str) {
    // 文字数チェック
    string outStr = str;
    if ((int)str.length() > num) {
        outStr = str.substr(0, num);
    }

    // 左詰めで指定文字数で出力
    ostringstream ostr;
    ostr << setfill(' ') << setw(num) << left << outStr;
    return ostr.str();
}

/** 文字列化:数値出力(整数)
 * @brief FormatWriter::toString
 * @param num 桁数
 * @param val 数値
 * @param fillChar 埋める文字
 * @param rightFlag true:右詰
 */
string FormatWriter::toString(int num, int val, char fillChar, bool rightFlag) {
    ostringstream ostr;
    if (rightFlag) {
        // 右詰めで数値を出力
        ostr << setfill(fillChar) << setw(num) << right <<  val;
    } else {
        // 左詰めで数値を出力
        ostr << setfill(fillChar) << setw(num) << left <<  val;
    }
    return ostr.str();
}

/** 文字列化:数値出力(少数以下あり)
 * @brief FormatWriter::toString
 * @param num 桁数
 * @param precision 少数以下の桁数
 * @param val 数値
 * @param rightFlag true:右詰
 */
string FormatWriter::toString(int num, int precision, double val, bool rightFlag) {
    ostringstream ostr;
    if (rightFlag) {
        // 右詰めで数値を出力
        ostr << setfill(' ') << setw(num) << right << fixed << setprecision(precision) << val;
    } else {
        // 左詰めで数値を出力
        ostr << setfill(' ') << setw(num) << left << fixed << setprecision(precision) << val;
    }
    return ostr.str();
}

/** 空白出力
 * @brief FormatWriter::writeSpace
 * @param num スペースの数
 */
string FormatWriter::toStringForSpace(int num) {
    ostringstream ostr;
    for (int i = 0; i < num; i++) {
        ostr << " ";
    }
    return ostr.str();
}
