﻿/*------------------------------------------------------------------------------
* GlonassEstimater.ccp : GLONASS受信機チャンネル間バイアス推定クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "GlonassEstimater.h"
#include "FormatWriter.h"
#include <math.h>
#include <stdio.h>
#include <time.h>

using namespace std;
#pragma warning(disable:4996)

#define SAFE_FREE(A)	if ((A)) {free((A));(A)=NULL;}
#define INITVAL_FOR_EST 1e-14   // 推定パラメータの初期値

// debug用
//#define DEBUG_EKF                   // カルマンフィルタの入力行列出力
static int countForEkfDebug = 0;    // デバッグ用カウンタ

/** コンストラクタ
 * @brief GlonassEstimater::GlonassEstimater
 */
GlonassEstimater::GlonassEstimater(void)
{
	m_init = false;
	m_X = NULL;
    m_navMng = NULL;
    m_nav = NULL;
	m_freq = 0;
	memset(m_referPos, 0, sizeof(m_referPos));
	memset(m_targetPos, 0, sizeof(m_targetPos));
	m_sigx = 0;
	m_sigp = 0;
}

/** デストラクタ
 * @brief GlonassEstimater::~GlonassEstimater
 */
GlonassEstimater::~GlonassEstimater(void)
{
    reset();
}

/**
 * @brief 初期化 
 *
 * @param	navMng	[IN] : 航法データ管理クラス
 */
void GlonassEstimater::initialize(NavigationManager* navMng)
{
    if (!navMng)
	{
		return;
	}
	
	m_navMng = navMng;
	
	m_init = true;
}
/**
 * @brief リセット処理
 *			L1 L2 計算前に呼び出すこと
 *
 */
void GlonassEstimater::reset()
{
	m_avrBF[0] = 0.0;
	m_avrBF[1] = 0.0;
	m_freq = 0;
	m_referObsMng  = NULL;
	m_targetObsMng = NULL;
	memset(m_referPos, 0, sizeof(m_referPos));
	memset(m_targetPos, 0, sizeof(m_targetPos));
	m_sigx = 0;
	m_sigp = 0;

    // マップクリア
    for (int i = 0; i < NSATGLO; i++) {
        m_mapPathInfo[i].clear();
    }

    // メモリ解放
    for (int i = 0; i < (int)m_vecSlopeData.size(); i++) {
        delete(m_vecSlopeData[i]);
    }
    m_vecSlopeData.clear();
}

/**
 * @brief GLONASS受信機チャンネル間バイアス推定 
 *
 * @param	freq				[IN] : 使用周波数(0：L1、1：L2)
 * @param	opt         		[IN] : 推定オプション
 * @param	referenceObsMng		[IN] : 基準観測データ管理クラス
 * @param	referencePosition	[IN] : 基準アンテナ位相中心
 * @param	targetObsMng		[IN] : ターゲット観測データ管理クラス
 * @param	targetPosition		[IN] : ターゲットアンテナ位相中心
 * @param	sigx				[IN] : アプリオリ共分散行列(標準偏差)
 * @param	sigp				[IN] : 観測誤差の標準偏差
 * @param	calcInterval		[IN] : 衛星バイアス推定をする間隔[sec]
 * @param	pathMinRange		[IN] : パスを有効とする最小時間[sec]
 * @param	outputFilepathForBias   [IN] : 出力ファイルパス(バイアス推定用)
 * @param	outputFilepath          [IN] : 出力ファイルパス
 * @param   sigNum
 * @param   repeat
 */
void GlonassEstimater::estimateBias(int freq, prcopt_t opt,
									ObservationManager *referenceObsMng, double* referencePosition, 
									ObservationManager *targetObsMng, double* targetPosition,
                                    double sigx, double sigp,
                                    int calcInterval, int pathMinRange,
                                    string outputFilepathForBias, string outputFilepath,
                                    double sigNum, int repeat, std::vector<std::string> &vecMessage)
{
    outputLog(TRACE_LEVEL_INFO, "m_init=", m_init, "GlonassEstimater::estimateBias");
    outputLog(TRACE_LEVEL_INFO, "sigp=", sigp, "GlonassEstimater::estimateBias");
    outputLog(TRACE_LEVEL_INFO, "sigx=", sigx, "GlonassEstimater::estimateBias");
    outputVector(TRACE_LEVEL_INFO, "referencePosition=", referencePosition, 3, 12, "GlonassEstimater::estimateBias");
    outputVector(TRACE_LEVEL_INFO, "targetPosition=", targetPosition, 3, 12, "GlonassEstimater::estimateBias");

    // 初期化チェック
	if (!m_init)
	{
        throw AntLibException("not initialized", "GlonassEstimater::estimateBias");
	}

	// 引数チェック
	if (!referenceObsMng	||
		!referencePosition	||
		!targetObsMng		||
		!targetPosition	)
	{
        throw AntLibException("position or manager is null", "GlonassEstimater::estimateBias");
    }
	
	if (freq != 0 && freq != 1)
	{
        ostringstream oStr;
        oStr << "invalid freq. freq=" << freq;
        throw AntLibException(oStr.str(), "GlonassEstimater::estimateBias");
    }

    m_freq = freq;							///< 使用周波数(0：L1、1：L2)
	m_referObsMng  = referenceObsMng;		///< 基準観測データ管理クラス
	m_targetObsMng = targetObsMng;			///< ターゲット観測データ管理クラス
	m_referPos[0]  = referencePosition[0];	///< 基準アンテナ位相中心
	m_referPos[1]  = referencePosition[1];	///< 基準アンテナ位相中心
	m_referPos[2]  = referencePosition[2];	///< 基準アンテナ位相中心
	m_targetPos[0] = targetPosition[0];		///< ターゲットアンテナ位相中心
	m_targetPos[1] = targetPosition[1];		///< ターゲットアンテナ位相中心
	m_targetPos[2] = targetPosition[2];		///< ターゲットアンテナ位相中心
    m_sigx = sigx;							///< アプリオリ共分散行列(標準偏差)
    m_sigp = sigp;                          ///< 観測誤差の標準偏差
    m_calcInterval = calcInterval;
    m_pathMinRange = pathMinRange;
	
    // 航法データ取得
    if (!m_navMng)
    {
        throw AntLibException("NavigationManager is null", "GlonassEstimater::estimateBias");
    }
    m_nav = m_navMng->getAll();
    if (!m_nav)
    {
        throw AntLibException("nav_t is null", "GlonassEstimater::estimateBias");
    }

	// Bruiの初期化
    m_X = zeros(NSATGLO, 1);

    // 初期値
    // ※初期値が０のままだとEKF(rtklib filter())が計算できないため
    //   とりあえず小さな値を初期値とする
    for (int i=0;i<NSATGLO;i++)
	{
        m_X[i] = INITVAL_FOR_EST;
        m_pathNo[i] = 0;
        m_freqNo[i] = FREQ_NO_MIN - 1;
	}

    // 周波数番号を取得
    int prn;
    for (int i=0;i<m_nav->ng;i++) {
        if (satsys(m_nav->geph[i].sat, &prn) != SYS_GLO)
        {
            outputLog(TRACE_LEVEL_WARNING, "not GLONASS. sat=",  (int)m_nav->geph[i].sat, "GlonassEstimater::estimateBias");
            continue;
        }

        int gloIdx =  prn - MINPRNGLO;
        if (m_freqNo[gloIdx] >= FREQ_NO_MIN) {
            // 既に周波数番号を取得しているなら無視
            continue;
        }

        // 周波数番号取得
        m_freqNo[gloIdx] = m_nav->geph[i].frq;
    }
    
    // 搬送波位相バイアス推定
    if (!CalcBias(&opt, outputFilepathForBias))
    {
		SAFE_FREE(m_X);
        throw AntLibException("error occurred in CalcBias()", "GlonassEstimater::estimateBias");
    }

    // 搬送波位相バイアスの読込
    readBiasFile(outputFilepathForBias);


	// 周波数間バイアスの傾き計算
    if (!CalcSlope(outputFilepath, sigNum, repeat, vecMessage))
	{
		SAFE_FREE(m_X);

        // エラーメッセージ生成
        stringstream errStr;
        errStr << "error occurred in CalcSlope()";
        if (m_numFreq0 == 0) {
            errStr << " ('FreqNo=0' is not found)";
        } else if (m_numFreq1 == 0) {
            errStr << " ('FreqNo=1 or -1' is not found)";
        } else {
            errStr << " (base combination of frequency is not found)";
        }
        throw AntLibException(errStr.str(), "GlonassEstimater::estimateBias");
    }
	
	// 終了処理
	SAFE_FREE(m_X);
}

/**
 * @brief 搬送波位相バイアス推定
 * @param	opt         		[IN] : 推定オプション
 * @param	outputFilepath		[IN] : 出力ファイルパス
 */
bool GlonassEstimater::CalcBias(prcopt_t *opt, string outputFilepath)
{
    // 出力ファイルのオープン
    ResultWriter resultWriter;
    resultWriter.open(outputFilepath, false);

    // 観測データの最終受信時刻
    gtime_t gtimeZero = { 0 };
    gtime_t lastDataTime[NSATGLO];
    for (int i = 0; i < NSATGLO; i++) {
        lastDataTime[i] = gtimeZero;
    }

    // rtkの初期化(slip検知のみに使用)
    rtk_t rtk = { 0 };
    rtk.opt = *opt;
    ssat_t ssat0={0};
    for (int i=0;i<MAXSAT;i++) rtk.ssat[i]=ssat0;

	// 全衛星数(M)
    int M = NSATGLO;

	// アプリオリ誤差共分散行列
    double workP[NSATGLO * NSATGLO];
    double *P = apriori();

	// アンテナ基準局の観測データの時刻を取得
	bool startLastIndex = false;
	vector<gtime_t> vecDataTime = m_referObsMng->getAllDataTime();
  
	// 時刻のループ
    double workX[NSATGLO];
	bool calcFlag = false;
    for (int timeIdx=0;timeIdx<(int)vecDataTime.size();timeIdx++) 
    {
		if (!m_targetObsMng->containTime(vecDataTime[timeIdx])) 
		{
			// 同じ時刻のデータがない場合は無視
			continue;
		}

		// データを取得
		m_referObs	= m_referObsMng->getData(&vecDataTime[timeIdx], NULL, startLastIndex);
		m_targetObs	= m_targetObsMng->getData(&vecDataTime[timeIdx], NULL, startLastIndex);
		startLastIndex = true;

		// GLONASSのみを抽出
        int baseIdx = 0;
        double maxEl = 0.0;
		vector<calc_sat_t> vecSat;
		for (int i=0;i<(int)m_referObs.size();i++) 
		{
			for (int j=0;j<(int)m_targetObs.size();j++) 
			{
				// カレント時刻のGLONASSの取得
				int rsat;
				int tsat;
				if ((satsys(m_referObs[i].sat, &rsat)  == SYS_GLO &&
					 satsys(m_targetObs[j].sat, &tsat) == SYS_GLO) &&
                     m_referObs[i].sat == m_targetObs[j].sat)
				{
                    int satIdx  = m_referObs[i].sat - 1;
                    if (m_nav->lam[satIdx][m_freq] == 0.0) {
                        // 航法MSGがない場合はスキップ
                        continue;
                    }

                    calc_sat_t csat;
                    double	satClock[MAXOBS];
                    double	satVariance[MAXOBS];
                    int		satHealth[MAXOBS];

                    // 基準局の観測データで、衛星位置計算
                    satposs(vecDataTime[timeIdx], &m_referObs[i], 1, m_nav, EPHOPT_BRDC,
                            csat.satPosRef, satClock, satVariance, satHealth);

                    // 基準局衛星間距離
                    double eRefer[3];
                    csat.distToRef = geodist(csat.satPosRef, m_referPos, eRefer);

                    // 仰角の算出
                    double pos[3];
                    ecef2pos(m_referPos,pos);
                    satazel(pos, eRefer, csat.azel);

#if 0               // 仰角チェック
                    double el = csat.azel[1] * R2D;
                    if (el < 20) {
                        continue;
                    }
#endif
#if 0               // 最大仰角のチェック
                    if (maxEl <csat.azel[1]) {
                        maxEl = csat.azel[1];
                        baseIdx = vecSat.size();
                    }
#endif

                    // 測定局の観測データで、衛星位置計算
                    satposs(vecDataTime[timeIdx], &m_targetObs[j], 1, m_nav, EPHOPT_BRDC,
                            csat.satPosTarget, satClock, satVariance, satHealth);

                    // ターゲット局衛星間距離
                    double eTarget[3];
                    csat.distToTarget = geodist(csat.satPosTarget, m_targetPos, eTarget);

                    // 計算対象に追加
                    csat.satIdx  = satIdx;
                    csat.gloIdx  = rsat - MINPRNGLO;
                    csat.referIdx  = i;
                    csat.targetIdx = j;
                    vecSat.push_back(csat);
				}
			}
		}
		
		// 計算対象がない場合次の時刻へ
        if (vecSat.size() == 0)
		{
			continue;
		}
		
		// 観測衛星数(m)
		int m = (int)vecSat.size();
	
        // 状態更新
        updateState(&rtk);
        for (int i = 0; i < m; i++) {
            int gloIdx = vecSat[i].gloIdx;
            int satIdx = vecSat[i].satIdx;
            double dt = timediff(vecDataTime[timeIdx], lastDataTime[gloIdx]);

            // 最終受信時刻を保管
            lastDataTime[gloIdx] = vecDataTime[timeIdx];

            // パス切断orサイクルスリップチェック
            vecSat[i].slip = rtk.ssat[satIdx].slip[m_freq];
            vecSat[i].gf = rtk.ssat[satIdx].gf;
            if (rtk.ssat[satIdx].slip[m_freq] == 0
                    && dt < ARC_BREAK_LIMIT) {
                // データに問題がない場合は、次へ
                continue;
            }

            // パス番号加算
            m_pathNo[gloIdx]++;

            // 衛星バイアス値のリセット
            vecSat[i].reset = 1;
#if 1       // (初期値あり)
            double dL = m_targetObs[vecSat[i].targetIdx].L[m_freq] - m_referObs[vecSat[i].referIdx].L[m_freq];
            double dP = m_targetObs[vecSat[i].targetIdx].P[m_freq] - m_referObs[vecSat[i].referIdx].P[m_freq];
            m_X[gloIdx] = dL - dP / m_nav->lam[satIdx][m_freq];
#else       // (初期値なし)
            m_X[gloIdx] = INITVAL_FOR_EST;
#endif
            for (int j = 0; j < M; j++) {
                P[gloIdx+j*M] = P[j+gloIdx*M] = (gloIdx==j)?  m_sigx * m_sigx: 0.0;
            }
        }

        // 観測感度行列の初期化
		double *Hur = zeros(m, M);

		// 一重差位相の観測残差ベクトルの初期化
		double *Vur = zeros(m, 1);
		
		// 衛星別に処理
		for (int i=0;i<(int)vecSat.size();i++)
		{
            // GLONASSのINDEX(衛星No)
            int gloIdx = vecSat[i].gloIdx;

			// 観測残差ベクトル
            Vur[i] = residualsValue(vecDataTime[timeIdx], &vecSat[i], -1, NULL);
		
			// 観測感度行列
            int idx = gloIdx * m + i;
            Hur[idx] = m_nav->lam[vecSat[i].satIdx][m_freq];
		}

		// １重差観測の誤差共分散行列
		double* Rur = covariance(m);
		
		// １重差行列
        double* D = matrix1(m, baseIdx);

		// 観測残差ベクトル(V)
		double* V = zeros(m-1, 1);
		matmul("NN", m-1, 1, m, 1.0, D, Vur, 0.0, V);
		
		// 観測感度行列(H)
        double* wH = zeros(m-1, M);
        matmul("NN", m-1, M, m, 1.0, D, Hur, 0.0, wH);
		
        // カルマンフィルタ関数に入力のはHの転置行列
        double *H = zeros(M, m - 1);
        matrixTrans(wH, H, m - 1, M);

		// 観測誤差共分散行列(R)
		double* wR = zeros(m-1, m);
		matmul("NN", m-1, m, m, 1.0, D, Rur, 0.0, wR);
		double* R = zeros(m-1, m-1);
		matmul("NT", m-1, m-1, m, 1.0, wR, D, 0.0, R);

#ifdef DEBUG_EKF    // debug
{
    if (countForEkfDebug > 0) {
        countForEkfDebug--;
        cout << " time= " << time_str(vecDataTime[timeIdx], 0) << endl;
        cout << " baseSat(gloIdx)=" << vecSat[baseIdx].gloIdx << endl;
        cout << " Hur= " << endl;
        matprint(Hur, m, M, 12, 6);
        cout << " Rur= " << endl;
        matprint(Rur, m, m, 12, 6);
        cout << " D= " << endl;
        matprint(D, m - 1, m, 12, 6);
        cout << " v= " << endl;
        matprint(V, m - 1, 1, 12, 6);
        cout << " wH= " << endl;
        matprint(wH, m - 1, M, 12, 6);
        cout << " H= " << endl;
        matprint(H, M, m - 1, 12, 6);
        cout << " R= " << endl;
        matprint(R, m - 1, m - 1, 12, 6);
    }
}
#endif
        // 拡張カルマンフィルター
        matcpy(workX, m_X, NSATGLO, 1);
        matcpy(workP, P, NSATGLO, NSATGLO);
        int status = filter(workX, workP, H, V, R, M, m-1, NULL);
		
#ifdef DEBUG_EKF    // debug
{
    static gtime_t debugStartTime = { 0 };
    debugStartTime.time = 1354364370;   // 2012/12/01 12:19:30
    double dt = timediff(debugStartTime, vecDataTime[timeIdx]);
    if (dt == 0) {
        countForEkfDebug = 10;
    }
    if (countForEkfDebug > 0) {
        countForEkfDebug--;
        cout << " time= " << time_str(vecDataTime[timeIdx], 0) << endl;
        cout << " P= " << endl;
        matprint(P, NSATGLO, NSATGLO, 12, 6);
        cout << " P(after)= " << endl;
        matprint(workP, NSATGLO, NSATGLO, 12, 6);
    }
}
#endif
        // メモリ解放
		SAFE_FREE(Hur);
		SAFE_FREE(Vur);
		SAFE_FREE(Rur);
		SAFE_FREE(D);
		SAFE_FREE(V);
		SAFE_FREE(H);
		SAFE_FREE(wR);
		SAFE_FREE(R);
		
		if (status < 0)
		{
			// 計算エラー
			continue;
		}

        matcpy(m_X, workX, NSATGLO, 1);
        matcpy(P, workP, NSATGLO, NSATGLO);
        calcFlag = true;
		
        // 推定結果を使って観測残差
        for (int i=0;i<(int)vecSat.size();i++)
        {
            // 観測残差ベクトル(ログに出力するだけ)
            residualsValue(vecDataTime[timeIdx], &vecSat[i], i, &resultWriter);
        }
    }

    // ファイルクローズ
    resultWriter.close();

    // メモリ解放
	SAFE_FREE(P);

	if (!calcFlag)
	{
		return false;
	}

    return true;
}

/**
 * @brief 周波数間バイアスの傾き計算
 * @param	outputFilepath		[IN] : 出力ファイルパス
 * @param   sigNum              [IN] :
 * @param   repeat              [IN] :
 */
bool GlonassEstimater::CalcSlope(string outputFilepath, double sigNum, int repeat, std::vector<std::string> &vecMessage)
{
	m_avrBF[m_freq] = 0.0;

    // 出力ファイルのオープン
    ResultWriter resultWriter;
    resultWriter.open(outputFilepath, false);

    // パス情報出力
    m_numFreq0 = 0;
    m_numFreq1 = 0;
    resultWriter.writeLine(RESULT_PATH_DATA_TAG);
    for (int gloIdx = 0; gloIdx < NSATGLO; gloIdx++) {
        map<int, path_info_t>::iterator ite;
        for (ite = m_mapPathInfo[gloIdx].begin(); ite != m_mapPathInfo[gloIdx].end(); ite++) {
            stringstream oStr;

            // パス情報の出力
            oStr << (gloIdx + MINPRNGLO);                           // GLONASS PRN
            oStr << "," << ite->first;                              // パス番号
            oStr << "," << ite->second.valid;                       // パスの有効/無効
            oStr << "," << time_str(ite->second.startTime, 0);      // パス開始時刻
            oStr << "," << time_str(ite->second.endTime, 0);        // パス終了時刻
            oStr << "," << setprecision(12) << ite->second.bias;    // バイアス値(パス終了時刻)
            resultWriter.writeLine(oStr.str());
        }
    }

    // -----< 傾きデータのループ >-----
    resultWriter.writeLine(RESULT_SLOPE_DATA_TAG);
    bool firstFlag = true;
    gtime_t nextTime;
    vector<double> resultChBias;
    vector<bool> resultChBiasFlag;
    for (int dataIdx = 0; dataIdx < (int)m_vecSlopeData.size(); dataIdx++) {
        vector<sat_bias_t> *vecSatData = &m_vecSlopeData[dataIdx]->vecSatData;

        // [[[ 計算タイミングの判定 ]]]
        // 統計処理間隔をチェック
        gtime_t dataTime = m_vecSlopeData[dataIdx]->dataTime;
        if (!firstFlag) {
            double dt = timediff(nextTime, dataTime);
            if (dt > 0) {
                // 統計処理間隔でなければスキップ
                continue;
            }

        } else {
            // 初回
            firstFlag = false;
        }
        nextTime = dataTime;
        nextTime.time += m_calcInterval;

        // チェック用時刻の設定
#if 0
        gtime_t *checkDataTime = NULL;      // パス終了時刻のバイアス値を使用
#else
        gtime_t *checkDataTime = &dataTime; // データ時刻のバイアス値を使用
#endif

        ostringstream strStream;
        int baseFreqNo=0;
        vector<int> vecFreqZero = searchFreqNo(0, vecSatData, checkDataTime);
        strStream << time_str(dataTime,0) << " 周波数番号\"0\"  vecFreqZero.size()=" << vecFreqZero.size() << "\n";
        //vector<int> vecFreqZero;
        // 基準衛星(周波数番号が"0"～)を検索
        for(int i = 0; i<=FREQ_ABS_MAX;i++){
            vecFreqZero = searchFreqNo(i, vecSatData, checkDataTime);
            if (vecFreqZero.size() != 0) {
                baseFreqNo=i;
                strStream << time_str(dataTime,0) << " break baseFreqNo=" << baseFreqNo << "\n";
                // 基準衛星が見つかった場合はループを抜ける
                break;
            }
            vecFreqZero = searchFreqNo(-i, vecSatData, checkDataTime);
            if (vecFreqZero.size() != 0) {
                baseFreqNo=-i;
                strStream << time_str(dataTime,0) << " break baseFreqNo=" << baseFreqNo << "\n";
                // 基準衛星が見つかった場合はループを抜ける
                break;
            }
        }
        if (vecFreqZero.size() == 0) {
            // 基準衛星がない場合はスキップ
            continue;
        }
        m_numFreq0 += vecFreqZero.size();

        // 周波数番号が基準周波数の"+1"/"-1"のものを検索
        vector<int> vecFreqPlusOne = searchFreqNo(baseFreqNo+1, vecSatData, checkDataTime);
        vector<int> vecFreqMinusOne = searchFreqNo(baseFreqNo-1, vecSatData, checkDataTime);
        if (vecFreqPlusOne.size() == 0
                && vecFreqMinusOne.size() == 0) {
            // 基準衛星に隣合う周波数番号がない場合はスキップ
            continue;
        }
        m_numFreq1 += vecFreqPlusOne.size() + vecFreqMinusOne.size();

        // [[[ 周波数間バイアス計算の準備 ]]]
        // 傾き計算用にデータを準備
        vector<ch_bias_data_t> vecChBiasData;
        for (int i = 0; i < (int)vecSatData->size(); i++) {
            int gloIdx = vecSatData->at(i).gloIdx;

            ch_bias_data_t chBiasData = { 0 };
            chBiasData.gloIdx = gloIdx;
            chBiasData.freqNo = vecSatData->at(i).freqNo;

            // 波長＆周波数を取得
            double lam = m_nav->lam[vecSatData->at(i).satIdx][m_freq];
            if (lam == 0.0) {
                // 0.0はエラー(航法MSGにデータなし)
                chBiasData.noUse = true;
            } else {
                chBiasData.lam = lam;
                chBiasData.freq =  CLIGHT / lam;
            }

            // バイアス値を取得
            if (checkDataTime == NULL) {
                // パス情報からバイアス値を取得
                map<int, path_info_t>::iterator ite = m_mapPathInfo[gloIdx].find(vecSatData->at(i).pathNo);
                if (ite == m_mapPathInfo[gloIdx].end()) {
                    // パス情報なし
                    chBiasData.noUse = true;
                } else {
                    chBiasData.bias = ite->second.bias;
                }
            } else {
                // その時刻のバイアス値を取得
                chBiasData.bias = vecSatData->at(i).bias;
            }

            // データを追加
            vecChBiasData.push_back(chBiasData);
        }

        // 基準衛星は計算対象から外す
        int baseIdx = -1;
        for (int i = 0; i < (int)vecFreqZero.size(); i++) {
            if (baseIdx < 0) {
                // 最初のものを基準衛星とする
                baseIdx = vecFreqZero[i];
            }

            // 計算済みとする
            vecChBiasData[vecFreqZero[i]].calcFlag = true;
        }

        // [[[ 周波数間バイアスの計算 ]]]
        // 周波数番号が基準周波数"+1"
        for (int i = 0; i < (int)vecFreqPlusOne.size(); i++) {
            int idx = vecFreqPlusOne[i];
            vecChBiasData[idx].calcFlag = true;
            vecChBiasData[idx].chBias = CalcBurij(vecChBiasData[idx].bias, vecChBiasData[baseIdx].bias);
        }

        // 周波数番号が基準周波数"-1"
        for (int i = 0; i < (int)vecFreqMinusOne.size(); i++) {
            int idx = vecFreqMinusOne[i];
            vecChBiasData[idx].calcFlag = true;
            vecChBiasData[idx].chBias = CalcBurij(vecChBiasData[idx].bias, vecChBiasData[baseIdx].bias);
        }

        // 周波数番号が±n(ｎは2以上)について、外挿ポイントを算出しつつ求めていく
        for (int n = 2; n <= FREQ_ABS_MAX+abs(baseFreqNo); n++) {
            // 該当の周波数番号があるか検索
            vector<int> vecPlus = searchFreqNo(baseFreqNo+n, vecSatData, checkDataTime);
            vector<int> vecMinas = searchFreqNo(baseFreqNo-n, vecSatData, checkDataTime);
            if (vecFreqPlusOne.size() == 0
                    && vecFreqMinusOne.size() == 0) {
                // 対象がなければスキップ
                continue;
            }

            // 計算済みのものを使って、係数pを求める
            double sum = 0.0;
            int num = 0;
            for (int i = 0; i < (int)vecChBiasData.size(); i++) {
                if (vecChBiasData[i].freqNo == baseFreqNo
                        || !vecChBiasData[i].calcFlag
                        || vecChBiasData[i].noUse) {
                    // 周波数番号"0"、未計算、未使用は除く
                    continue;
                }

                // 傾きを計算
                sum += vecChBiasData[i].chBias / (double)(vecSatData->at(i).freqNo-baseFreqNo);
                num++;
            }
            if (num == 0) {
                // 計算できていない場合はスキップ
                continue;
            }
            double p = sum / num;

            // 外層ポイントを算出
            double An_plus = p * n;
            double An_minus = p * (-n);

            // 周波数番号がn
            for (int i = 0; i < (int)vecPlus.size(); i++) {
                int idx = vecPlus[i];
                if (vecChBiasData[idx].noUse) {
                    continue;
                }
                vecChBiasData[idx].calcFlag = true;
                vecChBiasData[idx].chBias = CalcBurij(vecChBiasData[idx].bias, vecChBiasData[baseIdx].bias, An_plus);
            }

            // 周波数番号が-n
            for (int i = 0; i < (int)vecMinas.size(); i++) {
                int idx = vecMinas[i];
                if (vecChBiasData[idx].noUse) {
                    continue;
                }
                vecChBiasData[idx].calcFlag = true;
                vecChBiasData[idx].chBias = CalcBurij(vecChBiasData[idx].bias, vecChBiasData[baseIdx].bias, An_minus);
            }
        }


        // [[[ 全周波数間バイアス値から傾きを求める ]]]
        for (int i = 0; i < (int)vecSatData->size(); i++) {
            if (vecChBiasData[i].freqNo == baseFreqNo
                    || !vecChBiasData[i].calcFlag
                    || vecChBiasData[i].noUse) {
                // 周波数番号"0"、未計算、未使用は除く
                continue;
            }

            // 周波数間バイアスの単位変換(cycle->m)
            double chBias = vecChBiasData[i].chBias * vecChBiasData[i].lam;

            // 推定データの出力
            vector<double> vecData;
            vecData.push_back(vecChBiasData[baseIdx].gloIdx + MINPRNGLO);   // 基準衛星
            vecData.push_back(vecChBiasData[i].gloIdx + MINPRNGLO);         // 衛星i
            vecData.push_back(vecChBiasData[baseIdx].freq);                 // 周波数(基準)[Hz]
            vecData.push_back(vecChBiasData[i].freq);                       // 周波数(衛星i)[Hz]
            vecData.push_back(vecChBiasData[baseIdx].bias);                 // 衛星バイアス(基準)[cycle]
            vecData.push_back(vecChBiasData[i].bias);                       // 衛星バイアス(衛星i)[cycle]
            vecData.push_back(chBias);                                      // 周波数間バイアス[m]
            vecData.push_back(vecChBiasData[baseIdx].freqNo);               // 周波数番号(基準)
            vecData.push_back(vecChBiasData[i].freqNo);                     // 周波数番号(衛星i)
            vecData.push_back(chBias / (vecChBiasData[i].freqNo-baseFreqNo));
            resultWriter.writeLine(dataTime, vecData, 13);
            //resultWriter.writeLine(dataTime, vecData, 12);


            // 傾きを計算[m/No]
            resultChBias.push_back(chBias / (vecChBiasData[i].freqNo-baseFreqNo));
            resultChBiasFlag.push_back(true);
        }
    }
	
	// 全BFの平均
    if (resultChBias.size() == 0)
	{
		return false;
	}

    // 平均の傾きを取得
    double sumChBias = 0.0;
    int count = 0;
    // 平均計算
    for (int i = 0; i < (int)resultChBias.size(); i++) {
        if(resultChBiasFlag[i]){
            sumChBias +=  resultChBias[i];
            count++;
        }
    }
    for(int j = 0; j < repeat; j++){
        double ave = sumChBias / (double)count;
        // 標準偏差計算
        double sumSig=0;
        count = 0;
        for (int i = 0; i < (int)resultChBias.size(); i++){
            if(resultChBiasFlag[i]){
                sumSig += (resultChBias[i] - ave) * (resultChBias[i] - ave);
                count++;
            }
        }
        double sig = sqrt(sumSig / (double)count);
        stringstream ss;
        ss << "mean : " << ave << " , sigma : " << sig << ", num : " << count << "  [freq=" << m_freq+1 << "]" ;
        vecMessage.push_back(ss.str());
        count = 0;
        sumChBias = 0.0;
        int reject = 0;
        // 棄却判定および平均計算
        for(int i = 0; i < (int)resultChBias.size(); i++){
            if(resultChBiasFlag[i]){
                if(abs(resultChBias[i]-ave) >= sigNum * sig){
                    resultChBiasFlag[i] = false;
                    reject++;
                }
                else{
                    sumChBias += resultChBias[i];
                    count++;
                }
            }
        }
        if(reject == 0){
            break;
        }
    }
    // 棄却されなかった個数を確認
    if(count == 0){
        return false;
    }
    m_avrBF[m_freq] = sumChBias / (double)count;
    stringstream ss;
    ss << "mean : " << m_avrBF[m_freq] << ", num : " << count << "  [freq=" << m_freq+1 << "]" ;
    vecMessage.push_back(ss.str());

    // 推定結果の出力
    resultWriter.writeLine(RESULT_SLOPE_VAL_TAG);
    vector<double> vecResult;
    vecResult.push_back(m_avrBF[m_freq]);
    resultWriter.writeLine(vecResult);

    // ファイルクローズ
    resultWriter.close();

    return true;
}
/**
 * @brief 結果の出力
 *
 * @param	resultFile			[IN] : 結果出力ファイル名称
 *
 * @return	 処理結果
 * @retval	true	: 正常終了
 * @retval	false	: 異常終了
 *
 *	注意事項：
 *		1) L1およびL2の計算が終了後呼び出すこと
 *		2) L1とL2の計算は同一条件とすること
 */
bool GlonassEstimater::outputResult(std::string resultFile)
{
	// 計算結果の初期化
	m_result.clear();

	// 受信機名称の取得
	sta_t* sta1 = m_referObsMng->getStationParam();
	if (!sta1)
	{
		return false;
	}
	char name1[MAXANT];
	strcpy(name1, sta1->rectype);
	
	sta_t* sta2 = m_targetObsMng->getStationParam();
	if (!sta2)
	{
		return false;
	}
	char name2[MAXANT];
	strcpy(name2, sta2->rectype);
	
	RESULT_T result;
	result.Clear();
	strcpy(result.receiverName1, name1);
	strcpy(result.receiverName2, name2);

    // 傾きを変換(m/No -> m/MHz)
    result.bias[0] = m_avrBF[0] / (DFRQ1_GLO * 1e-6);
    result.bias[1] = m_avrBF[1] / (DFRQ2_GLO * 1e-6);
    //result.bias[1] = m_avrBF[1] / (DFRQ1_GLO * 1e-6);

	m_result.push_back(result);
	
	// ファイルの存在確認
	FILE* fp=fopen(resultFile.c_str(), "r");
	bool fileExist = false;
	if (fp)
	{
		fileExist = true;
		fclose(fp);
	}
	
	// 結果出力
	FormatWriter fw;
	fw.open(resultFile, true);
	if (!fw.isOpen())
	{
		return false;
	}

	// 新規時ヘッダ出力
	char buff[48];
	if (!fileExist)
	{
		// ヘッダ部出力
		fw.write("GLONASS IFB TABLE");
		fw.writeSpace(45);
		time_t now = time(NULL);
		struct tm *pnow = localtime(&now);
		sprintf(buff, "%d/%02d/%02d", pnow->tm_year+1900, pnow->tm_mon+1, pnow->tm_mday);
		fw.write(buff);
		fw.flush();
		
		// line
		for (int i=0;i<72;i++)
		{
			fw.write("-");
		}
		fw.flush();

		// 空白行
		fw.flush();
		
		// 項目名
		fw.write("RECEIVER TYPE 1");
		fw.writeSpace(7);
		fw.write("RECEIVER TYPE 2");
		fw.writeSpace(6);
		fw.write("FREQ L* BIAS(m/MHz)");
		fw.writeSpace(3);
		fw.write("FLAG");
		fw.flush();
		
		// **
		fw.write("********************");
		fw.writeSpace(2);
		fw.write("********************");
		fw.writeSpace(7);
		fw.write("*");
        fw.writeSpace(6);
		fw.write("********");
		fw.writeSpace(5);
		fw.write("*");
		fw.flush();
		
		// 空白行
		fw.flush();
	}
	
	// データ部の出力
	for (int i=0;i<(int)m_result.size();i++)
	{
		// L1
		fw.write(20, m_result[i].receiverName1);
		fw.writeSpace(2);
		fw.write(20, m_result[i].receiverName2);
		fw.writeSpace(7);
		fw.write(1, "1");
        fw.writeSpace(6);
        sprintf(buff, "%8.4f", m_result[i].bias[0]);  // [m/MHz]
		fw.write(buff);
		fw.flush();

		// L2
		fw.writeSpace(49);
		fw.write(1, "2");
        fw.writeSpace(6);
        sprintf(buff, "%8.4f", m_result[i].bias[1]);  // [m/MHz]
		fw.write(buff);
		fw.flush();
		
		// 空白行
		fw.flush();
	}
	
	fw.close();
	
	return true;
}

/**
 * @brief 一重差位相の観測残差ベクトル値の計算
 *
 */
double* GlonassEstimater::apriori()
{
    double* P = zeros(NSATGLO, NSATGLO);
    double val = m_sigx * m_sigx;
    
    for (int i=0;i<NSATGLO;i++)
    {
        int idx = i * NSATGLO + i;
        P[idx] = val;
    }
	
	return P;
}
/**
 * @brief １重差観測の誤差共分散行列の計算
 *
 * @param	numSat		[IN] : 観測衛星数
 */
double* GlonassEstimater::covariance(int numSat)
{
	double* Rur = zeros(numSat, numSat);
	double val = 2.0 * m_sigp * m_sigp;
    
    for (int i=0;i<numSat;i++)
    {
        int idx = i * numSat + i;
        Rur[idx] = val;
    }
	
	return Rur;
}

/** 状態更新
 * @brief AntennaEstimater::updateState
 * @param rtk [IN/OUT]結果
 */
void GlonassEstimater::updateState(rtk_t *rtk)
{
    // 共通の衛星を取得
    obsd_t referenceObsData[MAXOBS];
    obsd_t targetObsData[MAXOBS];
    int num = AntennaEstimater::getSameSat(m_referObs, m_targetObs, SYS_GLO, referenceObsData, targetObsData);

    // [[[ 初期設定 ]]]
    // 航法データ取得
    nav_t *nav = m_navMng->getAll();

    // サイクルスリップ状態のリセット(全衛星)
    for (int i = 0; i < MAXSAT; i++) {
        rtk->ssat[i].slip[m_freq] = 0;
    }

    // [[[ サイクルスリップ検出＆観測データ抜けのカウント ］］］
    // LLIによるサイクルスリップの検出
    ObservationManager::detectSlipByLLI(rtk, referenceObsData, num, m_freq);
    ObservationManager::detectSlipByLLI(rtk, targetObsData, num, m_freq);

    // 位相の飛びによるサイクルスリップの検出
    ObservationManager::detectSlipyGF2(rtk, referenceObsData, targetObsData, num, nav, m_freq);
}

/**
 * @brief 一重差位相の観測残差ベクトル値の計算
 *
 * @param	gtime			[IN] : 時刻
 * @param	csat            [IN] : 衛星情報
 * @param	nv		        [IN] : 観測残差ベクトルのINDEX
 * @param	resultWriter    [IN] : 出力ファイルクラス（NULL：出力しない)
 * @return 観測残差ベクトル
 */
double GlonassEstimater::residualsValue(gtime_t gtime, calc_sat_t *csat, int nv, ResultWriter *resultWriter)
{
    int referIdx = csat->referIdx;
    int targetIdx = csat->targetIdx;

	// ruri
    double ruri = csat->distToTarget - csat->distToRef;
	
	// suri
    double puri = m_targetObs[targetIdx].L[m_freq] - m_referObs[referIdx].L[m_freq];
	
	// ターゲット受信機と基準受信機の一重差位相の観測残差ベクトル
    int gloIdx = csat->gloIdx;
    int satIdx = csat->satIdx;
    double value = m_nav->lam[satIdx][m_freq] * puri - (ruri + m_nav->lam[satIdx][m_freq] * m_X[gloIdx]);
#if 1   // debug
    if (nv >= 0) {
        ostringstream oStr;
        oStr << time_str(gtime, 0)
                << setprecision(12)
                << " vur[" << nv << "]=" << value
                << " lam=" << m_nav->lam[satIdx][m_freq]
                << " ruri=" << ruri
                << "(tsDist=" << csat->distToTarget
                << " rsDist=" << csat->distToRef
                << ") puri=" << puri
                << "(Lu=" << m_targetObs[targetIdx].L[m_freq]
                << " Lr=" << m_referObs[referIdx].L[m_freq]
                << ") X=" << m_X[gloIdx]
                << " gloIdx=" << gloIdx
                << " m_freq=" << m_freq;
        outputLog(TRACE_LEVEL_INFO, oStr.str(), "GlonassEstimater::residualsValue");
    }
#endif
    // 推定データの出力
    if (resultWriter != NULL) {
        vector<double> vecData;
        vecData.push_back(satIdx);                          // 衛星のINDEX
        vecData.push_back(gloIdx);                          // GLONASSのINDEX
        vecData.push_back(m_nav->lam[satIdx][m_freq]);      // 波長
        vecData.push_back(csat->distToRef);                 // 基準局－衛星間距離
        vecData.push_back(csat->distToTarget);              // 測定局－衛星間距離
        vecData.push_back(m_referObs[referIdx].L[m_freq]);      // 搬送波位相(基準局)
        vecData.push_back(m_targetObs[targetIdx].L[m_freq]);    // 搬送波位相(測定局)
        vecData.push_back(m_X[gloIdx]);                     // 推定バイアス
        vecData.push_back(m_pathNo[gloIdx]);                // パス番号
        vecData.push_back(value);                           // 残差
        vecData.push_back(m_freqNo[gloIdx]);                // 周波数番号
        vecData.push_back(csat->slip);                      // サイクルスリップ(1:発生)
        vecData.push_back(csat->gf);                        // サイクルスリップ判定用
        vecData.push_back(csat->reset);                     // リセット(1:実施)
        resultWriter->writeLine(gtime, vecData, 12);

    }
	return value;
}

/** 衛星バイアスファイルの読込
 * @brief GlonassEstimater::readBiasFile
 * @param inputFilepath 入力ファイルパス
 */
void GlonassEstimater::readBiasFile(string inputFilepath) {

    // ファイルオープン
    ifstream ifs;
    ifs.open(inputFilepath.c_str(), ios::in);

    // パス情報初期化
    int nowPathNo[NSATGLO];
    path_info_t pathInfoZero = { 0 };
    path_info_t pathInfo[NSATGLO];
    for (int i = 0; i < NSATGLO; i++) {
        nowPathNo[i] = -1;
        pathInfo[i] = pathInfoZero;
    }

    // データ読込
    string buf;
    slope_data_t* slopeData = NULL;
    gtime_t lastDataTime;
    gtime_t lastDataTimeSat[NSATGLO];
    double lastBias[NSATGLO];
    while(ifs && getline(ifs, buf)) {
        // 対象データ読込
        vector<string> vecData = splitString(buf, ',');

        // GLONASS INDEXチェック
        int gloIdx = atoi(vecData[2].c_str());
        if (gloIdx < 0 || gloIdx >=NSATGLO) {
            // 不正なINDEXの場合は無視
            continue;
        }

        // 時刻取得
        gtime_t dataTime = stringToGtime(vecData[0]);
        if (slopeData == NULL) {
            // 初回
            slopeData = new slope_data_t;
            lastDataTime = dataTime;
            slopeData->dataTime = dataTime;

        } else {
            double dt = timediff(dataTime, lastDataTime);
            if (dt > 0) {
                // 時刻が変わったら、データを登録する
                m_vecSlopeData.push_back(slopeData);
                //delete slopeData;
                // 次の時刻のデータ準備
                slopeData = new slope_data_t;
                lastDataTime = dataTime;
                slopeData->dataTime = dataTime;
            }
        }

        // パス番号チェック
        int pathNo = atoi(vecData[9].c_str());
        if (nowPathNo[gloIdx] != pathNo) {
            if (nowPathNo[gloIdx] >= 0) {
                // 初回以外は終了時刻(前回の時刻)を設定して、マップに追加
                addPathInfo(gloIdx, nowPathNo[gloIdx], pathInfo[gloIdx], lastDataTimeSat[gloIdx], lastBias[gloIdx]);
            }

            // パス開始時刻を保管
            pathInfo[gloIdx].startTime = dataTime;
            nowPathNo[gloIdx] = pathNo;
        }


        // 衛星バイアス情報取得
        /* ref) GlonassEstimater::residualsValue()
         * 0:時刻
         * 1:衛星のINDEX
         * 2:GLONASSのINDEX
         * 3:波長
         * 4:基準局－衛星間距離
         * 5:測定局－衛星間距離
         * 6:搬送波位相(基準局)
         * 7:搬送波位相(測定局)
         * 8:推定バイアス
         * 9:パス番号
         *  etc
         */
        sat_bias_t satBias = { 0 };
        satBias.satIdx = atoi(vecData[1].c_str());
        satBias.gloIdx = gloIdx;
        satBias.bias = atof(vecData[8].c_str());
        satBias.pathNo = pathNo;
        satBias.freqNo = m_freqNo[gloIdx];

        // データ追加
        slopeData->vecSatData.push_back(satBias);
        lastDataTimeSat[gloIdx] = dataTime;
        lastBias[gloIdx] = satBias.bias;
    }

    // 最後のデータ登録
    if (slopeData != NULL) {
        m_vecSlopeData.push_back(slopeData);
    }
    for (int i = 0; i < NSATGLO; i++) {
        if (nowPathNo[i] >= 0) {
            addPathInfo(i, nowPathNo[i], pathInfo[i], lastDataTimeSat[i], lastBias[i]);
        }
    }

    // ファイルクローズ
    ifs.close();
}

/** パス情報をマップに追加
 * @brief GlonassEstimater::addPathInfo
 * @param gloIdx GLONASS INDEX
 * @param nowPathNo 現在パス番号
 * @param pathInfo パス情報(開始時刻は入っていること)
 * @param endTime パス終了時刻
 * @param bias バイアス値
 */
void GlonassEstimater::addPathInfo(int gloIdx, int nowPathNo, path_info_t &pathInfo, gtime_t endTime, double bias) {
    // 終了時刻(前回の時刻)＆バイアス値を設定
    pathInfo.endTime = endTime;
    pathInfo.bias = bias;

    // 有効/無効チェック
    double dt = timediff(pathInfo.endTime, pathInfo.startTime);
    pathInfo.valid = (dt >= m_pathMinRange);

    // マップに追加
    m_mapPathInfo[gloIdx].insert(pair<int, path_info_t>(nowPathNo, pathInfo));
}

/** 指定周波数番号を検索
 * @brief GlonassEstimater::searchFreqNo
 * @param targetNo 指定周波数番号
 * @param vecSlopeData 衛星バイアス情報
 * @param dataTime データ時刻(NULL:有効パスを採用、NULL以外：パス開始時刻からのチェック)
 * @return 指定周波数番号のあるデータINDEX(複数)
 */
vector<int> GlonassEstimater::searchFreqNo(int targetNo, vector<sat_bias_t> *vecSatBias, gtime_t *dataTime) {
    vector<int> vec;

    // 指定された周波数番号を探す
    for(int i = 0; i < (int)vecSatBias->size(); i++) {
        // 指定周波数番号のチェック
        if (vecSatBias->at(i).freqNo == targetNo) {
            int gloIdx = vecSatBias->at(i).gloIdx;

            // パス情報を取得
            map<int, path_info_t>::iterator ite = m_mapPathInfo[gloIdx].find(vecSatBias->at(i).pathNo);
            if (ite != m_mapPathInfo[gloIdx].end()) {
                if (dataTime != NULL) {
                    double dt = timediff(*dataTime, ite->second.startTime);
                    if (dt >= m_pathMinRange) {
                        // パス開始時刻から最小時間が経過しているなら追加
                        vec.push_back(i);
                    }

                } else {
                    if (ite->second.valid) {
                        // 有効なパスの場合は追加
                        vec.push_back(i);
                    }
                }
            }
        }
    }
    return vec;
}


static double round_d(double x) 
{
	if (x > 0.0) 
	{
		return floor(x + 0.5);
	} 
	else 
	{
		return -1.0 * floor(fabs(x) + 0.5);
	}
}

/**
 * @brief Burijの計算
 *
 * @param	Buri		[IN] : 衛星1のバイアス
 * @param	Buri		[IN] : 衛星2のバイアス
 * @param	An          [IN] : 外捜ポイントのバイアス
 */
double GlonassEstimater::CalcBurij(double Buri, double Burj, double An)
{
    double Nurij = round_d(Buri - Burj - An);
	
	return Buri - Burj - Nurij;
}

/**
 * @brief 周波数番号（-7～6）の周波数を昇順に取得
 *
 * @param	freqSats		[OUT]  : 昇順に並んだ周波数と周波数に対応する衛星列
 *
 * @return	  取得した周波数数
 */
int GlonassEstimater::GetFrequency(FREQ_SATS_T* freqSats)
{
	int numFreq = 0;
	
	for (int i=0;i<FREQ_NUM;i++)
	{
		freqSats[i].frequency.frequencyNo = 0;
		freqSats[i].frequency.frequency = 0;
		freqSats[i].sats.clear();
	}

	// 周波数および衛星の取り出し
	for (int i=FREQ_NO_MIN;i<=FREQ_NO_MAX;i++)
	{
		bool exist = false;
		for (int j=0;j<m_nav->ng;j++)
		{
			if (m_nav->geph[j].frq == i)
			{
				int prn;
                if (satsys(m_nav->geph[j].sat, &prn) != SYS_GLO)
				{
                    outputLog(TRACE_LEVEL_WARNING, "not GLONASS. sat=",  (int)m_nav->geph[j].sat, "GlonassEstimater::GetFrequency");
					continue;
				}

                int satNo =  prn - MINPRNGLO;
				if (!exist)
				{
                    int satIdx = m_nav->geph[j].sat - 1;
                    double lam = m_nav->lam[satIdx][m_freq];
                    if (lam == 0.0) {
                        // 0.0はエラー(航法MSGにデータなし)
                        continue;
                    }

					exist = true;
					freqSats[numFreq].frequency.frequencyNo = i;
                    freqSats[numFreq].frequency.frequency = CLIGHT / lam;
				}
				bool satExist = false;
				for (int k=0;k<(int)freqSats[numFreq].sats.size();k++)
				{
					if (freqSats[numFreq].sats[k] == satNo)
					{
						satExist = true;
						break;
					}
				}
				if (!satExist)
				{
					freqSats[numFreq].sats.push_back(satNo);
				}
			}
		}

		numFreq++;
	}
	
	// 周波数を昇順にソート
	for (int i=0;i<numFreq-1;i++)
	{
		for (int j=i+1;j<numFreq;j++)
		{
			if (freqSats[i].frequency.frequency > freqSats[j].frequency.frequency)
			{
				FREQ_SATS_T swap;
				swap = freqSats[i];
				freqSats[i] = freqSats[j];
				freqSats[j] = swap;
			}
		}
	}
	
	return numFreq;
}


