﻿/*------------------------------------------------------------------------------
* GlonassEstimater.h : GLONASS受信機チャンネル間バイアス推定クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#ifndef GLONASSESTIMATER_H
#define GLONASSESTIMATER_H

#include <vector>
#include <sstream>
#include "AntLibBase.h"
#include "ObservationManager.h"
#include "NavigationManager.h"
#include "ResultWriter.h"
#include "AntennaEstimater.h"

#define RESULT_PATH_DATA_TAG   "<path_data>"
#define RESULT_SLOPE_DATA_TAG   "<slope_data>"
#define RESULT_SLOPE_VAL_TAG    "<slope>"

/** 衛星情報
 * @brief The calc_sat_t struct
 */
struct calc_sat_t
{
    int     satIdx;             ///< 衛星INDEX
    int     gloIdx;             ///< GLONASS INDEX
    double  satPosRef[6];       ///< 衛星位置(基準局)
    double  satPosTarget[6];    ///< 衛星位置(測定局)
    double  distToRef;          ///< 衛星－基準局間距離
    double  distToTarget;       ///< 衛星－測定局間距離
    double  azel[2];            ///< 方位角/仰角[rad]
    int     referIdx;           ///< 当該エポックの基準局観測データINDEX
    int     targetIdx;          ///< 当該エポックの測定局観測データINDEX
    int     slip;               ///< サイクルスリップ(1：発生)
    double  gf;                 ///< サイクルスリップ判定用
    int     reset;              ///< リセット(1:実行)[サイクルスリップ時、パス切断時]
};

/** 衛星バイアス情報
 * @brief The slope_sat_data_t struct
 */
struct sat_bias_t
{
    int     satIdx;     ///< 衛星のINDEX
    int     gloIdx;     ///< GLONASS INDEX
    int     pathNo;     ///< パス番号
    int     freqNo;     ///< 周波数番号
    double  bias;       ///< バイアス値[cycle]
};

/** 傾きデータ情報(時刻別衛星バイアス情報)
 * @brief The slope_data_t struct
 */
struct slope_data_t
{
    gtime_t dataTime;                   ///< データ時刻
    std::vector<sat_bias_t> vecSatData; ///< 衛星別データ
};

/** 周波数間バイアス情報
 * @brief The calc_slope_t struct
 */
struct ch_bias_data_t {
    int     gloIdx;     ///< GLONASS INDEX
    int     freqNo;     ///< 周波数番号
    double  freq;       ///< 周波数[Hz]
    double  lam;        ///< 波長[m]
    bool    calcFlag;   ///< 計算フラグ(true：済)
    bool    noUse;      ///< 使用不可フラグ(true：不可)
    double  bias;       ///< バイアス値[cycle]
    double  chBias;     ///< 周波数間バイアス値[cycle]
};

/** パス情報
 * @brief The path_info_t struct
 */
struct path_info_t {
    gtime_t startTime;  ///< パス開始時刻
    gtime_t endTime;    ///< パス終了時刻
    bool    valid;      ///< パス有効/無効(true：有効)
    double  bias;       ///< バイアス値[cycle](パス終了時刻のもの)
};

/** GLONASS受信機チャンネル間バイアス推定
 * @brief The GlonassEstimater class
 */
class GlonassEstimater : public AntLibBase
{
public:
	GlonassEstimater(void);
    virtual ~GlonassEstimater(void);
	
    void initialize(NavigationManager* navMng);
	void reset();
    void estimateBias(int freq, prcopt_t opt,
						ObservationManager *referenceObsMng, double* referencePosition, 
						ObservationManager *targetObsMng, double* targetPosition, 
                        double sigx, double sigp,
                        int calcInterval, int pathMinRange,
                        std::string outputFilepathForBias, std::string outputFilepath,
                        double sigNum, int repaet, std::vector<std::string> &vecMessage);
	bool outputResult(std::string resultFile);

private:
	/**
	 * @brief 周波数番号データ
	 */
	struct FREQ_NO_DATA_T
	{
		int							frequencyNo;	///< 周波数番号
		double						frequency;		///< 周波数
	};
	/**
	 * @brief 周波数および対応衛星列データ
	 */
	struct FREQ_SATS_T
	{
		FREQ_NO_DATA_T				frequency;		///< 周波数番号
		std::vector<unsigned char>	sats;			///< 衛星番号列
	};
	/**
	 * @brief 推定結果データ
	 */
	struct RESULT_T
	{
		char						receiverName1[MAXANT];	///< 受信機名1
		char						receiverName2[MAXANT];	///< 受信機名2
		double						bias[2];				///< 周波数間バイアスの傾き

		void Clear()
		{
			memset(receiverName1, 0, sizeof(receiverName1));
			memset(receiverName2, 0, sizeof(receiverName2));
			bias[0] = 0;
			bias[1] = 0;
		}
	};

private:
    bool CalcBias(prcopt_t *opt, std::string outputFilepath);
    bool CalcSlope(std::string outputFilepath, double sigNum, int repeat, std::vector<std::string> &vecMessage);
    void updateState(rtk_t *rtk);
    double residualsValue(gtime_t gtime, calc_sat_t *csat, int nv, ResultWriter *resultWriter);
    void readBiasFile(std::string inputFilepath);
    void addPathInfo(int gloIdx, int nowPathNo, path_info_t &pathInfo, gtime_t endTime, double bias);
    std::vector<int> searchFreqNo(int targetNo, std::vector<sat_bias_t> *vecSlopeData, gtime_t *dataTime);
	double* apriori();
	double* covariance(int numSat);
    double CalcBurij(double Buri, double Burj, double An = 0);
	int GetFrequency(FREQ_SATS_T* freqSats);


private:
	typedef std::vector<obsd_t> OBS_DATA;
	typedef std::vector<RESULT_T> RESULT_DATA;
	
	bool				m_init;			///< 初期化フラグ
    NavigationManager*	m_navMng;		///< 航法データ管理クラス
    ObservationManager*	m_referObsMng;	///< 基準観測データ管理クラス
    ObservationManager*	m_targetObsMng;	///< ターゲット観測データ管理クラス
	int					m_freq;			///< 使用周波数(0：L1、1：L2)
	double				m_referPos[3];	///< 基準アンテナ位相中心
	double				m_targetPos[3];	///< ターゲットアンテナ位相中心
	double				m_sigx;			///<
	double				m_sigp;			///< 観測誤差の標準偏差
    OBS_DATA			m_referObs;		///< 基準観測データ
	OBS_DATA			m_targetObs;	///< ターゲット観測データ
	nav_t*				m_nav;			///< 航法データ
    int                 m_calcInterval; ///< 衛星バイアス推定をする間隔[sec]
    int                 m_pathMinRange; ///< パスを有効とする最小時間[sec]
    int                 m_pathNo[NSATGLO];      ///< パス番号
    int                 m_freqNo[NSATGLO];      ///< 周波数番号
    std::vector<slope_data_t*> m_vecSlopeData;  ///< 傾き計算用データ
    std::map<int, path_info_t> m_mapPathInfo[NSATGLO]; ///< 衛星別パス情報
	double*				m_X;			///< 搬送波移送バイアス
	double				m_avrBF[2];		///< 周波数間バイアスの傾きの平均[m/MHz]
										///< [0]:L1 [1]:L2
	RESULT_DATA			m_result;		///< 計算結果
    int                 m_numFreq0;       ///< 周波数番号0のデータ数
    int                 m_numFreq1;       ///< 周波数番号1or-1のデータ数

	
	
	static const int	FREQ_NO_MIN	= -7;								///< 周波数最少番号
	static const int	FREQ_NO_MAX	= 6;								///< 周波数最大番号
    static const int	FREQ_ABS_MAX = 7;								///< 周波数最大番号(絶対値)
	static const int	FREQ_NUM	= FREQ_NO_MAX - FREQ_NO_MIN + 1;	///< 周波数数
};


#endif
