﻿/*------------------------------------------------------------------------------
* NavErrEstimater.ccp : 観測誤差モデル推定クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "NavErrEstimater.h"

#define SQR(x)          ((x)*(x))
#define SAT_INDEX(s)    (s - 1)
#define INITVAL_FOR_EST 1e-10   // 推定パラメータの初期値
#define RMS_LIMIT_FOR_MP 10.0   // 観測残差の閾値(観測誤差モデル推定用)
#define CMP_ERROR_LIMIT 2       // コードマルチパスの異常値除去閾値[m]
#define PMP_ERROR_LIMIT 0.1     // 搬送波マルチパスの異常値除去閾値[m]


// debug用
//#define DEBUG_EKF                   // カルマンフィルタの入力行列出力
static int countForEkfDebug = 1;    // デバッグ用カウンタ

using namespace std;

/** コンストラクタ
 * @brief NavErrEstimater::NavErrEstimater
 */
NavErrEstimater::NavErrEstimater()
{
    receiverNo = RECEIVER_NO_TARGET;
}

/** デストラクタ
 * @brief NavErrEstimater::~NavErrEstimater
 */
NavErrEstimater::~NavErrEstimater()
{
}

/** 状態リセット(衛星システム別)
 * @brief NavErrEstimater::clearResult
 */
void NavErrEstimater::clearResult() {
    vecEstNavError.clear();
}

/** 状態リセット
 * @brief NavErrEstimater::reset
 */
void NavErrEstimater::reset() {
    AntennaBase::reset();

    // 推定状況のリセット
    estimatedErrorModelByCmp = false;
    estimatedErrorModelByPmp = false;
}

/** 搬送波位相マルチパス＋受信機雑音データ読込
 * @brief NavErrEstimater::readPhaseErrorFile
 * @param inputFilepath 入力ファイルパス
 */
void NavErrEstimater::readPhaseErrorFile(string inputFilepath) {

    // 衛星別観測データクリア
    for (int i = 0; i < MAXSAT; i++) {
        targetSatObsData[i].clear();
    }

    // ファイルオープン
    ifstream ifs;
    ifs.open(inputFilepath.c_str(), ios::in);

    // データ読込
    string buf;
    while(ifs && getline(ifs, buf)) {
        // 対象データ読込
        estPhaseErrorData_t data = SatelliteObsData::parsePhaseErrorData(buf);
        if (data.stat != 1) {
            // FIX解以外は使用しない
            continue;
        }
        if (fabs(data.phaseError) > PMP_ERROR_LIMIT) {
            // 大きすぎる値は使用しない
            continue;
        }

        // 衛星別観測データへセット
        int idx = data.sat - 1;
        if (idx < 0 || idx >= MAXSAT) {
            continue;
        }
        targetSatObsData[idx].setPhaseErrorData(data);
    }

    // ファイルクローズ
    ifs.close();
}

/** マルチパスを推定
 * @brief NavErrEstimater::estimateCodeMultipass
 * @param referenceObsMng 基準局の観測データ
 * @param opt RTKLIBのオプション
 * @param rangeForAmp 単純移動平均の対象データ数
 * @param minEl 推定対象衛星仰角下限値
 * @param maxEl 推定対象衛星仰角上限値
 * @param needNotPmp 搬送波マルチパス算出不要フラグ(true：不要)
 * @param outputDataDir 値確認用データファイル出力先
 */
void NavErrEstimater::estimateMultipass(ObservationManager *referenceObsMng, prcopt_t opt, int rangeForAMP,
                                         double minEl, double maxEl, bool needNotPmp, string outputDataDir) {
    outputLog(TRACE_LEVEL_INFO, "rangeForAMP=", rangeForAMP, "NavErrEstimater::estimateMultipass");
    outputLog(TRACE_LEVEL_INFO, "thresslip=", opt.thresslip, "NavErrEstimater::estimateMultipass");
    outputLog(TRACE_LEVEL_INFO, "minEl=", minEl, "NavErrEstimater::estimateMultipass");
    outputLog(TRACE_LEVEL_INFO, "maxEl=", maxEl, "NavErrEstimater::estimateMultipass");
    outputVector(TRACE_LEVEL_INFO, "position=", position, 3, 12,"NavErrEstimater::estimateMultipass");

    // L1の時のみ、確認用データを出力
    SatelliteObsData referenceSatObsData[MAXSAT];
    for (int i = 0; i < MAXSAT; i++) {
        referenceSatObsData[i].outputDataFile = true;
        referenceSatObsData[i].outputDataFileDir = outputDataDir;
        targetSatObsData[i].outputDataFile = true;
        targetSatObsData[i].outputDataFileDir = outputDataDir;
    }

    // 観測データを衛星別に振り分け(基準局)
    obs_t *referenceAllObsData = referenceObsMng->getAll();
    for (int i = 0; i < referenceAllObsData->n; i++) {
        if (!referenceObsMng->containTime(referenceAllObsData->data[i].time)) {
            // 無効データは使用しない(アンテナ位相中心推定時のFIXではないデータ)
            continue;
        }

        // 使用するシステムによる衛星の取捨
        int sys = satsys(referenceAllObsData->data[i].sat, NULL);
        if (!(sys & useSys)) {
            continue;
        }

        int satIdx = SAT_INDEX(referenceAllObsData->data[i].sat);
        referenceSatObsData[satIdx].setObsData(referenceAllObsData->data[i]);
    }

    // 観測データを衛星別に振り分け(測定局)
    obs_t *targetAllObsData = observationManager->getAll();
    for (int i = 0; i < targetAllObsData->n; i++) {
        if (!observationManager->containTime(targetAllObsData->data[i].time)) {
            // 無効データは使用しない
            continue;
        }

        // 使用するシステムによる衛星の取捨
        int sys = satsys(targetAllObsData->data[i].sat, NULL);
        if (!(sys & useSys)) {
            continue;
        }

        int satIdx = SAT_INDEX(targetAllObsData->data[i].sat);
        targetSatObsData[satIdx].setObsData(targetAllObsData->data[i]);
    }

#if 0   // debug
    // ログを出力する衛星を一つ指定
    for (int i = 0; i < MAXSAT; i++) {
        if (targetSatObsData[i].hasData()) {
            outputLog("set logStream at satIndex=", i, "NavErrEstimater::estimateMultipass");
            referenceSatObsData[i].setLogStream(outputStreamForLog);
            targetSatObsData[i].setLogStream(outputStreamForLog);
            break;
        }
    }
#endif
#if 1   // debug
    // アーク切れ目の情報を出力
    for (int i = 0; i < MAXSAT; i++) {
        if (referenceSatObsData[i].hasArcBreak()) {
            outputLog(TRACE_LEVEL_DEBUG_DETAIL, "satIndex(ref)=", i, "NavErrEstimater::estimateMultipass");
            referenceSatObsData[i].outputArcBreak();
        }
    }
    for (int i = 0; i < MAXSAT; i++) {
        if (targetSatObsData[i].hasArcBreak()) {
            outputLog(TRACE_LEVEL_DEBUG_DETAIL, "satIndex(target)=", i, "NavErrEstimater::estimateMultipass");
            targetSatObsData[i].outputArcBreak();
        }
    }
#endif
    // 基準局の電離層遅延量変動を算出
    calculateIono = false;
    for (int i = 0; i < MAXSAT; i++) {
        calculateIono |= referenceSatObsData[i].calculateIonosphericDelayVariations(navigationManager->getAll(), opt);
    }

    // 衛星別にマルチパスを算出
    mapCodeMp.clear();
    mapPhaseMp.clear();
    mapCodeMpForEl.clear();
    mapPhaseMpForEl.clear();
    for (int i = 0; i < MAXSAT; i++) {
        targetSatObsData[i].calculateMultipath(useFreq, navigationManager->getAll(), opt,
                                               position, &referenceSatObsData[i], rangeForAMP,
                                               CMP_ERROR_LIMIT, needNotPmp);

        // 結果を取得(コードマルチパス)
        vector<MultipathData> *vecCmp = targetSatObsData[i].getVecCodeMultipath();
        for (int j = 0; j < (int)vecCmp->size(); j++) {
            // 衛星毎のデータを時刻別/仰角[deg]別にセット
            mapCodeMp.insert(pair<DataTime, MultipathData>((*vecCmp)[j].time, (*vecCmp)[j]));

            double el = (*vecCmp)[j].azel[1] * R2D; // [deg]
            if (el >= minEl && el <= maxEl) {

                // 観測誤差モデル推定に使うデータをセット
                mapCodeMpForEl.insert(pair<Elevation, MultipathData>(el, (*vecCmp)[j]));
            }
        }

        if (needNotPmp) {
            // 搬送波マルチパスの算出が不要な場合は次へ
            continue;
        }

        // 結果を取得(搬送波位相マルチパス)
        vector<MultipathData> *vecPmp = targetSatObsData[i].getVecPhaseMultipath();
        for (int j = 0; j < (int)vecPmp->size(); j++) {
            // 衛星毎のデータを時刻別/仰角[deg]別にセット
            mapPhaseMp.insert(pair<DataTime, MultipathData>((*vecPmp)[j].time, (*vecPmp)[j]));

            double el = (*vecPmp)[j].azel[1] * R2D; // [deg]
            if (el >= minEl && el <= maxEl) {

                // 観測誤差モデル推定に使うデータをセット
                mapPhaseMpForEl.insert(pair<Elevation, MultipathData>(el, (*vecPmp)[j]));
            }
        }
    }
}

/** 観測誤差モデル推定
 * @brief NavErrEstimater::estimateErrorModel
 * @param deltaBeta 観測更新の単位とする仰角範囲
 * @param sigmaA アプリオリ共分散行列:σａ
 * @param sigmaB アプリオリ共分散行列:σb
 * @param cCode 観測コード(擬似距離)
 * @param pCode 観測コード(搬送波位相)
 * @param outputCmpFilepath 出力ファイルパス(プロット用ファイル、擬似距離マルチパス)
 * @param outputPmpFilepath 出力ファイルパス(プロット用ファイル、搬送波位相マルチパス)
 */
void NavErrEstimater::estimateErrorModel(double deltaBeta, double sigmaA, double sigmaB,
                                         string cCode, string pCode,
                                         string outputCmpFilepath, string outputPmpFilepath) {
    outputLog(TRACE_LEVEL_DEBUG, "deltaBeta=", deltaBeta, "NavErrEstimater::estimateErrorModel");
    outputLog(TRACE_LEVEL_DEBUG, "sigmaA=", sigmaA, "NavErrEstimater::estimateErrorModel");
    outputLog(TRACE_LEVEL_DEBUG, "sigmaB=", sigmaB, "NavErrEstimater::estimateErrorModel");

    estNavErrorData_t estZero = { 0 };
    estZero.sys = useSys;
    estCmpNavError = estZero;
    estPmpNavError = estZero;

    // [[[ 観測誤差モデルを推定(コードマルチパス) ]]]
#ifdef DEBUG_EKF    // debug
    countForEkfDebug = 1;
#endif
    outputLog(TRACE_LEVEL_DEBUG, "estimateErrorModel(CodeMp)", "NavErrEstimater::estimateErrorModel");
    estimatedErrorModelByCmp = estimateErrorModel(mapCodeMpForEl, deltaBeta, sigmaA, sigmaB, estCmpNavError.estCoef);

    // [[[ 観測誤差モデルを推定(搬送波位相マルチパス) ]]]
#ifdef DEBUG_EKF    // debug
    countForEkfDebug = 1;
#endif
    outputLog(TRACE_LEVEL_DEBUG, "estimateErrorModel(phaseMp)", "NavErrEstimater::estimateErrorModel");
    estimatedErrorModelByPmp= estimateErrorModel(mapPhaseMpForEl, deltaBeta, sigmaA, sigmaB, estPmpNavError.estCoef);

    // 推定パラメータの絶対値を取る
    for (int i = 0; i < 2; i++) {
        estCmpNavError.estCoef[i] = fabs(estCmpNavError.estCoef[i]);
        estPmpNavError.estCoef[i] = fabs(estPmpNavError.estCoef[i]);
    }

    // 観測コードセット
    estCmpNavError.code = cCode;
    estPmpNavError.code = pCode;

    // [[[ 推定結果を出力 ]]]
    if (estimatedErrorModelByCmp) {
        ResultWriter writer;
        writer.open(outputCmpFilepath, false);

        // コードマルチパスの誤差モデル出力
        writer.writeLine(RESULT_CMP_MODEL_TAG);
        writer.writeLine(estCmpNavError.estCoef, 2, 12);

        // コードマルチパス出力
        writer.writeLine(RESULT_CMP_TAG);
        outputMultipassData(writer, mapCodeMp);

        writer.close();

        // 観測誤差モデルテーブル出力用に保管
        vecEstNavError.push_back(estCmpNavError);
    }

    if (estimatedErrorModelByPmp) {
        ResultWriter writer;
        writer.open(outputPmpFilepath, false);

        // 搬送波位相マルチパスの誤差モデル出力
        writer.writeLine(RESULT_PMP_MODEL_TAG);
        writer.writeLine(estPmpNavError.estCoef, 2, 12);

        // 搬送波位相マルチパス出力
        writer.writeLine(RESULT_PMP_TAG);
        outputMultipassData(writer, mapPhaseMp);

        writer.close();

        // 観測誤差モデルテーブル出力用に保管
        vecEstNavError.push_back(estPmpNavError);
    }
}

/** マルチパスデータの出力
 * @brief NavErrEstimater::outputMultipassData
 * @param writer 出力クラス
 * @param mapData マルチパスデータ
 */
void NavErrEstimater::outputMultipassData(ResultWriter &writer, multimap<DataTime, MultipathData> &mapData) {
    typedef pair<DataTime, MultipathData> typePair;
    typedef multimap<DataTime, MultipathData>::iterator typeIteMap;

    pair<typeIteMap, typeIteMap> itePair;
    for (typeIteMap ite = mapData.begin(); ite != mapData.end();) {
        DataTime dataTimeClass = (*ite).first;

        // 時刻別に出力
        itePair = mapData.equal_range(dataTimeClass);
        for (ite = itePair.first; ite != itePair.second; ite++) {
            MultipathData *data = &(*ite).second;

            // 結果を取得
            vector<double> vec;
            vec.push_back(data->sat);       // 衛星No
            vec.push_back(data->azel[0]);   // AZ
            vec.push_back(data->azel[1]);   // EL
            vec.push_back(data->amp);       // 単純移動平均結果

            // ファイルに出力
            writer.writeLine(data->time, vec);
        }
    }

}

/** 観測誤差モデルの出力
 * @brief NavErrEstimater::outputErrorModel
 * @param outputFilepath 出力ファイルパス
 * @param atFirst true:先頭行の出力あり
 */
void NavErrEstimater::outputErrorModel(string outputFilepath, bool atFirst) {
    // ファイルオープン
    NavErrWriter writer;
    writer.open(outputFilepath);

    // 算出した観測誤差モデル数分繰り返し
    bool firstFlag = atFirst;
    sta_t *sta = observationManager->getStationParam();
    for (int i = 0; i < (int)vecEstNavError.size(); i++) {
        if (firstFlag) {
            // 1行目は受信機名、アンテナ名を出力
            writer.write(sta->rectype, sta->antdes, vecEstNavError[i]);
            firstFlag = false;

        } else {
            writer.write("", "", vecEstNavError[i]);
        }
    }

    // ファイルクローズ
    writer.close();
}

/** 観測誤差モデルの推定結果[観測誤差モデルテーブル出力用]取得
 * @brief NavErrEstimater::getVecEstNavError
 * @return 観測誤差モデルの推定結果[観測誤差モデルテーブル出力用]
 */
vector<estNavErrorData_t> NavErrEstimater::getVecEstNavError() {
    return vecEstNavError;
}

/** 電離層遅延量変動の計算状態取得
 * @brief NavErrEstimater::getCalculateIono
 * @return true：算出あり、false：算出なし
 */
bool NavErrEstimater::getCalculateIono() {
    return calculateIono;
}

/** EKF観測誤差モデル推定
 * @brief NavErrEstimater::estimateErrorModel
 * @param mapData 推定対象のマルチパスデータのマップ
 * @param deltaBeta 観測更新の単位とする仰角範囲
 * @param sigmaA アプリオリ共分散行列:σａ(標準偏差)
 * @param sigmaB アプリオリ共分散行列:σb(標準偏差)
 * @param estVal [OUT] 推定パラメータ[a,b]
 * @return true：推定実施、false：一回も推定できていない
 */
bool NavErrEstimater::estimateErrorModel(std::multimap<Elevation, MultipathData> &mapData, double deltaBeta,
                                          double sigmaA, double sigmaB, double *estVal) {
    const int estValNum = 2;
    typedef pair<Elevation, MultipathData> typePair;
    typedef multimap<Elevation, MultipathData>::iterator typeIteMap;

    // データ数チェック
    if (mapData.size() == 0) {
        // データなし
        return false;
    }

    // アプリオリ初期共分散行列
    double *tmpPp = zeros(estValNum, estValNum);
    double *Pp = zeros(estValNum, estValNum);
    Pp[0] = sigmaA * sigmaA;
    Pp[1 + 1 * 2] = sigmaB * sigmaB;

    // 推定パラメータ(計算処理用)
    double *tmpXp = zeros(estValNum, 1);

    // 推定パラメータの初期値
    for (int i = 0; i < estValNum; i++) {
        estVal[i] = INITVAL_FOR_EST;
    }

    // 仰角別のループ
    // ※マップデータは仰角[deg]で降順に格納されている
    bool calcFlag = false;
    double targetEl = 90 - deltaBeta;
    pair<typeIteMap, typeIteMap> itePair;
    vector<MultipathData> vecMultipathData;
    for (typeIteMap ite = mapData.begin(); ite != mapData.end();) {
        Elevation elevationClass = (*ite).first;
        // 仰角のチェック
        if (elevationClass.elevation >= targetEl) {
            // 指定の仰角以上なら
            itePair = mapData.equal_range(elevationClass);
            for (ite = itePair.first; ite != itePair.second; ite++) {
                // 推定データの収集
                vecMultipathData.push_back((*ite).second);
            }
            continue;
        }


        // データ数チェック
        int dataNum = vecMultipathData.size();
        if (dataNum < 2) {
            // データが少ない場合はスキップ
            targetEl -= deltaBeta;
            vecMultipathData.clear();
            continue;
        }
#if 0   // debug
        cout << "targetEl=" << targetEl << " dataNum=" << dataNum << endl;
#endif

        // 指定の仰角未満なら収集したデータで推定
        double *vur = mat(dataNum, 1);
        double *Hur = mat(dataNum, estValNum);
        double a2 = SQR(estVal[0]);
        double b2 = SQR(estVal[1]);
        double sum = 0.0;
        for (int i = 0; i < dataNum; i++) {
            // 観測残差ベクトル Vur
            double sinBeta2 = SQR(sin(vecMultipathData[i].azel[1]));
            vur[i] = vecMultipathData[i].mp - (a2 + b2 / sinBeta2);
            sum += vecMultipathData[i].mp;

            // 観測感度行列 Hur
            Hur[i] = 2 * estVal[0];
            Hur[i + dataNum] = 2 * estVal[1] / sinBeta2;

#if 0   // debug
        ostringstream oStr;
        oStr << setprecision(12)
                << "vur[" << i << "]=" << vur[i]
                << " mp=" << vecMultipathData[i].mp
                << " a2=" << a2
                << " b2=" << b2
                << " sinBeta2=" << sinBeta2
                << "(el=" << (vecMultipathData[i].azel[1] * R2D)
                << ") time=" << time_str(vecMultipathData[i].time, 0);
        outputLog(TRACE_LEVEL_DEBUG, oStr.str(), "NavErrEstimater::estimateErrorModel");
#endif
        }

        // 分散を算出
        double avg = sum / dataNum;
        sum = 0.0;
        for (int i = 0; i < dataNum; i++) {
            double val = vecMultipathData[i].mp - avg;
            sum += SQR(val);
        }
        double var = sum / dataNum;


        // 誤差共分散行列 R
        double *R = zeros(dataNum, dataNum);
        for (int i = 0; i < dataNum; i++) {
            R[i + i*dataNum] = var;
        }

        // カルマンフィルタ関数に入力のはHの転置行列
        double *H = zeros(estValNum, dataNum);
        matrixTrans(Hur, H, dataNum , estValNum);

#ifdef DEBUG_EKF    // debug
{
    if (countForEkfDebug > 0) {
        countForEkfDebug--;
        cout << " var=" << var << endl;
        cout << " Hur= " << endl;
        matprint(Hur, dataNum, estValNum, 12, 6);
        cout << " H= " << endl;
        matprint(H, estValNum, dataNum, 12, 6);
        cout << " R= " << endl;
        matprint(R, dataNum, dataNum, 12, 6);
    }
}
#endif


        // カルマンフィルタ
        matcpy(tmpXp, estVal, estValNum, 1);
        matcpy(tmpPp, Pp, estValNum, estValNum);
        int info=filter(tmpXp, tmpPp, H, vur, R, estValNum, dataNum, NULL);
        if (info < 0) {
            // エラー時
            ostringstream oStr;
            oStr <<  "filter error. targetEl=" << targetEl << ends;
            outputLog(TRACE_LEVEL_WARNING, oStr.str(), "NavErrEstimater::estimateErrorModel");

        } else {
#if 0       // 観測誤差を計算
            sum = 0.0;
            a2 = SQR(estVal[0]);
            b2 = SQR(estVal[1]);
            for (int i = 0; i < dataNum; i++) {
                double sinBeta2 = SQR(sin(vecMultipathData[i].azel[1]));
                double val = vecMultipathData[i].mp - (a2 + b2 / sinBeta2);
                sum += SQR(val);
            }
            double rms = sqrt(sum / dataNum);

            // 観測残差が大きい場合は、結果を使用しない
            if (rms < RMS_LIMIT_FOR_MP) {
                matcpy(estVal, tmpXp, estValNum, 1);
                matcpy(Pp, tmpPp, estValNum, estValNum);

                // 計算実施
                calcFlag = true;
            } else {
                // エラー時
                ostringstream oStr;
                oStr << "rms over. rms=" << rms << " targetEl=" << targetEl;
                outputLog(oStr.str(), "NavErrEstimater::estimateErrorModel");
            }
#else
            //結果を反映
            matcpy(estVal, tmpXp, estValNum, 1);
            matcpy(Pp, tmpPp, estValNum, estValNum);

            // 計算実施
            calcFlag = true;
#endif
        }

        // メモリ開放
        free(vur);
        free(Hur);
        free(H);
        free(R);

        // 次の仰角へ
        targetEl -= deltaBeta;
        vecMultipathData.clear();
    }

    // メモリ開放
    free(Pp);
    free(tmpPp);
    free(tmpXp);

    return calcFlag;
}
