﻿/*------------------------------------------------------------------------------
* NavErrEstimater.h : 観測誤差モデル推定クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#ifndef NAVERRESTIMATER_H
#define NAVERRESTIMATER_H
#include <map>
#include "AntennaBase.h"
#include "AntennaEstimater.h"
#include "AntLibException.h"
#include "SatelliteObsData.h"
#include "ResultWriter.h"
#include "NavErrWriter.h"

#define RESULT_CMP_TAG "<cmp>"                  // 結果ファイルの出力タグ：コードマルチパス
#define RESULT_PMP_TAG "<pmp>"                  // 結果ファイルの出力タグ：搬送波位相マルチパス
#define RESULT_CMP_MODEL_TAG "<cmp_model>"      // 結果ファイルの出力タグ：コードマルチパスの観測誤差モデルa,b
#define RESULT_PMP_MODEL_TAG "<pmp_model>"      // 結果ファイルの出力タグ：搬送波位相マルチパスの観測誤差モデルa,b


/** 時刻クラス
 * @brief The DataTime class
 */
class DataTime {
public:
    DataTime(gtime_t time) { dataTime = time; }
    ~DataTime() { }

    bool operator<(const DataTime& target) const {
        double dt = timediff(this->dataTime, target.dataTime);
        return (dt < 0.0);
    }

    gtime_t dataTime;
};

/** 仰角クラス
 * @brief The Elevation class
 */
class Elevation {
public:
    Elevation(double el) { elevation = el; }
    ~Elevation() { }

    bool operator<(const Elevation& target) const {
        double val = elevation -  target.elevation;
        return (val > 0.0); // 仰角の降順
    }

    double elevation;
};


/** 観測誤差モデル推定クラス
 * @brief The NavErrEstimater class
 */
class NavErrEstimater : public AntennaBase
{
public:
    NavErrEstimater();
    ~NavErrEstimater();

    void clearResult();
    virtual void reset();
    void readPhaseErrorFile(std::string inputFilepath);
    void estimateMultipass(ObservationManager *referenceObsMng, prcopt_t opt, int rangeForAMP,
                           double minEl, double maxEl, bool needNotPmp, std::string outputDataDir);
    void estimateErrorModel(double deltaBeta, double sigmaA, double sigmaB,
                            std::string cCode, std::string pCode,
                            std::string outputFilepath, std::string outputPmpFilepath);
    void outputMultipassData(ResultWriter &writer, std::multimap<DataTime, MultipathData> &mapData);
    void outputErrorModel(std::string outputFilepath, bool atFirst);
    std::vector<estNavErrorData_t> getVecEstNavError();
    bool getCalculateIono();

    // -----< 推定用パラメータ >-----
    /** 使用システム
     * @brief useSys
     */
    int useSys;

    /** 使用周波数(0：L1、1：L2、2：L5)
     * @brief useFreq
     */
    int useFreq;


private:
    bool estimateErrorModel(std::multimap<Elevation, MultipathData> &mapData, double deltaBeta,
                                              double sigmaA, double sigmaB, double *estVal);

    /** コードマルチパス(時系列データ)
     * @brief mapCodeMp
     */
    std::multimap<DataTime, MultipathData> mapCodeMp;

    /** 搬送波位相マルチパス(時系列データ)
     * @brief mapPhaseMp
     */
    std::multimap<DataTime, MultipathData> mapPhaseMp;

    /** コードマルチパス(仰角降順)
     * @brief mapCodeMp
     */
    std::multimap<Elevation, MultipathData> mapCodeMpForEl;

    /** 搬送波位相マルチパス(仰角降順)
     * @brief mapPhaseMp
     */
    std::multimap<Elevation, MultipathData> mapPhaseMpForEl;

    /** 衛星別観測データ(測定局)
     * @brief targetSatObsData
     */
    SatelliteObsData targetSatObsData[MAXSAT];

    /** 搬送波コードマルチパスの観測誤差モデルの推定状況(true：推定実施)
     * @brief estimatedErrorModelByCmp
     */
    bool estimatedErrorModelByCmp;

    /** 搬送波コードマルチパスの観測誤差モデルの推定状況(true：推定実施)
     * @brief estimatedErrorModelByPmp
     */
    bool estimatedErrorModelByPmp;

    /** 観測誤差モデルの推定結果(コードマルチパス)
     * @brief estCmpNavError
     */
    estNavErrorData_t estCmpNavError;

    /** 観測誤差モデルの推定結果(搬送波位相マルチパス)
     * @brief estCmpNavError
     */
    estNavErrorData_t estPmpNavError;

    /** 観測誤差モデルの推定結果[観測誤差モデルテーブル出力用]
     * @brief vecEstNavError
     */
    std::vector<estNavErrorData_t> vecEstNavError;

    /** 電離層遅延量変動の計算状態(true：算出あり、false：算出なし)
     * @brief calculateIono
     */
    bool calculateIono;
};

#endif // NAVERRESTIMATER_H
