﻿/*------------------------------------------------------------------------------
* NavigationManager.ccp : 航法データ管理クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "NavigationManager.h"

using namespace std;

/** コンストラクタ
 * @brief NavigationManager::NavigationManager
 */
NavigationManager::NavigationManager()
{
    nav_t navZero = { 0 };
    navData = navZero;
}

/** デストラクタ
 * @brief NavigationManager::~NavigationManager
 */
NavigationManager::~NavigationManager()
{
    // 航法データを削除
    freenav(&navData, SYS_ALL);
}

/** 航法データ設定
 * @brief NavigationManager::setData
 * @param nav 航法データ
 */
void NavigationManager::setData(nav_t *nav) {
    // 過去の航法データを削除
    freenav(&navData, SYS_ALL);

    // 新しい航法データを設定
    navData = *nav;

    // [[[ 航法データの初期化 ]]]
    // carrier wave lengths (m)の設定
    uniqnav(&navData);
}


/** 観測誤差モデルテーブル読込
 * @brief NavigationManager::readNavErr
 * @param filepath ファイルパス
 */
void NavigationManager::readNavErr(string filepath) {
    // 観測誤差モデルテーブルを航法データへ反映
    int ret = readerr(filepath.c_str(), &navData);
    if (ret != 0) {
        throw AntLibException("reading failure ", "NavigationManager::readNavErr");
    }
}

/** 全データ数取得
 * @brief NavigationManager::getTotalNum
 * @return 全データ数
 */
int NavigationManager::getTotalNum() {
    int totalNum = navData.n;
    totalNum += navData.ng;
    totalNum += navData.ns;
    totalNum += navData.ne;
    totalNum += navData.nc;
    totalNum += navData.na;
    totalNum += navData.nt;
    totalNum += navData.nn;
    return totalNum;
}

/** GLONASSの航法データ数取得
 * @brief NavigationManager::getGloNavNum
 * @return 航法データ数取得
 */
int NavigationManager::getGloNavNum() {
    int totalNum = navData.ng;
    return totalNum;
}

/** 全データ取得
 * @brief NavigationManager::getAll
 * @return 全データへのポインタ
 */
nav_t* NavigationManager::getAll() {
    return &navData;
}

/** 航法データをストリームに出力
 * @brief NavigationManager::outputInfo
 * @param oStr 出力ストリーム
 * @param startIndex 開始INDEX(GPS/QZS/GAL ephemeris)
 * @param num データ数(-1：ALL、GPS/QZS/GAL ephemeris)
 */
#define MIN_NUM(n, m) ((n > m)? m: n)
void NavigationManager::outputInfo(ostream &oStr, int startIndex, int num) {
    oStr << "[nav] n=" << navData.n << "," << navData.nmax
        << " ng=" << navData.ng << "," << navData.ngmax
        << " ns=" << navData.ns << "," << navData.nsmax
        << " ne=" << navData.ne << "," << navData.nemax
        << " nc=" << navData.nc << "," << navData.ncmax
        << " na=" << navData.na << "," << navData.namax
        << " nt=" << navData.nt << "," << navData.ntmax
        << " nn=" << navData.nn << "," << navData.nnmax
        << endl;

    int outNum = (num != -1)? MIN_NUM(num,  navData.n): navData.n;
    for (int i = startIndex; i < outNum; i++) {
        string toeStr = time_str(navData.eph[i].toe, 2);
        string tocStr = time_str(navData.eph[i].toc, 2);
        string ttrStr = time_str(navData.eph[i].ttr, 2);

        oStr << "eph i=" << i << " sat=" << navData.eph[i].sat
                << " iode=" << navData.eph[i].iode
                << " iodc=" << navData.eph[i].iodc
                << " sva=" << navData.eph[i].sva
                << " svh=" << navData.eph[i].svh
                << " week=" << navData.eph[i].week
                << " code=" << navData.eph[i].code
                << " flag=" << navData.eph[i].flag;
        oStr << " toe=" << toeStr.c_str()
             << " toc=" << tocStr.c_str()
             << " ttr=" << ttrStr.c_str();
        oStr << " A=" << navData.eph[i].A
                << " e=" << navData.eph[i].e
                << " i0=" << navData.eph[i].i0
                << " OMG0=" << navData.eph[i].OMG0
                << " omg=" << navData.eph[i].omg
                << " M0=" << navData.eph[i].M0
                << " deln=" << navData.eph[i].deln
                << " OMGd=" << navData.eph[i].OMGd
                << " idot=" << navData.eph[i].idot;
        oStr << " crc=" << navData.eph[i].crc
                << " crs=" << navData.eph[i].crs
                << " cuc=" << navData.eph[i].cuc
                << " cus=" << navData.eph[i].cus
                << " cic=" << navData.eph[i].cic
                << " cis=" << navData.eph[i].cis;
        oStr << " toes=" << navData.eph[i].toes
                << " fit=" << navData.eph[i].fit
                << " f0=" << navData.eph[i].f0
                << " f1=" << navData.eph[i].f1
                << " f2=" << navData.eph[i].f2
                << " tgd=" << navData.eph[i].tgd[0]
                    << "," << navData.eph[i].tgd[1]
                    << "," << navData.eph[i].tgd[2]
                    << "," << navData.eph[i].tgd[3];
        oStr << endl;
    }

    outNum = (num != -1)? MIN_NUM(num,  navData.ng): navData.ng;
    for (int i = startIndex; i < outNum; i++) {
        string toeStr = time_str(navData.geph[i].toe, 2);
        string tofStr = time_str(navData.geph[i].tof, 2);

        oStr << "geph i=" << i << " sat=" << navData.geph[i].sat
                << " iode=" << navData.geph[i].iode
                << " frq=" << navData.geph[i].frq
                << " svh=" << navData.geph[i].svh
                << " sva=" << navData.geph[i].sva
                << " age=" << navData.geph[i].age;
        oStr << " toe=" << toeStr.c_str()
                << " tof=" << tofStr.c_str();
        oStr << " pos=" << navData.geph[i].pos[0]
                << "," << navData.geph[i].pos[1]
                << "," << navData.geph[i].pos[2];
        oStr << " vel=" << navData.geph[i].vel[0]
                << "," << navData.geph[i].vel[1]
                << "," << navData.geph[i].vel[2];
        oStr << " acc=" << navData.geph[i].acc[0]
                << "," << navData.geph[i].acc[1]
                << "," << navData.geph[i].acc[2];
        oStr << " taun=" << navData.geph[i].taun
                << " gamn=" << navData.geph[i].gamn
                << " dtaun=" << navData.geph[i].dtaun;
        oStr << endl;
    }
}
