﻿/*------------------------------------------------------------------------------
* ObservationManager.ccp : 観測データ管理クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "ObservationManager.h"

using namespace std;

/** コンストラクタ
 * @brief ObservationManager::ObservationManager
 */
ObservationManager::ObservationManager()
{
    obs_t obsZero = { 0 };
    obsData = obsZero;

    gtime_t timeZero = { 0 };
    startTimeForObsData = timeZero;
    endTimeForObsData = timeZero;

    lastObsDataIndex = 0;

    pcvs_t pcvsZero = { 0 };
    pcvsData = pcvsZero;
}

/** デストラクタ
 * @brief ObservationManager::~ObservationManager
 */
ObservationManager::~ObservationManager()
{
    // 観測データを削除
    freeobs(&obsData);

    // PCVをクリア
    clearPcv();
}

/** 観測データ設定
 * @brief ObservationManager::setData
 * @param obs 観測データ
 * @param sta 局情報
 */
void ObservationManager::setData(obs_t *obs, sta_t *sta) {
    // 過去の観測データを削除
    freeobs(&obsData);

    // 新しい観測データを設定
    obsData = *obs;

    // 新しい局情報を設定
    stationParam = *sta;

    // データ開始時刻と終了時刻を取得
    vecObsDataTime.clear();
    bool initFlag = true;
    gtime_t lastTime = { 0 };
    for (int i = 0; i < obsData.n; i++) {
        // データ時刻を保管
        double dt = timediff(lastTime, obsData.data[i].time);
        if (dt != 0.0) {
            // 前回と違う時刻なら
            obsTimeData_t timeData;
            timeData.time = obsData.data[i].time;
            timeData.valid = true;

            vecObsDataTime.push_back(timeData);
            lastTime = obsData.data[i].time;
        }

        // 初回チェック
        if (initFlag) {
            // 初回データを仮のデータの先頭時刻とする
            initFlag = false;
            startTimeForObsData = obsData.data[i].time;
            endTimeForObsData = obsData.data[i].time;
            continue;
        }

        // データ時刻比較(開始時刻)
        if (startTimeForObsData.time > obsData.data[i].time.time) {
            // 先頭時刻をセット
            startTimeForObsData = obsData.data[i].time;
        }
        // データ時刻比較(終了時刻)
        if (endTimeForObsData.time < obsData.data[i].time.time) {
            // 先頭時刻をセット
            endTimeForObsData = obsData.data[i].time;
        }
    }
}

/** PCV(or ANTEX)ファイル読込
 * @brief ObservationManager::readPcv
 * @param filepath ファイルパス
 */
void ObservationManager::readPcv(string filepath) {
    // 過去のデータをクリア
    clearPcv();

    // PCV or ANTEX読込
    int ret = readpcv(filepath.c_str(), &pcvsData);
    if (ret == 0) {
        throw AntLibException("file open error", "ObservationManager::readPcv");
    }

    // データ数チェック
    if (pcvsData.n <= 0) {
        throw AntLibException("no antenna data", "AntennaBase::readRinex");
    }
}

/** アンテナパラメータ設定
 * @brief ObservationManager::setPcv
 * @param time 時刻
 * @param useSys 使用システム
 * @param antIdx アンテナINDEX
 * @param nav [OUT]航法データにアンテナモデル設定
 * @param popt [OUT]オプションにアンテナモデル設定
 */
void ObservationManager::setPcv(gtime_t time, int useSys, int antIdx, nav_t *nav, prcopt_t *popt) {
    pcv_t *pcv;

    // [[[ 衛星のアンテナパラメータを設定 ]]]
    for (int i = 0; i < MAXSAT; i++) {
        if (!(satsys(i+1,NULL) & useSys)){
            // 対象システムでない場合はスキップ
            continue;
        }
        if (!(pcv = searchpcv(i+1, "", time, &pcvsData))) {
            // 該当時刻の情報がない場合はスキップ
            continue;
        }
        nav->pcvs[i]=*pcv;
    }

    // [[[ 受信機のアンテナパラメータを設定 ]]]
    sta_t *sta = &stationParam;
    double pos[3],del[3];
    strcpy(popt->anttype[antIdx], sta->antdes);
    if (sta->deltype==1) { /* xyz */
        if (norm(sta->pos,3)>0.0) {
            ecef2pos(sta->pos, pos);
            ecef2enu(pos, sta->del, del);
            for (int j = 0; j < 3; j++){
                popt->antdel[antIdx][j] = del[j];
            }
        }
    }
    else { /* enu */
        for (int j = 0; j < 3; j++) {
            popt->antdel[antIdx][j] = sta->del[j];
        }
    }

    if (!(pcv = searchpcv(0, sta->antdes, time, &pcvsData))) {
        // アンテナモデルが見つからない場合
        ostringstream oStr;
        oStr << "not found pcv. [type=" << sta->antdes << "]";
        throw AntLibException(oStr.str(), "AntennaManager::setPcv");
    }
    strcpy(popt->anttype[antIdx], pcv->type);
    popt->pcvr[antIdx]=*pcv;

    // debug
    cout << "### pcv type=" << pcv->type << endl;
    for (int i = 0; i < NFREQ; i++) {
        cout << "###  enu=" << pcv->off[i][0] << "," << pcv->off[i][1] << "," << pcv->off[i][2] << endl;
    }
}

/** 全データ数取得
 * @brief ObservationManager::getTotalNum
 * @return 全データ数
 */
int ObservationManager::getTotalNum() {
    return obsData.n;
}

/** 全観測データ取得
 * @brief ObservationManager::getAll
 * @return 全観測データ
 */
obs_t* ObservationManager::getAll() {
    return &obsData;
}

/** 局情報取得
 * @brief ObservationManager::getStationParam
 * @return 局情報
 */
sta_t* ObservationManager::getStationParam() {
    return &stationParam;
}

/** 観測データを取得(対象時刻)
 * @brief ObservationManager::getData
 * @param targetTime 対象時刻（NULL:データの先頭時刻)
 * @param useDataTime [OUT] 使用したデータ時刻（NULL時は何も返さない)
 * @param startLastIndex true：前回使用した観測データINDEXから、false：最初から
 * @return 観測データ
 */
vector<obsd_t> ObservationManager::getData(gtime_t *targetTime, gtime_t *useDataTime, bool startLastIndex) {
    vector<obsd_t> vecObsdPtr;

    // 開始時刻が指定されていない場合は、データの先頭時刻とする
    gtime_t dataTime = startTimeForObsData;
    if (targetTime != NULL) {
        // 指定された時刻に一番近い時刻を取得
        bool findFlag = false;
        for (int i = 0; i < (int)vecObsDataTime.size(); i++) {
            double dt = timediff(vecObsDataTime[i].time, *targetTime);
            if (dt == 0.0) {
                // 同じ時刻が見つかったら採用する
                dataTime = *targetTime;
                findFlag = true;
                break;
            }
            if (dt > 0.0) {
                // データ時刻が指定された時刻より後だったら、最初のものを採用する
                dataTime = vecObsDataTime[i].time;
                findFlag = true;
                break;
            }
        }

        if (!findFlag) {
            return vecObsdPtr;
        }
    }

    // 前回の使用した観測データの最後から使うかチェック
    if (!startLastIndex) {
        // 継続ではないなら、0クリア
        lastObsDataIndex = 0;
    }

    // 時刻を検索
    // ※検索期間は秒オーダーとする
    for (int i = lastObsDataIndex; i < obsData.n; i++) {
        // 時刻チェック
        double dt = timediff(obsData.data[i].time, dataTime);
        if (dt == 0.0) {
            // 対象時刻を取得
            vecObsdPtr.push_back(obsData.data[i]);
            lastObsDataIndex = i;

        } else if (dt > 0.0) {
            // 未来のデータになったら抜ける
            break;
        }
    }

    // 使用した時刻を返す
    if (useDataTime != NULL) {
        *useDataTime = dataTime;
    }
    return vecObsdPtr;
}

/** データ時刻が観測データ期間に含まれるかチェック(秒オーダー)
 * @brief ObservationManager::checkPeriod
 * @param dataTime データ時刻
 * @return true：期間内、false：期間外
 */
bool ObservationManager::checkPeriod(gtime_t dataTime) {
    if ((dataTime.time >= startTimeForObsData.time)
        && (dataTime.time <= endTimeForObsData.time)) {
        return true;
    }
    return false;
}

/** 指定されたデータ時刻のデータがあるかチェック
 * @brief ObservationManager::containTime
 * @param dataTime
 * @param retIndex 存在した時刻データのINDEX番号(NULL時：戻り値を返さない)
 * @param checkValid true:有効チェック
 * @return true：有、false：無
 */
bool ObservationManager::containTime(gtime_t dataTime, int *retIndex, bool checkValid) {
    for (int i = 0; i < (int)vecObsDataTime.size(); i++) {
        if (checkValid) {
            if (!vecObsDataTime[i].valid) {
                // 無効ならスキップ
                continue;
            }
        }

        double dt = timediff(vecObsDataTime[i].time, dataTime);
        if (dt == 0.0) {
            if (retIndex != NULL) {
                *retIndex = i;
            }
            return true;
        }
    }
    return false;
}

/** 時刻データの有効/無効を設定
 * @brief ObservationManager::setTimeValid
 * @param idx 時刻データのINDEX
 * @param valid 有効/無効(true：有効)
 */
void ObservationManager::setTimeValid(int idx, bool valid) {
    vecObsDataTime[idx].valid = valid;
}

/** LLIによるサイクルスリップ検出
 * @brief ObservationManager::detectSlipbyLLI
 * @param rtk [IN/OUT] スリップ検出結果
 * @param obs 観測データ
 * @param n 観測データ数
 * @param freq 周波数INDEX
 */
void ObservationManager::detectSlipByLLI(rtk_t *rtk, const obsd_t *obs, int n, int freq)
{
    // ppp.c-detslp_ll()を流用
    for (int i = 0; i < n && i < MAXOBS; i++) {
        if (obs[i].L[freq] == 0.0 || !(obs[i].LLI[freq]&3)) {
            continue;
        }

        rtk->ssat[obs[i].sat-1].slip[freq]=1;
    }
}

/** geometry-free phase measurementによるサイクルスリップ検出
 * @brief ObservationManager::detectSlipyGF
 * @param rtk [IN/OUT] スリップ検出結果
 * @param obs 観測データ
 * @param n 観測データ数
 * @param nav 航法データ
 * @param freq 検出対象の周波数
 */
void ObservationManager::detectSlipyGF(rtk_t *rtk, const obsd_t *obs, int n, const nav_t *nav, int freq)
{
    // L1と比較する周波数INDEX
    int useFreqIdx = ANTLIB_FREQ_L2;
    if (freq == ANTLIB_FREQ_L5) {
        useFreqIdx = ANTLIB_FREQ_L5;
    }

    // ppp.c-detslp_gf()を流用
    for (int i = 0; i < n && i < MAXOBS; i++) {
        int satIdx = obs[i].sat-1;
        const double *lam = nav->lam[satIdx];

        if (lam[0] == 0.0
                || lam[useFreqIdx] == 0.0
                || obs->L[0] == 0.0
                || obs->L[useFreqIdx] == 0.0) {
            continue;
        }

        // geometry-free phase measurement
        double g1 = lam[0]*obs->L[0] - lam[useFreqIdx]*obs->L[useFreqIdx];
        double g0 = rtk->ssat[satIdx].gf;
        rtk->ssat[satIdx].gf = g1;

        // 前回値との差からスリップを検出
        if (g0 != 0.0
                && fabs(g1-g0) > rtk->opt.thresslip) {
            rtk->ssat[satIdx].slip[freq] |= 1;
        }
    }
}

/** geometry-free phase measurementによるサイクルスリップ検出
 * @brief ObservationManager::detectSlipyGF2
 * @param rtk [IN/OUT] スリップ検出結果
 * @param obsForRef 観測データ(基準局)
 * @param obsForTarget 観測データ(測定局)
 * @param n 観測データ数
 * @param nav 航法データ
 * @param freq 検出対象の周波数
 */
void ObservationManager::detectSlipyGF2(rtk_t *rtk, const obsd_t *obsForRef, const obsd_t *obsForTarget,
                                       int n, const nav_t *nav, int freq)
{
    // L1と比較する周波数INDEX
    int useFreqIdx = ANTLIB_FREQ_L2;
    if (freq == ANTLIB_FREQ_L5) {
        useFreqIdx = ANTLIB_FREQ_L5;
    }

    // rtkpos.c-detslp_gf()を流用
    for (int i = 0; i < n && i < MAXOBS; i++) {
        int satIdx = obsForRef[i].sat-1;
        const double *lam = nav->lam[satIdx];

        if (lam[0] == 0.0
                || lam[useFreqIdx] == 0.0
                || obsForRef->L[0] == 0.0
                || obsForRef->L[useFreqIdx] == 0.0
                || obsForTarget->L[0] == 0.0
                || obsForTarget->L[useFreqIdx] == 0.0) {
            continue;
        }

        // geometry-free phase measurement
        double g1 = lam[0]*(obsForRef->L[0] - obsForTarget->L[0])
                - lam[useFreqIdx]*(obsForRef->L[useFreqIdx] - obsForTarget->L[useFreqIdx]);
        double g0 = rtk->ssat[satIdx].gf;
        rtk->ssat[satIdx].gf = g1;

        // 前回値との差からスリップを検出
        if (g0 != 0.0
                && fabs(g1-g0) > rtk->opt.thresslip) {
            rtk->ssat[satIdx].slip[freq] |= 1;
        }
    }
}

/** 全データ時刻を取得
 * @brief ObservationManager::getAllDataTime
 * @param checkValid true:有効チェック
 * @return 全データ時刻
 */
vector<gtime_t> ObservationManager::getAllDataTime(bool checkValid) {
    vector<gtime_t> allTime;
    for (int i = 0; i < (int)vecObsDataTime.size(); i++) {
        if (checkValid) {
            if (!vecObsDataTime[i].valid) {
                // 無効ならスキップ
                continue;
            }
        }

        allTime.push_back(vecObsDataTime[i].time);
    }
    return allTime;
}

/** 観測データをストリームに出力
 * @brief ObservationManager::outputInfo
 * @param oStr 出力ストリーム
 * @param startIndex 開始INDEX
 * @param num データ数(-1：ALL)
 */
void ObservationManager::outputInfo(ostream &oStr, int startIndex, int num) {
    oStr << "[obs] n=" << obsData.n << "," << obsData.nmax << endl;

#define FREQ_DATA_UCHAR(name, array) \
    oStr << " " << name << "=" << (int)array[0];    \
    for (int j = 1; j < NFREQ; j++) {               \
        oStr << "," << (int)array[j];               \
    }

#define FREQ_DATA(name, array) \
    oStr << " " << name << "=" << setprecision(12) << array[0];     \
    for (int j = 1; j < NFREQ; j++) {           \
        oStr << "," << array[j];                \
    }

    int outNum = (num != -1)? num: obsData.n;
    for (int i = startIndex; i < outNum; i++) {
        string dataTime = time_str(obsData.data[i].time, 2);

        oStr << "i=" << i << " " << dataTime;
        oStr << " sat=" << (int)obsData.data[i].sat
            << " rcv=" << (int)obsData.data[i].rcv;

        FREQ_DATA_UCHAR("SNR", obsData.data[i].SNR);
        FREQ_DATA_UCHAR("LLI", obsData.data[i].LLI);
        FREQ_DATA_UCHAR("code", obsData.data[i].code);

        FREQ_DATA("L", obsData.data[i].L);
        FREQ_DATA("P", obsData.data[i].P);
        FREQ_DATA("D", obsData.data[i].D);

        oStr << endl;
    }
}

/** PCVをストリームに出力
 * @brief ObservationManager::outputPcv
 * @param oStr 出力ストリーム
 */
void ObservationManager::outputPcv(ostream &oStr) {
    oStr << "[pcv] n=" << pcvsData.n << "," << pcvsData.nmax << endl;
    for (int i = 0; i < pcvsData.n ; i++) {
        // 開始時刻/終了時刻取得
        string startTime = "";
        string endTime = "";
        if (pcvsData.pcv[i].ts.time != 0) {
            startTime = time_str(pcvsData.pcv[i].ts, 2);
        }
        if (pcvsData.pcv[i].te.time != 0) {
            endTime = time_str(pcvsData.pcv[i].te, 2);
        }

        // データ出力
        oStr << " i=" << i
            << " sat=" << pcvsData.pcv[i].sat
            << " type=" << pcvsData.pcv[i].type
            << " code=" << pcvsData.pcv[i].code
            << " ts=" << startTime
            << " te=" << endTime;
        for (int j = 0; j < NFREQ; j++) {
            oStr << " off[" << j << "]=" << pcvsData.pcv[i].off[j][0]
                << "," << pcvsData.pcv[i].off[j][1]
                << "," << pcvsData.pcv[i].off[j][2];
        }
        for (int j = 0; j < NFREQ; j++) {
            oStr << " var[" << j << "]=" << pcvsData.pcv[i].var[j][0];
            for (int k = 1; k < 19; k++) {
                oStr << "," << pcvsData.pcv[i].var[j][k];
            }
        }
        oStr << endl;
    }
}

/** アンテナ位相特性データクリア
 * @brief ObservationManager::clearPcv
 */
void ObservationManager::clearPcv() {
    // PCVデータを解放
    free(pcvsData.pcv);
    pcvsData.pcv = NULL;

    pcvsData.n = pcvsData.nmax = 0;
}
