﻿/*------------------------------------------------------------------------------
* ObservationManager.h : 観測データ管理クラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#ifndef OBSERVATIONMANAGER_H
#define OBSERVATIONMANAGER_H

#include <iostream>
#include <sstream>
#include <vector>
#include <iomanip>
#include "rtklib.h"
#include "AntLibException.h"

#define ANTLIB_FREQ_L1 0               // 使用周波数：L1
#define ANTLIB_FREQ_L2 1               // 使用周波数：L2
#define ANTLIB_FREQ_L5 2               // 使用周波数：L5
#define ARC_BREAK_LIMIT 60  // アーク切れ目の閾値[sec]

/** 時刻データ
 * @brief The obsTimeData_t struct
 */
struct obsTimeData_t {
    /** 時刻
     * @brief time
     */
    gtime_t time;

    /** 有効/無効(true：有効)
     * @brief valid
     */
    bool valid;
};

/** 観測データ管理クラス
 * @brief The ObservationManager class
 */
class ObservationManager
{
public:
    ObservationManager();
    ~ObservationManager();

    void setData(obs_t *obs, sta_t *sta);
    void readPcv(std::string filepath);
    void setPcv(gtime_t time, int useSys, int antIdx, nav_t *nav, prcopt_t *popt);

    int getTotalNum();
    obs_t* getAll();
    sta_t* getStationParam();
    std::vector<obsd_t> getData(gtime_t *targetTime, gtime_t *useDataTime, bool startLastIndex);
    bool checkPeriod(gtime_t dataTime);
    bool containTime(gtime_t dataTime, int *retIndex = NULL, bool checkValid = false);
    std::vector<gtime_t> getAllDataTime(bool checkValid = false);

    // setter
    void setTimeValid(int idx, bool valid);

    // static
    static void detectSlipByLLI(rtk_t *rtk, const obsd_t *obs, int n, int freq);
    static void detectSlipyGF(rtk_t *rtk, const obsd_t *obs, int n, const nav_t *nav, int freq);
    static void detectSlipyGF2(rtk_t *rtk, const obsd_t *obsForRef, const obsd_t *obsForTarget,
                              int n, const nav_t *nav, int freq);

    // debug
    void outputInfo(std::ostream &oStr, int startIndex, int num);
    void outputPcv(std::ostream &oStr);

private:
    void clearPcv();

    /** 観測データ
     * @brief obsData
     */
    obs_t obsData;

    /** 局情報
     * @brief stationParam
     */
    sta_t stationParam;

    /** アンテナ位相特性データ
     * @brief pcvsData
     */
    pcvs_t pcvsData;

    /** 観測データの時刻配列
     * @brief vecObsDataTime
     */
    std::vector<obsTimeData_t> vecObsDataTime;

    /** 観測データ開始時刻
     * @brief startTimeForObsData
     */
    gtime_t startTimeForObsData;

    /** 観測データ終了時刻
     * @brief endTimeForObsData
     */
    gtime_t endTimeForObsData;

    /** 前回取得した最後の観測データINDEX
     * @brief lastObsDataNo
     */
    int lastObsDataIndex;
};

#endif // OBSERVATIONMANAGER_H
