﻿/*------------------------------------------------------------------------------
* PeoridData.ccp : 期間データクラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "PeoridData.h"

/** コンストラクタ
 * @brief PeoridData::PeoridData
 */
PeoridData::PeoridData()
{
    gtime_t timeZero = { 0 };
    startTime = timeZero;
    endTime = timeZero;
}

/** デストラクタ
 * @brief PeoridData::~PeoridData
 */
PeoridData::~PeoridData()
{
}

/** データ期間の設定
 * @brief PeoridData::setPeorid
 * @param sTime データ期間の開始時刻
 * @param eTime データ期間の終了時刻
 */
void PeoridData::setPeorid(gtime_t sTime, gtime_t eTime) {
    startTime = sTime;
    endTime = eTime;
}

/** データ値を設定
 * @brief PeoridData::setData
 * @param dataTime データ時刻
 * @param value 値
 */
void PeoridData::setData(gtime_t dataTime, double value) {
    vecTime.push_back(dataTime);
    vecData.push_back(value);
}

/** 対象期間かチェック
 * @brief PeoridData::contain
 * @param dataTime データ時刻
 * @return true：期間内、false：期間外
 */
bool PeoridData::contain(gtime_t dataTime) {
    double dt = timediff(dataTime, startTime);
    if (dt < 0.0) {
        // 開始時刻より前なら期間外
        return false;
    }

    dt = timediff(dataTime, endTime);
    if (dt > 0.0) {
        //終了時刻より後なら期間外
        return false;
    }

    return true;
}

/** データ値を取得
 * @brief PeoridData::getData
 * @param dataTime データ時刻
 * @return 値
 */
double PeoridData::getData(gtime_t dataTime) {
    // データが一つしかない場合
    if (vecTime.size() == 0) {
        throw AntLibException("no data", "PeoridData::getData");
    } else if (vecTime.size() == 1) {
        return vecData[0];
    }

    // データ範囲内の場合
    int index = -1;
    for (int i = 0; i < (int)vecTime.size(); i++) {
        double dt = timediff(dataTime, vecTime[i]);
        if (dt == 0.0) {
            // 時刻が一致した場合
            return vecData[i];

        } else if (dt < 0.0) {
            // 期間内のデータ時刻が対象のデータ時刻を超えた場合、線形補完
            if (i == 0) {
                // 最初のデータ時刻より前の場合(外挿)
                index  = 0;
            } else {
                index = i - 1;
            }
            break;
        }
    }

    if (index == -1) {
        // 最後ののデータ時刻より後の場合(外挿)
        index = vecTime.size() - 2;
    }

    // 線形近似
    return linear(index, dataTime);
}

/** 全データ取得
 * @brief PeoridData::getAllData
 * @return 全データ
 */
std::vector<double>* PeoridData::getAllData() {
    return &vecData;
}

/** 線形近似
 * @brief PeoridData::linear
 * @param index データINDEX
 * @param dataTime データ時刻
 * @return 近似値
 */
double PeoridData::linear(int index, gtime_t dataTime) {
    double dt = timediff(vecTime[index + 1], vecTime[index]);
    if (dt == 0.0) {
        return vecData[index];
    }

    double dx = vecData[index + 1] - vecData[index];
    double a = dx / dt;
    double b = vecData[index] - a * (vecTime[index].time + vecTime[index].sec);

    return (a * (dataTime.time + dataTime.sec) + b);
}
