﻿/*------------------------------------------------------------------------------
* SatelliteObsData.ccp : 衛星用観測データクラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#include "SatelliteObsData.h"

using namespace std;

//#define DEBUG_SAT_OBSDATA   // デバッグ出力モード

// -----< 値確認データファイル用定義 >-----
#ifdef DEBUG_OUTPUT_DATA
#define DEBUG_OUTPUT_DATA_OPEN(outstream, prefix, sat, suffix, hdrstr)  \
    ofstream outstream;                                                 \
    if (outputDataFile && (vecArcBreakIndex.size() > 0)) {              \
        string filepath = outputDataFileDir + getDataFilename(prefix, sat, suffix); \
        outstream.open(filepath.c_str());                                       \
        outstream << hdrstr << endl;                                    \
    }
#define DEBUG_OUTPUT_DATA_CLOSE(outstream)          \
    if (outputDataFile && outstream.is_open()) {    \
        outstream.close();                          \
    }

#else
#define DEBUG_OUTPUT_DATA_OPEN(outstream, prefix, sat, arcIdx, suffix, hdrstr)
#define DEBUG_OUTPUT_DATA_CLOSE(outstream)
#endif
// -----< 値確認データファイル用定義(end) >-----

#define SQR(x)      ((x)*(x))
#define LAM2FREQ(lam) ((lam != 0.0)? CLIGHT / lam: 1.0)

/** コンストラクタ
 * @brief SatelliteObsData::SatelliteObsData
 */
SatelliteObsData::SatelliteObsData()
{
    outputDataFile = false;
}

/** デストラクタ
 * @brief SatelliteObsData::~SatelliteObsData
 */
SatelliteObsData::~SatelliteObsData()
{
}

/** 内部保持データのクリア
 * @brief SatelliteObsData::clear
 */
void SatelliteObsData::clear() {
    vecObsData.clear();
    vecArcBreakIndex.clear();
    vecIonoVar.clear();
    vecEstPhaseErrorData.clear();
    vecCodeMultipath.clear();
    vecPhaseMultipath.clear();
}

/** 観測データの有無
 * @brief SatelliteObsData::hasData
 * @return true：有
 */
bool SatelliteObsData::hasData() {
    return (vecObsData.size() > 0);
}

/** アーク切れ目有無
 * @brief SatelliteObsData::hasArcBreak
 * @return true：有
 */
bool SatelliteObsData::hasArcBreak() {
    return (vecArcBreakIndex.size() > 0);
}

/** 観測データ設定
 * @brief SatelliteObsData::setObsData
 * @param obsData 観測データ
 */
void SatelliteObsData::setObsData(obsd_t obsData) {
    if (vecObsData.size() == 0) {
        // 観測データを保存するだけ
        vecObsData.push_back(obsData);
        return;
    }

    // 二つ目以降はアーク切れ目をチェック
    int idx = vecObsData.size() - 1;
    double dt = timediff(obsData.time, vecObsData[idx].time);
    if (dt > ARC_BREAK_LIMIT) {
        // アーク切れ目を保存
        vecArcBreakIndex.push_back(idx);
    }

    // 観測データを保存
    vecObsData.push_back(obsData);
}

/** 搬送波位相マルチパス算出用データをセット
 * @brief SatelliteObsData::setPhaseErrorData
 * @param data
 */
void SatelliteObsData::setPhaseErrorData(estPhaseErrorData_t data) {
    // 時刻の並びが逆なので、先頭にセット
    vecEstPhaseErrorData.insert(vecEstPhaseErrorData.begin(), data);
}

/** 電離層遅延量変動の算出
 * @brief SatelliteObsData::calculateIonosphericDelayVariations
 * @param nav 航法データ
 * @param opt オプション(サイクルスリップ検出)
 * @return true:算出なし、false：算出なし(データなし)
 */
bool SatelliteObsData::calculateIonosphericDelayVariations(const nav_t *nav, prcopt_t opt) {
    bool ret = false;
    if (vecObsData.size() == 0) {
        // データがない場合は何もしない
        return ret;
    }

    // 衛星INDEXを取得(全データ衛星IDは同じなので、一つ目を取得)
    int sat = vecObsData[0].sat;
    int satIdx = sat - 1;

    // 波長λ
    const double *satLam = nav->lam[satIdx];

    // 1 - f1^2/f2^2
    int freq1 = ANTLIB_FREQ_L1;
    int freq2 = ANTLIB_FREQ_L2;
    double f1 = LAM2FREQ(satLam[freq1]);
    double f2 = LAM2FREQ(satLam[freq2]);
    double gamma = 1 - SQR(f1)/SQR(f2);

    // 値確認用データファイルのオープン
//    DEBUG_OUTPUT_DATA_OPEN(ofstreamF110, "F110", sat, "PR", "time,arc,P,phai1,phai2,lam1,lam2,Iik")
//    DEBUG_OUTPUT_DATA_OPEN(ofstreamF210, "F210", sat, "PR", "time,arc,dIk,Iik,Ii")

    // アーク切れ目数分処理
    for (int arcIdx = 0; arcIdx <= (int)vecArcBreakIndex.size(); arcIdx++) {
        // 開始INDEX取得
        int startIdx = 0;
        if (arcIdx > 0) {
            startIdx = vecArcBreakIndex[arcIdx - 1] + 1;
        }

        // 終了INDEX取得
        int endIdx = vecObsData.size() - 1;
        if (arcIdx < (int)vecArcBreakIndex.size()) {
            endIdx = vecArcBreakIndex[arcIdx];
        }

        // オプションの設定
        rtk_t rtk = { 0 };
        rtk.opt = opt;

        // 電離層遅延量変動差＆サイクルスリップ検出
        int freqForSlip = ANTLIB_FREQ_L1;
        vector<int> vecSlipIndex;
        double lastIono = 0.0;
        for (int idx = startIdx; idx <= endIdx; idx++) {
            if (vecObsData[idx].L[freq1] == 0.0
                    || vecObsData[idx].L[freq2] == 0.0) {
                continue;
            }

            // iono算出
            double iono = (-vecObsData[idx].L[freq1] * satLam[freq1] + vecObsData[idx].L[freq2] * satLam[freq2]) / gamma;
            if (idx > startIdx) {
                double dt = timediff(vecObsData[idx].time, vecObsData[idx - 1].time);
                double dIono = fabs(iono - lastIono);
                if (dIono > (0.01 * dt)) {
                    // 変動差が超過
                    vecSlipIndex.push_back(idx - 1);
                    lastIono = iono;
                    continue;
                }
            }
            lastIono = iono;

            // サイクルスリップの結果を初期化
            rtk.ssat[satIdx].slip[freqForSlip] = 0;

            // LLIによるサイクルスリップ検出
            ObservationManager::detectSlipByLLI(&rtk, &vecObsData[idx], 1, freqForSlip);

            // geometry-free phase measurementによるサイクルスリップ検出
            ObservationManager::detectSlipyGF(&rtk, &vecObsData[idx], 1, nav, freqForSlip);

            // サイクルスリップ検出チェック
            if (rtk.ssat[satIdx].slip[freqForSlip] == 1) {
                vecSlipIndex.push_back(idx - 1);
            }
        }

#ifdef DEBUG_SAT_OBSDATA
    if (outputStreamForLog != NULL
            && vecSlipIndex.size() > 0) {
        for (int i = 0; i < (int)vecSlipIndex.size(); i++) {
            ostringstream oStr;
            oStr << "vecSlipIndex[" << i << "]=";
            outputLog(oStr.str(), vecObsData[vecSlipIndex[i]].time, "SatelliteObsData::calculateIonosphericDelayVariations");
        }
    }
#endif

        // サイクルスリップ数分処理
        for (int slipIdx = 0; slipIdx <= (int)vecSlipIndex.size(); slipIdx++) {
            // 開始INDEX取得
            int startDataIdx = startIdx;
            if (slipIdx > 0) {
                startDataIdx = vecSlipIndex[slipIdx - 1] + 1;
            }

            // 終了INDEX取得
            int endDataIdx = endIdx;
            if (slipIdx < (int)vecSlipIndex.size()) {
                endDataIdx = vecSlipIndex[slipIdx];
            }

            // 電離層遅延量変動[m]を算出
            double sum = 0;
            vector<double> vecIono;
            vector<int> vecIonoIndex;
            for (int idx = startDataIdx; idx <= endDataIdx; idx++) {
                if (vecObsData[idx].L[freq1] == 0.0
                        || vecObsData[idx].L[freq2] == 0.0) {
                    continue;
                }

                // iono算出
                ret = true;
                double iono = (-vecObsData[idx].L[freq1] * satLam[freq1] + vecObsData[idx].L[freq2] * satLam[freq2]) / gamma;
                sum += iono;

#ifdef DEBUG_SAT_OBSDATA
    if (outputStreamForLog != NULL
            && arcIdx == 0
            && slipIdx == 0) {
        ostringstream oStr;
        oStr << setprecision(12)
            << "iono[" << idx << "]=" << iono
            << " L1=" << vecObsData[idx].L[freq1]
            << " lam1=" << satLam[freq1]
            << " L2=" << vecObsData[idx].L[freq2]
            << " lam2=" << satLam[freq2]
            << " gamma=" << gamma;
        outputLog(oStr.str(), "SatelliteObsData::calculateIonosphericDelayVariations");
    }
#endif
#ifdef DEBUG_OUTPUT_DATA
    if (outputDataFile) {
        // 各アークデータを出力(F110)
        ofstreamF110 << time_str(vecObsData[idx].time, 0)
                    << "," << arcIdx
                    << setprecision(12)
                    << "," << vecObsData[idx].P[0]
                    << "," << vecObsData[idx].L[0]
                    << "," << vecObsData[idx].L[1]
                    << "," << satLam[0]
                    << "," << satLam[1]
                    << "," << iono
                    << endl;
    }
#endif

                // データ追加
                vecIono.push_back(iono);
                vecIonoIndex.push_back(idx);

            }

            if (vecIono.size() < 2) {
                // データ数が少ない場合は無視
                continue;
            }

            // 電離層遅延量の期間を設定
            PeoridData ionoVar;
            ionoVar.setPeorid(vecObsData[startDataIdx].time, vecObsData[endDataIdx].time);

            // 平均値算出
            double avg = sum / vecIono.size();

            // 電離層遅延量の変動を設定
            for (int i = 0; i < (int)vecIono.size(); i++) {
                double iono = vecIono[i] - avg;
                ionoVar.setData(vecObsData[vecIonoIndex[i]].time, iono);
            }

#ifdef DEBUG_OUTPUT_DATA
    if (outputDataFile) {
        // 各アークデータを出力(F210)
        for (int i = 0; i < (int)vecIono.size(); i++) {
            double iono = vecIono[i] - avg;
            ofstreamF210 << time_str(vecObsData[vecIonoIndex[i]].time, 0)
                        << "," << arcIdx
                        << setprecision(12)
                        << "," << iono
                        << "," << vecIono[i]
                        << "," << avg
                        << endl;
        }
    }
#endif

#ifdef DEBUG_SAT_OBSDATA
    if (outputStreamForLog != NULL
            && arcIdx == 0
            && slipIdx == 0) {
        vector<double> *vec = ionoVar.getAllData();
        for (int i = 0; i < (int)vec->size(); i++) {
            ostringstream oStr;
            oStr << "ionoVar[" << i << "]=" << setprecision(12) << vec->at(i);
            outputLog(oStr.str(), "SatelliteObsData::calculateIonosphericDelayVariations");
        }
    }
#endif

            // 結果を保管
            vecIonoVar.push_back(ionoVar);
        }
    }

    // 値確認用データファイルのクローズ
    DEBUG_OUTPUT_DATA_CLOSE(ofstreamF110)
    DEBUG_OUTPUT_DATA_CLOSE(ofstreamF210)

    return ret;
}

/** コード/搬送波位相マルチパス計算
 * @brief SatelliteObsData::calculateMultipath
 * @param freq 使用周波数
 * @param nav 航法データ
 * @param opt オプション(サイクルスリップ検出)
 * @param posTarget 測定局位置(ECEF)
 * @param satRefObsData 衛星用観測データクラス(基準局)
 * @param rangeForAmp 単純移動平均の対象データ数
 * @param cmpLimit コードマルチパスの異常値の閾値[m]
 * @param needNotPmp 搬送波マルチパス算出不要フラグ(true：不要)
 * @return
 */
void SatelliteObsData::calculateMultipath(int freq, const nav_t *nav, prcopt_t opt,
                                        double *posTarget, SatelliteObsData* satRefObsData, int rangeForAmp,
                                        double cmpLimit, bool needNotPmp) {
    if (vecObsData.size() == 0) {
        // データがない場合は何もしない
        return;
    }

    // 衛星INDEXを取得(全データ衛星IDは同じなので、一つ目を取得)
    int sat = vecObsData[0].sat;
    int satIdx = vecObsData[0].sat - 1;

    // 波長を取得
    const double *satLam=nav->lam[satIdx];

    // 係数を取得
    double fact = 1;
    double f1 = LAM2FREQ(satLam[ANTLIB_FREQ_L1]);
    if (freq == ANTLIB_FREQ_L2) {
        // f1^2/f2^2
        double f2 = LAM2FREQ(satLam[ANTLIB_FREQ_L2]);
        fact = SQR(f1)/SQR(f2);
    } else if (freq == ANTLIB_FREQ_L5) {
        // f1^2/f5^2
        double f5 = LAM2FREQ(satLam[ANTLIB_FREQ_L5]);
        fact = SQR(f1)/SQR(f5);
    }

    // 測地座標を取得
    double enuTarget[3];
    ecef2pos(posTarget, enuTarget);

    // オプションの設定
    rtk_t rtk = { 0 };
    rtk.opt = opt;

#ifdef DEBUG_SAT_OBSDATA
    int targetArcForDebug = 1;
#endif

    // 値確認用データファイルのオープン
//    DEBUG_OUTPUT_DATA_OPEN(ofstreamF211, "F211", sat, "PR", "time,arc,slip,P1ui,phai1ui,MP1ik,dMP1ik,MP1i")
//    DEBUG_OUTPUT_DATA_OPEN(ofstreamF222, "F222", sat, "PH", "time,arc,slip,baseSat,Mp1ik,dMp1ik,Mp1i")
//    DEBUG_OUTPUT_DATA_OPEN(ofstreamF310, "F310", sat, "PR", "time,arc,slip,sat,az,el,amp")
//    DEBUG_OUTPUT_DATA_OPEN(ofstreamF320, "F320", sat, "PH", "time,arc,slip,sat,az,el,amp")

    // アーク切れ目数分処理
    int phaseErrorLastIdx = 0;
    for (int arcIdx = 0; arcIdx <= (int)vecArcBreakIndex.size(); arcIdx++) {
        // 開始INDEX取得
        int startIdx = 0;
        if (arcIdx > 0) {
            startIdx = vecArcBreakIndex[arcIdx - 1] + 1;
        }

        // 終了INDEX取得
        int endIdx = vecObsData.size() - 1;
        if (arcIdx < (int)vecArcBreakIndex.size()) {
            endIdx = vecArcBreakIndex[arcIdx];
        }

        // gf変動差＆サイクルスリップ検出
        int freqForSlip = ANTLIB_FREQ_L1;
        vector<int> vecSlipIndex;
        double lastGf = 0.0;
        for (int idx = startIdx; idx <= endIdx; idx++) {
            if (vecObsData[idx].L[freq] == 0.0
                    || vecObsData[idx].P[freq] == 0.0) {
                // 無効データはスキップ
                continue;
            }

            // code-phase
            double gf = vecObsData[idx].P[freq] - vecObsData[idx].L[freq] * satLam[freq];
            if (idx > startIdx) {
                double dGf = fabs(gf - lastGf);
                if (dGf > 3.0) {
                    // 変動差が超過
                    vecSlipIndex.push_back(idx - 1);
                    lastGf = gf;
                    continue;
                }
            }
            lastGf = gf;

            // サイクルスリップの結果を初期化
            rtk.ssat[satIdx].slip[freqForSlip] = 0;

            // LLIによるサイクルスリップ検出
            ObservationManager::detectSlipByLLI(&rtk, &vecObsData[idx], 1, freqForSlip);

            // geometry-free phase measurementによるサイクルスリップ検出
            ObservationManager::detectSlipyGF(&rtk, &vecObsData[idx], 1, nav, freqForSlip);

            // サイクルスリップ検出チェック
            if (rtk.ssat[satIdx].slip[freqForSlip] == 1) {
                vecSlipIndex.push_back(idx - 1);
            }
        }

#ifdef DEBUG_SAT_OBSDATA
    if (outputStreamForLog != NULL
            && vecSlipIndex.size() > 0) {
        for (int i = 0; i < (int)vecSlipIndex.size(); i++) {
            ostringstream oStr;
            oStr << "vecSlipIndex[" << i << "]=";
            outputLog(oStr.str(), vecObsData[vecSlipIndex[i]].time, "SatelliteObsData::calculateMultipath");
        }
    }
#endif

        // サイクルスリップ数分処理
        for (int slipIdx = 0; slipIdx <= (int)vecSlipIndex.size(); slipIdx++) {
            // 開始INDEX取得
            int startDataIdx = startIdx;
            if (slipIdx > 0) {
                startDataIdx = vecSlipIndex[slipIdx - 1] + 1;
            }

            // 終了INDEX取得
            int endDataIdx = endIdx;
            if (slipIdx < (int)vecSlipIndex.size()) {
                endDataIdx = vecSlipIndex[slipIdx];
            }


            // [[[ コードマルチパス＋受信機雑音の変動を算出 ]]]
            double sum = 0.0;
            vector<double> vecMp;
            vector<int> vecMpIndex;
            for (int idx = startDataIdx; idx <= endDataIdx; idx++) {
                if (vecObsData[idx].L[freq] == 0.0
                        || vecObsData[idx].P[freq] == 0.0) {
                    // 無効データはスキップ
                    continue;
                }

                // 電離層遅延量変動を取得
                PeoridData *ionoVar = satRefObsData->getIonoVar(vecObsData[idx].time);
                if (ionoVar == NULL) {
                    // 電離層遅延量変動がない場合はスキップ
                    continue;
                }

                // コードマルチパス＋受信機雑音の計算
                double gf = vecObsData[idx].P[freq] - vecObsData[idx].L[freq] * satLam[freq];
                double iono = ionoVar->getData(vecObsData[idx].time);
                double cmp = gf - (2 * fact * iono);
                if (fabs(cmp) > cmpLimit) {
                    // 異常値はスキップ
                    continue;
                }

#ifdef DEBUG_SAT_OBSDATA
    if (outputStreamForLog != NULL
            && arcIdx == targetArcForDebug
            && slipIdx == 0) {
        ostringstream oStr;
        oStr << setprecision(12)
            << "cmp[" << idx << "]=" << cmp
            << " gf=" << gf
            << "(P=" << vecObsData[idx].P[freq]
            << " L=" << vecObsData[idx].L[freq]
            << " lam=" << satLam[freq]
            << ") iono=" << iono
            << "(time=" << time_str(vecObsData[idx].time, 0)
            << ") fact=" << fact;
        outputLog(oStr.str(), "SatelliteObsData::calculateMultipath");
    }
#endif
                // 値を保管
                sum += cmp;
                vecMp.push_back(cmp);
                vecMpIndex.push_back(idx);
            }

            if (vecMp.size() >= 2) {
                // データ数がある場合のみ、平均値算出
                double avg = sum / vecMp.size();

                // 結果を保管
                double satPos[MAXOBS * 6];
                double satClock[MAXOBS];
                double satVariance[MAXOBS];
                int satHealth[MAXOBS];
                int cmpStartIdx = vecCodeMultipath.size();
                for (int i = 0; i < (int)vecMp.size(); i++) {
                    MultipathData codeMp;
                    int idx = vecMpIndex[i];

                    // 衛星位置算出
                    satposs(vecObsData[idx].time, &vecObsData[idx], 1, nav, EPHOPT_BRDC,
                            satPos, satClock, satVariance, satHealth);

                    // 方向ベクトルを取得
                    double eTarget[3];
                    geodist(satPos, posTarget, eTarget);

                    // AZELを計算
                    satazel(enuTarget, eTarget, codeMp.azel);

                    // vectorに追加
                    codeMp.time = vecObsData[idx].time;
                    codeMp.sat = sat;
                    codeMp.mp = vecMp[i] - avg;

                    vecCodeMultipath.push_back(codeMp);

#ifdef DEBUG_OUTPUT_DATA
                if (outputDataFile) {
                    // 各アークデータを出力(F211)
                    ofstreamF211 << time_str(vecObsData[idx].time, 0)
                                << "," << arcIdx
                                << "," << slipIdx
                                << setprecision(12)
                                << "," << vecObsData[idx].P[freq]
                                << "," << vecObsData[idx].L[freq]
                                << "," << vecMp[i]
                                << "," << codeMp.mp
                                << "," << avg
                                << endl;
                }
#endif
                }

                // 単純移動平均を算出
                int num = vecCodeMultipath.size() - cmpStartIdx;
#ifdef DEBUG_SAT_OBSDATA
    if (outputStreamForLog != NULL
            && arcIdx == targetArcForDebug
            && slipIdx == 0) {
        for (int i = cmpStartIdx; i < (int)vecCodeMultipath.size(); i++) {
            ostringstream oStr;
            oStr << "[pre]vecCodeMultipath[" << i << "]=" << setprecision(12) << vecCodeMultipath[i].mp;
            outputLog(oStr.str(), "SatelliteObsData::calculateMultipath");
        }
    }
#endif
                if (num > 0) {
                    MultipathData::calculateAMP(vecCodeMultipath, cmpStartIdx, vecCodeMultipath.size() - 1, rangeForAmp);

#ifdef DEBUG_OUTPUT_DATA
                    if (outputDataFile) {
                        for (int i = cmpStartIdx; i < (int)vecCodeMultipath.size(); i++) {
                            // 各アークデータを出力(F310)
                            ofstreamF310 << time_str(vecCodeMultipath[i].time, 0)
                                        << "," << arcIdx
                                        << "," << slipIdx
                                        << setprecision(12)
                                        << "," << vecCodeMultipath[i].sat
                                        << "," << vecCodeMultipath[i].azel[0]
                                        << "," << vecCodeMultipath[i].azel[1]
                                        << "," << vecCodeMultipath[i].amp
                                        << endl;
                        }
                    }
#endif
                }
#ifdef DEBUG_SAT_OBSDATA
    if (outputStreamForLog != NULL
            && arcIdx == targetArcForDebug
            && slipIdx == 0) {
        for (int i = cmpStartIdx; i < (int)vecCodeMultipath.size(); i++) {
            ostringstream oStr;
            oStr << "[amp]vecCodeMultipath[" << i << "]=" << setprecision(12) << vecCodeMultipath[i].mp;
            outputLog(oStr.str(), "SatelliteObsData::calculateMultipath");
        }
    }
#endif
            }

        }   // END：サイクルスリップのループ(コードマルチパス)

        if (needNotPmp || vecEstPhaseErrorData.size() == 0) {
            // 搬送波位相マルチパスの算出が不要な場合は次へ
            continue;
        }

        // [[[ 搬送波位相マルチパス＋受信機雑音の変動を算出 ]]]
        // サイクルスリップ数分処理
        for (int slipIdx = 0; slipIdx <= (int)vecSlipIndex.size();) {
            int pmpStartIdx = vecPhaseMultipath.size();
            vector<int> vecPhaseMpIndex;

            // 開始INDEX取得
            int startDataIdx = startIdx;
            if (slipIdx > 0) {
                startDataIdx = vecSlipIndex[slipIdx - 1] + 1;
            }

            // 終了INDEX取得
            int endDataIdx = endIdx;
            if (slipIdx < (int)vecSlipIndex.size()) {
                endDataIdx = vecSlipIndex[slipIdx];
                if (endDataIdx < 0) {
                    // 初回スリップ時
                    slipIdx++;
                    continue;
                }
            }

            // サイクルスリップ別の開始時刻＆終了時刻を取得
            gtime_t startTime = vecObsData[startDataIdx].time;
            gtime_t endTime = vecObsData[endDataIdx].time;

            // 期間内のデータを検索
            bool timeOverFlag = false;
            bool baseSatChange = false;
            int baseSat = vecEstPhaseErrorData[phaseErrorLastIdx].baseSat;
            double sum = 0;
            for (int i = phaseErrorLastIdx; i < (int)vecEstPhaseErrorData.size(); i++) {
                // 時刻チェック
                double dt = timediff(vecEstPhaseErrorData[i].time, startTime);
                if (dt < 0) {
                    // 開始時刻より前なら無視
                    slipIdx++;
                    continue;
                }
                dt = timediff(vecEstPhaseErrorData[i].time, endTime);
                if (dt > 0) {
                    timeOverFlag = true;
                    if (i == phaseErrorLastIdx) {
                        // 期間内にデータがない場合は何もしない
                        break;
                    }
                    // 終了時刻より後なら抜ける
                    phaseErrorLastIdx = i;
                    break;
                }

                // 基準衛星が変わった場合
                if (baseSat != vecEstPhaseErrorData[i].baseSat) {
                    phaseErrorLastIdx = i;
                    timeOverFlag = true;
                    baseSatChange = true;
                    break;
                }

                // 値を保管
                sum += vecEstPhaseErrorData[i].phaseError;
                vecPhaseMpIndex.push_back(i);
            }
            if (!timeOverFlag) {
                // 時間データの最後までいったら、末尾をセット
                phaseErrorLastIdx = vecEstPhaseErrorData.size();
            }

            if (vecPhaseMpIndex.size() >= 2) {
                // データ数がある場合のみ、平均値算出
                double avg = sum / vecPhaseMpIndex.size();

                for (int i = 0; i < (int)vecPhaseMpIndex.size(); i++) {
                    int idx = vecPhaseMpIndex[i];

                    // 期間内のデータをセット
                    MultipathData phaseMp;
                    phaseMp.time = vecEstPhaseErrorData[idx].time;
                    phaseMp.sat = vecEstPhaseErrorData[idx].sat;
                    phaseMp.azel[0] = vecEstPhaseErrorData[idx].azel[0];
                    phaseMp.azel[1] = vecEstPhaseErrorData[idx].azel[1];
                    phaseMp.mp = vecEstPhaseErrorData[idx].phaseError - avg;

                    vecPhaseMultipath.push_back(phaseMp);
#ifdef DEBUG_OUTPUT_DATA
                    if (outputDataFile) {
                        // 各アークデータを出力(F222)
                        ofstreamF222 << time_str(vecEstPhaseErrorData[idx].time, 0)
                                    << "," << arcIdx
                                    << "," << slipIdx
                                    << "," << baseSat
                                    << setprecision(12)
                                    << "," << vecEstPhaseErrorData[idx].phaseError
                                    << "," << phaseMp.mp
                                    << "," << avg
                                    << endl;
                    }
#endif
                }

                // 単純移動平均を算出
                int num = vecPhaseMultipath.size() - pmpStartIdx;
#ifdef DEBUG_SAT_OBSDATA
if (outputStreamForLog != NULL
    && arcIdx == targetArcForDebug
    && slipIdx == 0) {
for (int i = pmpStartIdx; i < (int)vecPhaseMultipath.size(); i++) {
    ostringstream oStr;
    oStr << "[pre]vecPhaseMultipath[" << i << "]=" << setprecision(12) << vecPhaseMultipath[i].mp;
    outputLog(oStr.str(), "SatelliteObsData::calculateMultipath");
}
}
#endif
                if (num > 0) {
                    MultipathData::calculateAMP(vecPhaseMultipath, pmpStartIdx, vecPhaseMultipath.size() - 1, rangeForAmp);

#ifdef DEBUG_OUTPUT_DATA
                    if (outputDataFile) {
                        for (int i = pmpStartIdx; i < (int)vecPhaseMultipath.size(); i++) {
                            // 各アークデータを出力(F320)
                            ofstreamF320 << time_str(vecPhaseMultipath[i].time, 0)
                                        << "," << arcIdx
                                        << "," << slipIdx
                                        << setprecision(12)
                                        << "," << vecPhaseMultipath[i].sat
                                        << "," << vecPhaseMultipath[i].azel[0]
                                        << "," << vecPhaseMultipath[i].azel[1]
                                        << "," << vecPhaseMultipath[i].amp
                                        << endl;
                        }
                    }
#endif
                }
#ifdef DEBUG_SAT_OBSDATA
if (outputStreamForLog != NULL
    && arcIdx == targetArcForDebug
    && slipIdx == 0) {
for (int i = pmpStartIdx; i < (int)vecPhaseMultipath.size(); i++) {
    ostringstream oStr;
    oStr << "[amp]vecPhaseMultipath[" << i << "]=" << setprecision(12) << vecPhaseMultipath[i].mp;
    outputLog(oStr.str(), "SatelliteObsData::calculateMultipath");
}
}
#endif
            }

            if (!baseSatChange) {
                slipIdx++;
            }
        }   // END：サイクルスリップのループ(搬送波マルチパス)
    }   // END：アークの切れ目のループ

    // 値確認用データファイルのクローズ
    DEBUG_OUTPUT_DATA_CLOSE(ofstreamF211)
    DEBUG_OUTPUT_DATA_CLOSE(ofstreamF222)
    DEBUG_OUTPUT_DATA_CLOSE(ofstreamF310)
    DEBUG_OUTPUT_DATA_CLOSE(ofstreamF320)
}

/** 電離層遅延量変動を取得
 * @brief SatelliteObsData::getIonoVar
 * @param dataTime データ時刻
 * @return 電離層遅延量変動データクラス(NULL：該当なし)
 */
PeoridData* SatelliteObsData::getIonoVar(gtime_t dataTime) {
    for (int i = 0; i < (int)vecIonoVar.size(); i++) {
        if (vecIonoVar[i].contain(dataTime)) {
            return &vecIonoVar[i];
        }
    }
    return NULL;
}

/** コードマルチパス・データ取得
 * @brief SatelliteObsData::getVecCodeMultipath
 * @return コードマルチパス・データ
 */
std::vector<MultipathData>* SatelliteObsData::getVecCodeMultipath() {
    return &vecCodeMultipath;
}

/** 搬送波位相マルチパス・データ取得
 * @brief SatelliteObsData::geVecPhaseMultipath
 * @return 搬送波位相マルチパス・データ
 */
std::vector<MultipathData>* SatelliteObsData::getVecPhaseMultipath() {
    return &vecPhaseMultipath;
}

/** アークの切れ目の情報出力
 * @brief SatelliteObsData::outputArcBreak
 */
void SatelliteObsData::outputArcBreak() {
    // アークの切れ目の情報を出力
    for (int arcIdx = 0; arcIdx <= (int)vecArcBreakIndex.size(); arcIdx++) {
        // 開始INDEX取得
        int startIdx = 0;
        if (arcIdx > 0) {
            startIdx = vecArcBreakIndex[arcIdx - 1] + 1;
        }

        // 終了INDEX取得
        int endIdx = vecObsData.size() - 1;
        if (arcIdx < (int)vecArcBreakIndex.size()) {
            endIdx = vecArcBreakIndex[arcIdx];
        }

        ostringstream oStr;
        string startTime = time_str(vecObsData[startIdx].time, 0);
        string endTime = time_str(vecObsData[endIdx].time, 0);
        oStr << "data_seq[" << arcIdx << "]="
             << startTime << " -> " << endTime;
        outputLog(TRACE_LEVEL_DEBUG_DETAIL, oStr.str(), "SatelliteObsData::outputArcBreak");
    }
}

/** 文字列から搬送波位相マルチパス＋受信機雑音データを取得
 * @brief SatelliteObsData::parsePhaseErrorData
 * @param str 文字列(カンマ区切り)
 * @return 搬送波位相マルチパス＋受信機雑音データ
 */
estPhaseErrorData_t SatelliteObsData::parsePhaseErrorData(string str) {
    /* [format]
     * 0:time, 1:satid, 2:satno, 3:base_satno, 4:Q,
     * 5:az, 6:el, 7:phase-error
     */

    // 文字列分割
    vector<string> vecData = splitString(str, ',');

    // 日付を取得
    // YYYY/MM/DD hh:mm:ss
    gtime_t datatime = stringToGtime(vecData[0]);

    // [[[ データ取得 ]]]
    estPhaseErrorData_t data = {0};
    data.time = datatime;
    data.stat = atoi(vecData[4].c_str());
    data.sat = atoi(vecData[2].c_str());
    data.baseSat = atoi(vecData[3].c_str());
    data.azel[0] = atof(vecData[5].c_str());
    data.azel[1] = atof(vecData[6].c_str());
    data.phaseError = atof(vecData[7].c_str());

    return data;
}

/** データファイル名を取得
 * @brief SatelliteObsData::getDataFilename
 * @param prefix 先頭文字列
 * @param satNo 衛星No
 * @param suffix 終了文字列(.dataを除く)
 * @return ファイル名
 */
string SatelliteObsData::getDataFilename(string prefix, int satNo, string suffix) {
    // 右詰めは数値を出力
    ostringstream oStr;
    oStr << prefix << "_";
    oStr << "S" << setfill('0') << setw(2) << right << satNo;
    oStr << "_" << suffix << ".data";

    return oStr.str();
}

