﻿/*------------------------------------------------------------------------------
* SatelliteObsData.h : 衛星用観測データクラス
*
*    Copyright (C) 2014 by Geospatial Information Authority of Japan,
*    All rights reserved.
*    
*    Released under the GPL Licenses.
*
*
*  Original software: ANTTOOL ver.2.1
*
*    Copyright(c) 2007-2011 by T.Takasu, all rights reserved.
*
*
* references :
*
* history : 2015/01/08  1.0  new
*-----------------------------------------------------------------------------*/

#ifndef SATELLITEOBSDATA_H
#define SATELLITEOBSDATA_H

#include <iostream>
#include <fstream>
#include <vector>
#include "rtklib.h"
#include "ObservationManager.h"
#include "AntLibException.h"
#include "AntLibBase.h"
#include "PeoridData.h"
#include "AntLibTool.h"
#include "MultipathData.h"



/** 搬送波位相マルチパス＋受信機雑音データ 保管用構造体
 * @brief The estPhaseErrorData_t struct
 */
struct estPhaseErrorData_t {
    /** データ時刻
     * @brief time
     */
    gtime_t time;

    /** 解ステータス
     * @brief stat
     */
    int stat;

    /** 衛星
     * @brief sat
     */
    int sat;

    /** 基準衛星
     * @brief baseSat
     */
    int baseSat;

    /** AZ/EL
     * @brief azel
     */
    double azel[2];

    /** 搬送波位相マルチパス＋受信機雑音
     * @brief phaseError
     */
    double phaseError;
};


/** 衛星用観測データクラス
 * @brief The SatelliteObsData class
 */
class SatelliteObsData : public AntLibBase
{
public:
    SatelliteObsData();
    virtual ~SatelliteObsData();

    void clear();
    bool hasData();
    bool hasArcBreak();
    void setObsData(obsd_t obsData);
    void setPhaseErrorData(estPhaseErrorData_t data);
    bool calculateIonosphericDelayVariations(const nav_t *nav, prcopt_t opt);
    void calculateMultipath(int freq, const nav_t *nav, prcopt_t opt,
                            double *posTarget, SatelliteObsData* satRefObsData, int rangeForAmp,
                            double cmpLimit, bool needNotPmp);

    // getter
    PeoridData* getIonoVar(gtime_t datatime);
    std::vector<MultipathData>* getVecCodeMultipath();
    std::vector<MultipathData>* getVecPhaseMultipath();

    // log
    void outputArcBreak();

    // static
    static estPhaseErrorData_t parsePhaseErrorData(std::string str);
    static std::string getDataFilename(std::string prefix, int satNo, std::string suffix);

    /** 値確認データファイルへの出力(true：出力)
     * @brief outputDataFile
     */
    bool outputDataFile;

    /** 値確認データファイルの出力先('/'で終わること)
     * @brief outputDataFileDir
     */
    std::string outputDataFileDir;

private:
    /** 観測データ
     * @brief vecObsData
     */
    std::vector<obsd_t> vecObsData;

    /** アーク切れ目の観測データのINDEX(観測データの一番最後は含まない)
     * @brief vecArcBreakIndex
     */
    std::vector<int> vecArcBreakIndex;

    /** 電離層遅延量の変動
     * @brief vecIonoData
     */
    std::vector<PeoridData> vecIonoVar;

    /** 搬送波位相マルチパス＋受信機雑音データ
     * @brief vecEstPhaseErrorData
     */
    std::vector<estPhaseErrorData_t> vecEstPhaseErrorData;

    /** コードマルチパス・データ
     * @brief vecCodeMultipath
     */
    std::vector<MultipathData> vecCodeMultipath;

    /** 搬送波位相マルチパス・データ
     * @brief vecPhaseMultipath
     */
    std::vector<MultipathData> vecPhaseMultipath;
};

#endif // SATELLITEOBSDATA_H
